
/*
#    Sfront, a SAOL to C translator    
#    This file: Included file in sfront runtime
#
# Copyright (c) 1999-2006, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  Redistributions of source code must retain the above copyright
#  notice, this list of conditions and the following disclaimer.
#
#  Redistributions in binary form must reproduce the above copyright
#  notice, this list of conditions and the following disclaimer in the
#  documentation and/or other materials provided with the distribution.
#
#  Neither the name of the University of California, Berkeley nor the
#  names of its contributors may be used to endorse or promote products
#  derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#    Maintainer: John Lazzaro, lazzaro@cs.berkeley.edu
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <signal.h>

/********************************/
/* readabiliy-improving defines */
/********************************/

#define ROUND(x) ( ((x) > 0.0F) ? ((int) ((x) + 0.5F)) :  ((int) ((x) - 0.5F)))
#define POS(x)   (((x) > 0.0F) ? x : 0.0F)
#define RMULT ((float)(1.0F/(RAND_MAX + 1.0F)))

#define NOTUSEDYET 0
#define TOBEPLAYED 1
#define PAUSED     2
#define PLAYING    3
#define ALLDONE    4

#define NOTLAUNCHED 0
#define LAUNCHED 1

#define ASYS_DONE        0
#define ASYS_EXIT        1
#define ASYS_ERROR       2

#define IPASS 1
#define KPASS 2
#define APASS 3

#define IOERROR_RETRY 256 

/* integer type that holds a cast pointer */

#define PSIZE long 

/************************************/
/*  union for a data stack element  */
/************************************/

typedef union {

float f;
int i;
unsigned int ui;
PSIZE ps;

} dstack;

/************************************/
/* ntables: table entries for notes */
/************************************/

typedef struct tableinfo {

int    len;                /* length of table             */
float  lenf;               /* length of table, as a float */

int    start;              /* loop start position  */
int    end;                /* loop end position    */
float  sr;                 /* table sampling rate  */
float  base;               /* table base frequency */

                           /* precomputed constants       */
int tend;                  /* len -1 if end==0            */
float oconst;              /* len*ATIME                   */

unsigned int dint;         /* doscil: 64-bit phase incr   */
unsigned int dfrac;
                           /* doscil: sinc interpolation        */
unsigned int sfui;         /* scale_factor as unsigned int      */
float sffl;                /* scale_factor as a float           */
unsigned int dsincr;       /* sinc pointer increment (d=doscil) */

float  *t;                 /* pointer to table entries    */
float stamp;               /* timestamp on table contents */
char llmem;                /* 1 if *t was malloced        */

} tableinfo; 

/********************/
/*   tempo lines    */
/********************/

typedef struct stempo_lines {

  float t;          /* trigger time */
  float newtempo;   /* new tempo */ 

} stempo_lines;

/********************/
/*   table lines    */
/********************/

typedef struct stable_lines {

  float t;          /* trigger time */
  int gindex;       /* global table to target */
  int size;         /* size of data */
  void (*tmake) (); /* function   */
  void * data;      /* data block */

} stable_lines;


/* defines for a rentrant sa.c file */

#define ENGINE_NONREENTRANT  0
#define ENGINE_REENTRANT     1

#define ENGINE_STYLE  ENGINE_REENTRANT
#define ENGINE_PTR_TYPE  engine *
#define ENGINE_PTR_TYPE_COMMA  engine *,
#define ENGINE_PTR_DECLARE  engine * eng
#define ENGINE_PTR_DECLARE_COMMA  engine * eng,
#define ENGINE_PTR_DECLARE_SEMICOLON  engine * eng;
#define ENGINE_PTR  eng
#define ENGINE_PTR_COMMA  eng,
#define ENGINE_PTR_ASSIGNED_TO  eng = 
#define ENGINE_PTR_CREATE_SEMICOLON  engine * eng = (engine *) calloc(1, sizeof(engine));
#define ENGINE_SIZE  (sizeof(engine))
#define ENGINE_PTR_NULLRETURN_SEMICOLON  if (!eng) return NULL;
#define ENGINE_PTR_IS_NULL  (eng == NULL)
#define ENGINE_PTR_IS_NOT_NULL  (eng != NULL)
#define ENGINE_PTR_DESTROY_SEMICOLON  free(eng);
#define ENGINE_PTR_RETURN_SEMICOLON  return eng;
#define EV(x)   eng->x

#define NV(x)   nstate->v[x].f
#define NVI(x)  nstate->v[x].i
#define NVUI(x) nstate->v[x].ui
#define NVPS(x) nstate->v[x].ps
#define NVU(x)  nstate->v[x]
#define NT(x)   nstate->t[x]
#define NS(x)   nstate->x
#define NSP     ENGINE_PTR_COMMA nstate
#define NP(x)   nstate->v[x].f
#define NPI(x)  nstate->v[x].i
#define NPUI(x) nstate->v[x].ui

#define NPPS(x) nstate->v[x].ps

#define NG(x)   EV(global)[x].f
#define NGI(x)  EV(global)[x].i
#define NGUI(x) EV(global)[x].ui
#define NGU(x)  EV(global)[x]
#define TB(x)   EV(bus[x])
#define STB(x)  EV(sbus[x])

#define MAXPFIELDS 1

typedef struct instr_line { 

float starttime;  /* score start time of note */
float endtime;    /* score end time of note */
float startabs;   /* absolute start time of note */
float endabs;     /* absolute end time of note */
float abstime;    /* absolute time extension */
float time;       /* time of note start (absolute) */
float itime;      /* elapsed time (absolute) */
float sdur;       /* duration of note in score time*/

int kbirth;       /* kcycleidx for note launch */
int released;     /* flag for turnoff*/
int turnoff;      /* flag for turnoff */
int noteon;       /* NOTYETUSED,TOBEPLAYED,PAUSED,PLAYING,ALLDONE */
int notestate;    /* index into state array */
int launch;       /* only for dynamic instruments */
int numchan;      /* only for MIDI notes */
int preset;       /* only for MIDI notes */
int notenum;      /* only for MIDI notes */
int label;        /* SASL label index + 1, or 0 (no label) */
                  /* for static MIDI, all-sounds noteoff */

float p[MAXPFIELDS];  /* parameters */

struct ninstr_types * nstate; /* pointer into state array */

} instr_line;

typedef struct scontrol_lines {

float t;                  /* trigger time */
int label;                /* index into label array */
int siptr;                /* score instr line to control */
instr_line * iline;       /* pointer to score line */
int imptr;                /* position of variable in v[] */
float imval;              /* value to import into v[] */

} scontrol_lines;

typedef struct csys_table_ptr {

int num;                  /* number of floats in array */
float * t;                /* pointer to array */

} csys_table_ptr;


#define MAXCNOTES 256

#define CSYS_INSTRNUM 1

#define CSYS_IRATE 0
#define CSYS_KRATE 1
#define CSYS_ARATE 2
#define CSYS_TABLE 3

#define CSYS_NORMAL 0
#define CSYS_IMPORT 1
#define CSYS_EXPORT 2
#define CSYS_IMPORTEXPORT 3
#define CSYS_PFIELD 4
#define CSYS_INTERNAL 5

#define CSYS_UNUSED 0
#define CSYS_READ 1
#define CSYS_WRITTEN 2
#define CSYS_WRITTEN_AND_READ 3

#define CSYS_STATUS_EFFECTS    1 
#define CSYS_STATUS_SCORE      2 
#define CSYS_STATUS_MIDI       4 
#define CSYS_STATUS_DYNAMIC    8 
#define CSYS_STATUS_STARTUP   16 

typedef struct csys_varstruct {
  int index;
  char * name;
  int token;
  int type;
  int tag;
  int width;
  int use;
} csys_varstruct;

typedef struct csys_instrstruct {
  int index;
  char * name;
  int token;
  int numvars;
  csys_varstruct * vars;
  int outwidth;
  int status;
} csys_instrstruct;

typedef struct csys_labelstruct {
  int index;
  char * name;
  int token;
  int iflag[CSYS_INSTRNUM];
} csys_labelstruct;

typedef struct csys_presetstruct {
  int index;
  int preset;
} csys_presetstruct;

typedef struct csys_samplestruct {
  int index;
  int token;
  char * name;
  char * fname;
} csys_samplestruct;

typedef struct csys_routestruct {
  int bus;
  int ninstr;
  int * instr;
} csys_routestruct;

typedef struct csys_sendstruct {
  int instr;
  int nbus;
  int * bus;
} csys_sendstruct;

typedef struct csys_busstruct {
  int index;
  char * name;
  int width;
  int oflag;
} csys_busstruct;

typedef struct csys_targetstruct {
char * name;
int token;
int numinstr;
int * instrindex;
int * varindex;
} csys_targetstruct;


int csys_sfront_argc = 21;

char * csys_sfront_argv[21] = {
"sfront",
"-gcc",
"-cpp",
"-ain",
"audiounit",
"-cin",
"aucontrol",
"-aout",
"audiounit",
"-au_ui_name",
"Spectral Median Filter",
"-au_component_subtype",
"spcf",
"-au_component_manu",
"ucBe",
"-au_ui_manu",
"John Lazzaro",
"-au_manu_url",
"edu.berkeley.eecs",
"-orc",
"spc.saol"};



typedef struct engine {

/* audio and control rates */

float globaltune;
float invglobaltune;
float tempo;
float scoremult;
float scorebeats;          /* current score beat */
float absolutetime;        /* current absolute time */
int kbase;                 /* kcycle of last tempo change */
float scorebase;           /* scorebeat of last tempo change */

/* counters & bounds acycles and kcycles */

int endkcycle;
int kcycleidx;
int acycleidx;
int pass;
int beginflag;
sig_atomic_t graceful_exit;

int nextstate;     /* counter for active instrument state */
int oldstate;      /* detects loops in nextstate updates  */
float cpuload;     /* current cpu-time value              */

#define APPNAME "sfront"
#define APPVERSION "--IDSTRING--"
#define NSYS_NET 0
#define INTERP_LINEAR 0
#define INTERP_SINC 1
#define INTERP_TYPE INTERP_LINEAR

#define SAOL_SRATE 44100.0F
#define SAOL_KRATE 100.0F

float ARATE;       /* audio sample rate                   */
float ATIME;       /* audio sampling period               */
unsigned int ACYCLE;  /* number of samples in an a-cycle  */

float KRATE;       /* control sample rate                 */
float KTIME;       /* control sampling period             */
float KMTIME;      /* control sampling period (ms)        */
int KUTIME;        /* control sampling period (us)        */

#define ASYS_RENDER   0
#define ASYS_PLAYBACK 1
#define ASYS_TIMESYNC 2

#define ASYS_SHORT   0
#define ASYS_FLOAT   1

#define ASYS_AUDIOUNIT_AU_BUNDLECF CFSTR("edu.berkeley.eecs.audiounit.spc")

#define ASYS_AUDIOUNIT_SUPPORT_PROPERTY_RENDERQUALITY  (0)
#define ASYS_AUDIOUNIT_SUPPORT_PROPERTY_STREAMFROMDISK (0)
#define ASYS_AUDIOUNIT_SUPPORT_PROPERTY_CPULOAD (0)

/* audio i/o */

#define ASYS_OCHAN 2
#define ASYS_OTYPENAME  ASYS_FLOAT
#define ASYS_OTYPE  float

int asys_exit_status;

#define ASYS_ICHAN 2
#define ASYS_ITYPENAME  ASYS_FLOAT
#define ASYS_ITYPE  float
#define ASYS_ACTIVE_IO

#define ASYS_USERLATENCY  0
#define ASYS_LOWLATENCY   0
#define ASYS_HIGHLATENCY  1
#define ASYS_LATENCYTYPE  ASYS_LOWLATENCY
#define ASYS_LATENCY 0.002000F
#define ASYS_TIMEOPTION ASYS_PLAYBACK

/* AIF or WAV output file wordsize */

#define ASYS_OUTFILE_WORDSIZE_8BIT  0
#define ASYS_OUTFILE_WORDSIZE_16BIT  1
#define ASYS_OUTFILE_WORDSIZE_24BIT  2
#define ASYS_OUTFILE_WORDSIZE  1

int asys_argc;
char ** asys_argv;

int csys_argc;
char ** csys_argv;

void * csys_state;

float position[3];
float direction[3];
float listenerPosition[3];
float listenerDirection[3];
float minFront;
float maxFront;
float minBack;
float maxBack;
float params[128];

#define BUS_output_bus 0
#define ENDBUS_output_bus 2
#define BUS_input_bus 2
#define ENDBUS_input_bus 4
#define ENDBUS 4

float bus[ENDBUS];

float fakeMIDIctrl[256];

#define GBL_STARTVAR 0
#define GBL_aup_Spectral_Parameters 0
#define CSYS_SASL_GBL_aup_Spectral_Parameters 0
#define GBL_aup_Spectral_Parameters_display_menu 1
#define CSYS_SASL_GBL_aup_Spectral_Parameters_display_menu 1
#define GBL_aup_Spectral_Parameters_idx0__ 2
#define CSYS_SASL_GBL_aup_Spectral_Parameters_idx0__ 2
#define GBL_aup_Spectral_Parameters_pinfo 3
#define GBL_aup_Spectral_Center 4
#define CSYS_SASL_GBL_aup_Spectral_Center 4
#define GBL_aup_Spectral_Center_slider_log 5
#define CSYS_SASL_GBL_aup_Spectral_Center_slider_log 5
#define GBL_aup_Spectral_Center_unit_Hz 6
#define CSYS_SASL_GBL_aup_Spectral_Center_unit_Hz 6
#define GBL_aup_Spectral_Center_pinfo 7
#define GBL_aup_Treble_Meter 8
#define CSYS_SASL_GBL_aup_Treble_Meter 8
#define GBL_aup_Treble_Meter_slider_linear 9
#define CSYS_SASL_GBL_aup_Treble_Meter_slider_linear 9
#define GBL_aup_Treble_Meter_unit_dB 10
#define CSYS_SASL_GBL_aup_Treble_Meter_unit_dB 10
#define GBL_aup_Treble_Meter_pinfo 11
#define GBL_aup_Servo_Loop_Parameters 12
#define CSYS_SASL_GBL_aup_Servo_Loop_Parameters 12
#define GBL_aup_Servo_Loop_Parameters_display_menu 13
#define CSYS_SASL_GBL_aup_Servo_Loop_Parameters_display_menu 13
#define GBL_aup_Servo_Loop_Parameters_idx0__ 14
#define CSYS_SASL_GBL_aup_Servo_Loop_Parameters_idx0__ 14
#define GBL_aup_Servo_Loop_Parameters_pinfo 15
#define GBL_aup_Tracking_Time 16
#define CSYS_SASL_GBL_aup_Tracking_Time 16
#define GBL_aup_Tracking_Time_slider_log 17
#define CSYS_SASL_GBL_aup_Tracking_Time_slider_log 17
#define GBL_aup_Tracking_Time_unit_s 18
#define CSYS_SASL_GBL_aup_Tracking_Time_unit_s 18
#define GBL_aup_Tracking_Time_pinfo 19
#define GBL_aup_Threshold 20
#define CSYS_SASL_GBL_aup_Threshold 20
#define GBL_aup_Threshold_slider_log 21
#define CSYS_SASL_GBL_aup_Threshold_slider_log 21
#define GBL_aup_Threshold_unit_dB 22
#define CSYS_SASL_GBL_aup_Threshold_unit_dB 22
#define GBL_aup_Threshold_pinfo 23
#define GBL_ENDVAR 24

/* SAOL global block variables */

dstack global[GBL_ENDVAR+1];

#define TBL_GBL_aup_Spectral_Parameters_pinfo 0
#define CSYS_SASL_GBL_aup_Spectral_Parameters_pinfo 0
#define TBL_GBL_aup_Spectral_Center_pinfo 1
#define CSYS_SASL_GBL_aup_Spectral_Center_pinfo 1
#define TBL_GBL_aup_Treble_Meter_pinfo 2
#define CSYS_SASL_GBL_aup_Treble_Meter_pinfo 2
#define TBL_GBL_aup_Servo_Loop_Parameters_pinfo 3
#define CSYS_SASL_GBL_aup_Servo_Loop_Parameters_pinfo 3
#define TBL_GBL_aup_Tracking_Time_pinfo 4
#define CSYS_SASL_GBL_aup_Tracking_Time_pinfo 4
#define TBL_GBL_aup_Threshold_pinfo 5
#define CSYS_SASL_GBL_aup_Threshold_pinfo 5
#define GBL_ENDTBL 6

tableinfo gtables[GBL_ENDTBL+1];

#define centroid_loudness_bass 0
#define centroid_loudness_treble 1
#define centroid_logfloor 2
#define centroid_gfactor 3
#define centroid_sc_reljump 4
#define centroid_linfloor 5
#define centroid_onedb_time 6
#define centroid_db_per_kcycle 7
#define centroid_capture_range 8
#define centroid_tracking_range 9
#define centroid_amp 10
#define centroid_pi 11
#define centroid_sr3 12
#define centroid_acycle 13
#define centroid_in_rms 14
#define centroid_lp_rms 15
#define centroid_out_rms 16
#define centroid_lo_db 17
#define centroid_hi_db 18
#define centroid_in_db 19
#define centroid_out_db 20
#define centroid_state 21
#define centroid_tilt_db 22
#define centroid_last_tilt_db 23
#define centroid_lgain_start 24
#define centroid_hgain_start 25
#define centroid_lmult 26
#define centroid_hmult 27
#define centroid_sc_update 28
#define centroid_sc_interp 29
#define centroid_omega_start 30
#define centroid_omega_mult 31
#define centroid_last_Spectral_Center 32
#define centroid_in_log 33
#define centroid_lp_log 58
#define centroid_out_log 83
#define centroid_bptr 108
#define centroid_bidx 109
#define centroid_mlog 110
#define centroid_mmass 230
#define centroid_nextm 231
#define centroid_mptr 232
#define centroid_midx 233
#define centroid_qcount 234
#define centroid_quick_set 235
#define centroid_quick_go 236
#define centroid_last_mmass 237
#define centroid_mup_start 238
#define centroid_mup_mult 239
#define centroid_lgain 240
#define centroid_hgain 241
#define centroid_acount 242
#define centroid_mup 243
#define centroid_omega 244
#define centroid_n 245
#define centroid_a0 246
#define centroid_b1 247
#define centroid_lp 248
#define centroid_out 249
#define centroid_in_mono 250
#define centroid_in_perceptual 251
#define centroid_lp_mono 252
#define centroid_lp_perceptual 253
#define centroid_out_mono 254
#define centroid_out_perceptual 255
#define centroid__tvr10 256
#define centroid__tvr9 257
#define centroid__tvr8 258
#define centroid__tvr7 259
#define centroid__tvr6 260
#define centroid__tvr5 261
#define centroid__tvr4 262
#define centroid__tvr3 263
#define centroid__tvr2 264
#define centroid__tvr1 265
#define centroid__tvr0 266
#define centroid_log1_return 267
#define centroid_ampdb2_return 268
#define centroid_pow3_return 269
#define centroid_rms4_buffer 270
#define centroid_rms4_scale 271
#define centroid_rms4_kcyc 272
#define centroid_rms4_x 273
#define centroid_rms4_return 274
#define centroid_rms5_buffer 275
#define centroid_rms5_scale 276
#define centroid_rms5_kcyc 277
#define centroid_rms5_x 278
#define centroid_rms5_return 279
#define centroid_rms6_buffer 280
#define centroid_rms6_scale 281
#define centroid_rms6_kcyc 282
#define centroid_rms6_x 283
#define centroid_rms6_return 284
#define centroid_sqrt7_return 285
#define centroid_sqrt8_return 286
#define centroid_sqrt9_return 287
#define centroid_dbamp10_return 288
#define centroid_dbamp11_return 289
#define centroid_dbamp12_return 290
#define centroid_dbamp13_return 291
#define centroid_sgn14_x 292
#define centroid_sgn14_return 293
#define centroid_int15_return 294
#define centroid_int16_return 295
#define centroid_pow17_return 296
#define centroid_pow18_return 297
#define centroid_abs19_return 298
#define centroid_abs20_return 299
#define centroid_sgn21_x 300
#define centroid_sgn21_return 301
#define centroid_abs22_return 302
#define centroid_sgn23_x 303
#define centroid_sgn23_return 304
#define centroid_exp24_return 305
#define centroid_exp25_return 306
#define centroid_exp26_return 307
#define centroid_exp27_return 308
#define centroid_exp28_return 309
#define centroid_exp29_return 310
#define centroid_exp30_return 311
#define centroid_exp31_return 312
#define centroid_exp32_return 313
#define centroid_exp33_return 314
#define centroid_exp34_return 315
#define centroid_exp35_return 316
#define centroid_exp36_return 317
#define centroid_exp37_return 318
#define centroid_exp38_return 319
#define centroid_exp39_return 320
#define centroid_exp40_return 321
#define centroid_exp41_return 322
#define centroid_exp42_return 323
#define centroid_exp43_return 324
#define centroid_lopass44_d2 325
#define centroid_lopass44_d1 326
#define centroid_lopass44_a2 327
#define centroid_lopass44_a1 328
#define centroid_lopass44_b2 329
#define centroid_lopass44_b1 330
#define centroid_lopass44_b0 331
#define centroid_lopass44_ocut 332
#define centroid_lopass44_kcyc 333
#define centroid_lopass44_cut 334
#define centroid_lopass44_input 335
#define centroid_lopass44_return 336
#define centroid_hipass45_d2 337
#define centroid_hipass45_d1 338
#define centroid_hipass45_a2 339
#define centroid_hipass45_a1 340
#define centroid_hipass45_b2 341
#define centroid_hipass45_b1 342
#define centroid_hipass45_b0 343
#define centroid_hipass45_ocut 344
#define centroid_hipass45_kcyc 345
#define centroid_hipass45_cut 346
#define centroid_hipass45_input 347
#define centroid_hipass45_return 348
#define centroid_lopass46_d2 349
#define centroid_lopass46_d1 350
#define centroid_lopass46_a2 351
#define centroid_lopass46_a1 352
#define centroid_lopass46_b2 353
#define centroid_lopass46_b1 354
#define centroid_lopass46_b0 355
#define centroid_lopass46_ocut 356
#define centroid_lopass46_kcyc 357
#define centroid_lopass46_cut 358
#define centroid_lopass46_input 359
#define centroid_lopass46_return 360
#define centroid_hipass47_d2 361
#define centroid_hipass47_d1 362
#define centroid_hipass47_a2 363
#define centroid_hipass47_a1 364
#define centroid_hipass47_b2 365
#define centroid_hipass47_b1 366
#define centroid_hipass47_b0 367
#define centroid_hipass47_ocut 368
#define centroid_hipass47_kcyc 369
#define centroid_hipass47_cut 370
#define centroid_hipass47_input 371
#define centroid_hipass47_return 372
#define centroid_lopass48_d2 373
#define centroid_lopass48_d1 374
#define centroid_lopass48_a2 375
#define centroid_lopass48_a1 376
#define centroid_lopass48_b2 377
#define centroid_lopass48_b1 378
#define centroid_lopass48_b0 379
#define centroid_lopass48_ocut 380
#define centroid_lopass48_kcyc 381
#define centroid_lopass48_cut 382
#define centroid_lopass48_input 383
#define centroid_lopass48_return 384
#define centroid_hipass49_d2 385
#define centroid_hipass49_d1 386
#define centroid_hipass49_a2 387
#define centroid_hipass49_a1 388
#define centroid_hipass49_b2 389
#define centroid_hipass49_b1 390
#define centroid_hipass49_b0 391
#define centroid_hipass49_ocut 392
#define centroid_hipass49_kcyc 393
#define centroid_hipass49_cut 394
#define centroid_hipass49_input 395
#define centroid_hipass49_return 396
#define centroid_ENDVAR 397

#define TBL_centroid_rms4_buffer 0
#define TBL_centroid_rms5_buffer 1
#define TBL_centroid_rms6_buffer 2
#define centroid_ENDTBL 3

#define centroid1_loudness_bass 0
#define centroid1_loudness_treble 1
#define centroid1_logfloor 2
#define centroid1_gfactor 3
#define centroid1_sc_reljump 4
#define centroid1_linfloor 5
#define centroid1_onedb_time 6
#define centroid1_db_per_kcycle 7
#define centroid1_capture_range 8
#define centroid1_tracking_range 9
#define centroid1_amp 10
#define centroid1_pi 11
#define centroid1_sr3 12
#define centroid1_acycle 13
#define centroid1_in_rms 14
#define centroid1_lp_rms 15
#define centroid1_out_rms 16
#define centroid1_lo_db 17
#define centroid1_hi_db 18
#define centroid1_in_db 19
#define centroid1_out_db 20
#define centroid1_state 21
#define centroid1_tilt_db 22
#define centroid1_last_tilt_db 23
#define centroid1_lgain_start 24
#define centroid1_hgain_start 25
#define centroid1_lmult 26
#define centroid1_hmult 27
#define centroid1_sc_update 28
#define centroid1_sc_interp 29
#define centroid1_omega_start 30
#define centroid1_omega_mult 31
#define centroid1_last_Spectral_Center 32
#define centroid1_in_log 33
#define centroid1_lp_log 58
#define centroid1_out_log 83
#define centroid1_bptr 108
#define centroid1_bidx 109
#define centroid1_mlog 110
#define centroid1_mmass 230
#define centroid1_nextm 231
#define centroid1_mptr 232
#define centroid1_midx 233
#define centroid1_qcount 234
#define centroid1_quick_set 235
#define centroid1_quick_go 236
#define centroid1_last_mmass 237
#define centroid1_mup_start 238
#define centroid1_mup_mult 239
#define centroid1_lgain 240
#define centroid1_hgain 241
#define centroid1_acount 242
#define centroid1_mup 243
#define centroid1_omega 244
#define centroid1_n 245
#define centroid1_a0 246
#define centroid1_b1 247
#define centroid1_lp 248
#define centroid1_out 250
#define centroid1_in_mono 252
#define centroid1_in_perceptual 253
#define centroid1_lp_mono 254
#define centroid1_lp_perceptual 255
#define centroid1_out_mono 256
#define centroid1_out_perceptual 257
#define centroid1__tvr10 258
#define centroid1__tvr9 259
#define centroid1__tvr8 260
#define centroid1__tvr7 261
#define centroid1__tvr6 262
#define centroid1__tvr5 263
#define centroid1__tvr4 264
#define centroid1__tvr3 265
#define centroid1__tvr2 266
#define centroid1__tvr1 267
#define centroid1__tvr0 268
#define centroid1_log1_return 269
#define centroid1_ampdb2_return 270
#define centroid1_pow3_return 271
#define centroid1_rms4_buffer 272
#define centroid1_rms4_scale 273
#define centroid1_rms4_kcyc 274
#define centroid1_rms4_x 275
#define centroid1_rms4_return 276
#define centroid1_rms5_buffer 277
#define centroid1_rms5_scale 278
#define centroid1_rms5_kcyc 279
#define centroid1_rms5_x 280
#define centroid1_rms5_return 281
#define centroid1_rms6_buffer 282
#define centroid1_rms6_scale 283
#define centroid1_rms6_kcyc 284
#define centroid1_rms6_x 285
#define centroid1_rms6_return 286
#define centroid1_sqrt7_return 287
#define centroid1_sqrt8_return 288
#define centroid1_sqrt9_return 289
#define centroid1_dbamp10_return 290
#define centroid1_dbamp11_return 291
#define centroid1_dbamp12_return 292
#define centroid1_dbamp13_return 293
#define centroid1_sgn14_x 294
#define centroid1_sgn14_return 295
#define centroid1_int15_return 296
#define centroid1_int16_return 297
#define centroid1_pow17_return 298
#define centroid1_pow18_return 299
#define centroid1_abs19_return 300
#define centroid1_abs20_return 301
#define centroid1_sgn21_x 302
#define centroid1_sgn21_return 303
#define centroid1_abs22_return 304
#define centroid1_sgn23_x 305
#define centroid1_sgn23_return 306
#define centroid1_exp24_return 307
#define centroid1_exp25_return 308
#define centroid1_exp26_return 309
#define centroid1_exp27_return 310
#define centroid1_exp28_return 311
#define centroid1_exp29_return 312
#define centroid1_exp30_return 313
#define centroid1_exp31_return 314
#define centroid1_exp32_return 315
#define centroid1_exp33_return 316
#define centroid1_exp34_return 317
#define centroid1_exp35_return 318
#define centroid1_exp36_return 319
#define centroid1_exp37_return 320
#define centroid1_exp38_return 321
#define centroid1_exp39_return 322
#define centroid1_exp40_return 323
#define centroid1_exp41_return 324
#define centroid1_exp42_return 325
#define centroid1_exp43_return 326
#define centroid1_lopass44_d2 327
#define centroid1_lopass44_d1 328
#define centroid1_lopass44_a2 329
#define centroid1_lopass44_a1 330
#define centroid1_lopass44_b2 331
#define centroid1_lopass44_b1 332
#define centroid1_lopass44_b0 333
#define centroid1_lopass44_ocut 334
#define centroid1_lopass44_kcyc 335
#define centroid1_lopass44_cut 336
#define centroid1_lopass44_input 337
#define centroid1_lopass44_return 338
#define centroid1_hipass45_d2 339
#define centroid1_hipass45_d1 340
#define centroid1_hipass45_a2 341
#define centroid1_hipass45_a1 342
#define centroid1_hipass45_b2 343
#define centroid1_hipass45_b1 344
#define centroid1_hipass45_b0 345
#define centroid1_hipass45_ocut 346
#define centroid1_hipass45_kcyc 347
#define centroid1_hipass45_cut 348
#define centroid1_hipass45_input 349
#define centroid1_hipass45_return 350
#define centroid1_lopass46_d2 351
#define centroid1_lopass46_d1 352
#define centroid1_lopass46_a2 353
#define centroid1_lopass46_a1 354
#define centroid1_lopass46_b2 355
#define centroid1_lopass46_b1 356
#define centroid1_lopass46_b0 357
#define centroid1_lopass46_ocut 358
#define centroid1_lopass46_kcyc 359
#define centroid1_lopass46_cut 360
#define centroid1_lopass46_input 361
#define centroid1_lopass46_return 362
#define centroid1_hipass47_d2 363
#define centroid1_hipass47_d1 364
#define centroid1_hipass47_a2 365
#define centroid1_hipass47_a1 366
#define centroid1_hipass47_b2 367
#define centroid1_hipass47_b1 368
#define centroid1_hipass47_b0 369
#define centroid1_hipass47_ocut 370
#define centroid1_hipass47_kcyc 371
#define centroid1_hipass47_cut 372
#define centroid1_hipass47_input 373
#define centroid1_hipass47_return 374
#define centroid1_lopass48_d2 375
#define centroid1_lopass48_d1 376
#define centroid1_lopass48_a2 377
#define centroid1_lopass48_a1 378
#define centroid1_lopass48_b2 379
#define centroid1_lopass48_b1 380
#define centroid1_lopass48_b0 381
#define centroid1_lopass48_ocut 382
#define centroid1_lopass48_kcyc 383
#define centroid1_lopass48_cut 384
#define centroid1_lopass48_input 385
#define centroid1_lopass48_return 386
#define centroid1_hipass49_d2 387
#define centroid1_hipass49_d1 388
#define centroid1_hipass49_a2 389
#define centroid1_hipass49_a1 390
#define centroid1_hipass49_b2 391
#define centroid1_hipass49_b1 392
#define centroid1_hipass49_b0 393
#define centroid1_hipass49_ocut 394
#define centroid1_hipass49_kcyc 395
#define centroid1_hipass49_cut 396
#define centroid1_hipass49_input 397
#define centroid1_hipass49_return 398
#define centroid1_ENDVAR 399

#define TBL_centroid1_rms4_buffer 0
#define TBL_centroid1_rms5_buffer 1
#define TBL_centroid1_rms6_buffer 2
#define centroid1_ENDTBL 3

#define CSYS_GIVENENDTIME 0

#define MAXENDTIME 1E+37

float endtime;
instr_line e_centroid[1];

#define CSYS_SASL_INSTR_centroid 0

instr_line cs_centroid[MAXCNOTES];
instr_line * cs_centroid___first;
instr_line * cs_centroid___last;
instr_line * cs_centroid___end;
instr_line * cs_centroid___next;


#define CSYS_MAXSASLINSTR 1
#define CSYS_SASL_NOINSTR 1
#define CSYS_SASL_MAXCONTROL 64
#define CSYS_SASL_MAXTABLES 64

instr_line **cs_first[CSYS_MAXSASLINSTR];
instr_line **cs_last[CSYS_MAXSASLINSTR];
instr_line **cs_end[CSYS_MAXSASLINSTR];
instr_line **cs_next[CSYS_MAXSASLINSTR];

struct {
unsigned short id;
unsigned short label;
unsigned int fptr;
float fval;
} saslcontrol[CSYS_SASL_MAXCONTROL];

int maxsc;

struct table_cdtype {
unsigned short index; /* gtables index */
unsigned short tgen;  /* table generator */
unsigned int pnum;  /* number of params */
float * p;            /* param list */
} sasltable[CSYS_SASL_MAXTABLES];

int maxtb;

#define MAXLINES 256
#define MAXSTATE 264


#define MAXVARSTATE 399
#define MAXTABLESTATE 3

/* ninstr: used for score, effects, */
/* and dynamic instruments          */

struct ninstr_types {

instr_line * iline; /* pointer to score line */
dstack v[MAXVARSTATE];     /* parameters & variables*/
tableinfo t[MAXTABLESTATE]; /* tables */

} ninstr[MAXSTATE];


} engine;

extern void sigint_handler(int);

extern ENGINE_PTR_TYPE system_init(int argc, char **argv, float sample_rate);
extern void effects_init(ENGINE_PTR_DECLARE);
extern void main_initpass(ENGINE_PTR_DECLARE);
extern void shut_down(ENGINE_PTR_DECLARE);

extern void main_apass(ENGINE_PTR_DECLARE);
extern int  main_kpass(ENGINE_PTR_DECLARE);
extern void main_ipass(ENGINE_PTR_DECLARE);
extern void main_control(ENGINE_PTR_DECLARE);

float centroid__sym_rms4(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
void centroid__sym_rms4_spec(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_rms5(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
void centroid__sym_rms5_spec(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_rms6(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
void centroid__sym_rms6_spec(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_sgn14(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_sgn21(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_sgn23(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_lopass44(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_hipass45(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_lopass46(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_hipass47(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_lopass48(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid__sym_hipass49(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
float centroid1__sym_rms4(ENGINE_PTR_TYPE);
void centroid1__sym_rms4_spec(ENGINE_PTR_TYPE);
float centroid1__sym_rms5(ENGINE_PTR_TYPE);
void centroid1__sym_rms5_spec(ENGINE_PTR_TYPE);
float centroid1__sym_rms6(ENGINE_PTR_TYPE);
void centroid1__sym_rms6_spec(ENGINE_PTR_TYPE);
float centroid1__sym_sgn14(ENGINE_PTR_TYPE);
float centroid1__sym_sgn21(ENGINE_PTR_TYPE);
float centroid1__sym_sgn23(ENGINE_PTR_TYPE);
float centroid1__sym_lopass44(ENGINE_PTR_TYPE);
float centroid1__sym_hipass45(ENGINE_PTR_TYPE);
float centroid1__sym_lopass46(ENGINE_PTR_TYPE);
float centroid1__sym_hipass47(ENGINE_PTR_TYPE);
float centroid1__sym_lopass48(ENGINE_PTR_TYPE);
float centroid1__sym_hipass49(ENGINE_PTR_TYPE);
void centroid_ipass(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
void centroid_kpass(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);
void centroid_apass(ENGINE_PTR_TYPE_COMMA struct ninstr_types *);

void centroid1_ipass(ENGINE_PTR_TYPE);
void centroid1_kpass(ENGINE_PTR_TYPE);
void centroid1_apass(ENGINE_PTR_TYPE);

float finput0(float);
float finGroup0(float);

extern float table_global_aup_Spectral_Parameters_pinfo[];
extern float table_global_aup_Spectral_Center_pinfo[];
extern float table_global_aup_Treble_Meter_pinfo[];
extern float table_global_aup_Servo_Loop_Parameters_pinfo[];
extern float table_global_aup_Tracking_Time_pinfo[];
extern float table_global_aup_Threshold_pinfo[];

csys_table_ptr csys_table_catalog[] = {
	4, table_global_aup_Spectral_Parameters_pinfo,
	4, table_global_aup_Spectral_Center_pinfo,
	4, table_global_aup_Treble_Meter_pinfo,
	4, table_global_aup_Servo_Loop_Parameters_pinfo,
	4, table_global_aup_Tracking_Time_pinfo,
	4, table_global_aup_Threshold_pinfo };

#define CSYS_TABLE_CATALOG_SIZE 6

int csys_instrtablesize[CSYS_MAXSASLINSTR] = {
centroid_ENDTBL};

int csys_instrvarsize[CSYS_MAXSASLINSTR] = {
centroid_ENDVAR};


void engine_init(ENGINE_PTR_DECLARE_COMMA float sample_rate)
{
  int int_krate = (int)(SAOL_KRATE);

  int int_srate = (int)(sample_rate);

  EV(globaltune) = 440.0F;
  EV(invglobaltune) = 2.272727e-03F;
  EV(kbase) = 1;
  EV(kcycleidx) = 1;
  EV(pass) = IPASS;
  EV(tempo) = 60.0F;
  EV(ARATE) = sample_rate;
  EV(ATIME) = 1.0F/sample_rate;

  if (int_srate % int_krate)
    while (int_srate % (++int_krate));

  EV(ACYCLE) = int_srate/int_krate;
  EV(KRATE) = (float) int_krate;
  EV(KTIME) = 1.0F/EV(KRATE);
  EV(KMTIME) = 1000.0F*EV(KTIME);
  EV(KUTIME) = (int)(1000000.0F*EV(KTIME) + 0.5F);

  EV(scoremult) = EV(KTIME);
  EV(asys_exit_status) = ASYS_DONE;

  EV(endtime) = 60.0F;

  EV(cs_first)[0] = &EV(cs_centroid___first);

  EV(cs_last)[0] = &EV(cs_centroid___last);

  EV(cs_end)[0] = &EV(cs_centroid___end);

  EV(cs_next)[0] = &EV(cs_centroid___next);

  EV(cs_centroid___first) = &EV(cs_centroid)[1];
  EV(cs_centroid___last) = &EV(cs_centroid)[0];
  EV(cs_centroid___end) = &EV(cs_centroid)[MAXCNOTES-1];
}


/*  Reflection interface starts here */

#define CSYS_GLOBALNUM 24

csys_varstruct csys_global[CSYS_GLOBALNUM] = {
0, "aup_Spectral_Parameters", 2, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
1, "aup_Spectral_Parameters_display_menu", 3, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
2, "aup_Spectral_Parameters_idx0__", 4, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
0, "aup_Spectral_Parameters_pinfo", 5, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED, 
4, "aup_Spectral_Center", 7, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_READ, 
5, "aup_Spectral_Center_slider_log", 8, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
6, "aup_Spectral_Center_unit_Hz", 9, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
1, "aup_Spectral_Center_pinfo", 10, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED, 
8, "aup_Treble_Meter", 11, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN, 
9, "aup_Treble_Meter_slider_linear", 12, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
10, "aup_Treble_Meter_unit_dB", 13, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
2, "aup_Treble_Meter_pinfo", 14, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED, 
12, "aup_Servo_Loop_Parameters", 15, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
13, "aup_Servo_Loop_Parameters_display_menu", 16, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
14, "aup_Servo_Loop_Parameters_idx0__", 17, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
3, "aup_Servo_Loop_Parameters_pinfo", 18, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED, 
16, "aup_Tracking_Time", 19, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_READ, 
17, "aup_Tracking_Time_slider_log", 20, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
18, "aup_Tracking_Time_unit_s", 21, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
4, "aup_Tracking_Time_pinfo", 22, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED, 
20, "aup_Threshold", 23, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_READ, 
21, "aup_Threshold_slider_log", 24, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
22, "aup_Threshold_unit_dB", 25, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_UNUSED, 
5, "aup_Threshold_pinfo", 26, CSYS_TABLE, CSYS_NORMAL, 1, CSYS_UNUSED };

#define ASYS_AUDIOUNIT_PARAMETERS_WRITTEN 1
#define ASYS_AUDIOUNIT_PARAMETERS_READ 1

#define CSYS_TARGETNUM 0

csys_targetstruct csys_target[1];

#define CSYS_NOLABEL 0

#define CSYS_LABELNUM 0

csys_labelstruct csys_labels[1];

#define CSYS_PRESETNUM 0

csys_presetstruct csys_presets[1];

#define CSYS_SAMPLENUM 0

csys_samplestruct csys_samples[1];

#define CSYS_BUSNUM 2

csys_busstruct csys_bus[CSYS_BUSNUM] = {
0, "output_bus",2, 0 ,
1, "input_bus",2, 0 };

#define CSYS_ROUTENUM 0

csys_routestruct csys_route[1];

#define CSYS_SEND_BNUM_0 1

int csys_send0_bus[CSYS_SEND_BNUM_0] = {
 1};

#define CSYS_SENDNUM 1

csys_sendstruct csys_send[CSYS_SENDNUM] = {
0, CSYS_SEND_BNUM_0, &(csys_send0_bus[0]) };

#define CSYS_centroid_VARNUM 69

csys_varstruct csys_centroid_vars[CSYS_centroid_VARNUM] = {
1, "loudness_bass", 27, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "loudness_treble", 28, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "logfloor", 29, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "gfactor", 30, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "sc_reljump", 31, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
-1, "aup_Spectral_Center", 7, CSYS_KRATE, CSYS_IMPORTEXPORT, 1, CSYS_READ ,
-1, "aup_Treble_Meter", 11, CSYS_KRATE, CSYS_IMPORTEXPORT, 1, CSYS_READ ,
-1, "aup_Tracking_Time", 19, CSYS_KRATE, CSYS_IMPORTEXPORT, 1, CSYS_READ ,
-1, "aup_Threshold", 23, CSYS_KRATE, CSYS_IMPORTEXPORT, 1, CSYS_READ ,
1, "linfloor", 32, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "onedb_time", 33, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "db_per_kcycle", 34, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "capture_range", 35, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "tracking_range", 36, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "amp", 37, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
1, "pi", 38, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "sr3", 39, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "acycle", 40, CSYS_IRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "in_rms", 41, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lp_rms", 42, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "out_rms", 43, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lo_db", 44, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "hi_db", 45, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "in_db", 46, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "out_db", 47, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "state", 48, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "tilt_db", 49, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "last_tilt_db", 50, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lgain_start", 51, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "hgain_start", 52, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lmult", 53, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "hmult", 54, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "sc_update", 55, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "sc_interp", 56, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "omega_start", 57, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "omega_mult", 58, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "last_Spectral_Center", 59, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "in_log", 60, CSYS_KRATE, CSYS_NORMAL, 25, CSYS_WRITTEN_AND_READ ,
0, "lp_log", 61, CSYS_KRATE, CSYS_NORMAL, 25, CSYS_WRITTEN_AND_READ ,
0, "out_log", 62, CSYS_KRATE, CSYS_NORMAL, 25, CSYS_WRITTEN_AND_READ ,
0, "bptr", 63, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "bidx", 64, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "mlog", 65, CSYS_KRATE, CSYS_NORMAL, 120, CSYS_WRITTEN_AND_READ ,
0, "mmass", 66, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "nextm", 67, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "mptr", 68, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "midx", 69, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "qcount", 70, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "quick_set", 71, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "quick_go", 72, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "last_mmass", 73, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "mup_start", 74, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "mup_mult", 75, CSYS_KRATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lgain", 76, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "hgain", 77, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "acount", 78, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "mup", 79, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "omega", 80, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "n", 81, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "a0", 82, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "b1", 83, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lp", 84, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "out", 85, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "in_mono", 86, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "in_perceptual", 87, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lp_mono", 88, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "lp_perceptual", 89, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "out_mono", 90, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ ,
0, "out_perceptual", 91, CSYS_ARATE, CSYS_NORMAL, 1, CSYS_WRITTEN_AND_READ  };

csys_instrstruct csys_instr[CSYS_INSTRNUM] = {
0, "centroid",0, CSYS_centroid_VARNUM, &(csys_centroid_vars[0]),2, 1 };


/*  Reflection interface ends here */


/* handles termination in case of error */

extern void epr(int linenum, char * filename, char * token, char * message);

void csys_terminate(char * message)
{
   epr(0,NULL,"aucontrol control driver", message);
}


/*
#    Sfront, a SAOL to C translator    
#    This file: Code for table generators
#
# Copyright (c) 1999-2006, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  Redistributions of source code must retain the above copyright
#  notice, this list of conditions and the following disclaimer.
#
#  Redistributions in binary form must reproduce the above copyright
#  notice, this list of conditions and the following disclaimer in the
#  documentation and/or other materials provided with the distribution.
#
#  Neither the name of the University of California, Berkeley nor the
#  names of its contributors may be used to endorse or promote products
#  derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#    Maintainer: John Lazzaro, lazzaro@cs.berkeley.edu
*/


int tgen_init(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)

{
  int i;

  if (ti->llmem)
    free(ti->t); 
  ti->start = 0;
  ti->end = 0;
  ti->sr = 0;
  ti->dint = 0;
  ti->dfrac = 0;
  ti->base = 0;
  ti->stamp = EV(scorebeats);
  if ((pnum < 1) || ((i = ti->len = ROUND(p[0])) < 1))
    epr(0,"control driver","tgen_init","Table length < 1");
  ti->lenf = (float) i;
  ti->oconst = ((float) i)*EV(ATIME);
  ti->tend = i - 1;
  ti->t = (float *) calloc(i+1, sizeof(float));
  ti->llmem = 1;

#if INTERP_TYPE == INTERP_SINC  
  ti->sffl = 1.0F;
  ti->sfui = 0x00010000;
  ti->dsincr = SINC_PILEN;
#endif

  return i;
}


void tgen_buzz(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int size, nharm, loharm, i, j;
  float scale, base, index, acc, f;

  if (pnum < 4)
    epr(0,"control driver","tgen_buzzwave","Too few parameters");

  size = ROUND(p[0]);
  nharm = ROUND(p[1]);
  loharm = ROUND(p[2]);

  if (loharm < 0)
    epr(0,"control driver","tgen_buzzwave","Loharm must be >= 0");

  if ((size < 1) && (nharm < 1))
    epr(0,"control driver","tgen_buzzwave","Size or nharm must be given");
  if (nharm < 1)
    nharm = (int) floor(size/2.0 - loharm);
  else
    if (size < 1)
      size = 2*(loharm + nharm);

  if (((f = fabs(p[3])) < 1.0F) && (f != 0.0F))
    {
      j = -(int)(17.0F/log(f)) + loharm + 1;
      nharm = (nharm > j) ? j : nharm;
    }

  p[0] = (float) size;
  p[1] = (float) nharm;
  p[2] = (float) loharm;

  tgen_init(ENGINE_PTR_COMMA ti, pnum, p);

  if ((p[3] == 1.0F)||(p[3] == -1.0F))
    scale = 1.0F;
  else
    {
      if (p[3] == 0.0F)
	scale = 1.0F;
      else
	scale = (1.0F-(float)fabs(p[3]))/
	  (1-(float)fabs((float)pow(p[3],p[1]+1.0F)));
    }

   base = 6.283185F/p[0];
   for (i = 0; i < size; i++)
     {
       index = i*base;
       acc = 1.0F;
       j = loharm + 1;
       while (j <= (loharm + nharm))
	 {
	   ti->t[i] += acc*(float)cos(index*j);
	   acc *= p[3];
	   j++;
	 }
       ti->t[i] *= scale;
     }
   ti->t[ti->len] = ti->t[0];

}


void tgen_concat_global(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int i, j, k, len, tidx;

  if (pnum < 2)
    epr(0,"control driver","tgen_concat","Too few parameters");

  len = 0;
  for (j = 1; j < pnum; j++)
    {
      i = ROUND(p[j]);
      if ( (!EV(gtables)[i].t) || (EV(gtables)[i].len < 1))
	epr(0,"control driver","tgen_concat","Uninitialized/zero-lenth table");
      len += EV(gtables)[i].len;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  i = 0;
  tidx = ROUND(p[k = 1]);
  j = 0;
  while ((i < len) && (k < pnum))
    {
      if (j < EV(gtables)[tidx].len)
	ti->t[i++] = EV(gtables)[tidx].t[j++];
      else
	{
	  j = 0;
	  if (++k < pnum)
	    tidx = ROUND(p[k]);
	}
    }
  ti->t[ti->len] = ti->t[0];

}

void tgen_cubicseg(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p) 

{
  int i, j, len, endpoint;
  float Q, R, S, T, a, b, c, d, xf;

  if (pnum < 7)
    epr(0,"control driver","tgen_cubicseg","Too few parameters");

  if ((pnum - 3) % 4)
    epr(0,"control driver","tgen_cubicseg",
	"Parmeter list must end with (infl, y)");
    
  if ((len = ROUND(p[1])) != 0)
    epr(0,"control driver","tgen_cubicseg","x0 != 0");
  i = 3;

  while (i < pnum)
    {
      j = ROUND(p[i]);
      if (j < len)
	epr(0,"control driver","tgen_cubicseg",
	    "Consecutive x and infl values decreasing."); 
      p[i] = (float)(len = j);
      i += 2;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  i = 0;
  xf = 0.0F;

  for (j = 1; j < pnum; j += 4)
    {
      endpoint = p[j+4];

      Q = p[j]*p[j]*p[j] - p[j+2]*p[j+2]*p[j+2] 
	- 3.0F*p[j+2]*p[j+2]*(p[j]-p[j+2]);
      if (Q == 0.0F)
	epr(0,"control driver","tgen_cubicseg",
	    "No cubic solution for these parameters");
      R = p[j]*p[j] - p[j+2]*p[j+2] - 2.0F*p[j+2]*(p[j]-p[j+2]);
      if (R == 0.0F)
	epr(0,"control driver","tgen_cubicseg",
	    "No cubic solution for these parameters");
      
      S = p[j+2]*p[j+2]*p[j+2] - p[j+4]*p[j+4]*p[j+4]
	- 3.0F*p[j+2]*p[j+2]*(p[j+2]-p[j+4]);

      if (S == 0.0F)
	epr(0,"control driver","tgen_cubicseg",
	    "No cubic solution for these parameters");

      T = p[j+2]*p[j+2] - p[j+4]*p[j+4] - 2.0F*p[j+2]*(p[j+2]-p[j+4]);

      if ((T == 0.0F) || ((Q/R) == (S/T)))
	epr(0,"control driver","tgen_cubicseg",
	    "No cubic solution for these parameters");

      a = 1.0F/((Q/R) - (S/T));
      a *= (p[j+1]-p[j+3])/R - (p[j+3]-p[j+5])/T;
      b = 1.0F/((R/Q) - (T/S));
      b *= (p[j+1]-p[j+3])/Q - (p[j+3]-p[j+5])/S;
      c = - 3.0F*a*p[j+2]*p[j+2] - 2.0F*b*p[j+2];
      d = p[j+3] - a*p[j+2]*p[j+2]*p[j+2];
      d += - b*p[j+2]*p[j+2] - c*p[j+2];
    
      while (i < endpoint)
	{
	  ti->t[i] =  xf*xf*xf*a + xf*xf*b + xf*c + d;
	  xf = (++i);
	}
    }
  ti->t[ti->len] = ti->t[0];

}

void tgen_data(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{  
  int len, i;
  
  if (pnum < 1)
    epr(0,"control driver","tgen_data","Too few parameters");

  if (p[0] <= 0.0F)
    p[0] = pnum - 1;

  len = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
  len = (pnum - 1 < len) ? pnum - 1 : len;

  i = 0;
  while (i < len)
    {
      ti->t[i] = p[i+1];
      i++;
    }
  ti->t[ti->len] = ti->t[0];

}

void tgen_destroy(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{  
  ti->len = ti->start = ti->end = ti->tend = 0;
  ti->lenf = ti->oconst = ti->sr = ti->base = 0.0F;
  ti->dint = ti->dfrac = 0;
  ti->sffl = 0;
  ti->sfui = 0;
  ti->dsincr = 0;
  if (ti->llmem)
    free(ti->t); 
  ti->llmem = 0;
  ti->stamp = EV(scorebeats);
}

void tgen_empty(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
}


void tgen_expseg(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p) 

{
  int i, j, len, dx, nx, sgn;
  float diff;

  if (pnum < 5)
    epr(0,"control driver","tgen_expseg","Too few parameters");

  if (!(pnum % 2))
    epr(0,"control driver","tgen_expseg","Last parameter is an xval");
  
  if ((len = ROUND(p[1])) != 0)
    epr(0,"control driver","tgen_expseg","x0 != 0");

  if (p[2] == 0.0F)
    epr(0,"control driver","tgen_expseg","illegal yval");
  sgn = (p[2] > 0);

  i = 3;

  while (i < pnum)
    {
      if ((p[i+1] == 0.0F) || ((p[i-1] > 0) != sgn))
	epr(0,"control driver","tgen_expseg","illegal yval");

      j = ROUND(p[i]);
      if (j < len)
	epr(0,"control driver","tgen_expseg","Decreasing xval in sequence");
      len = j;
      i += 2;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  i = j = 1;
  while ((j + 2) < pnum)
    if (!(dx = ((nx = ROUND(p[j+2])) - ROUND(p[j]))))
      j += 2;
    else
      {
	diff = (float)pow(p[j+3]/p[j+1], 1.0F/dx);
	ti->t[ti->len] = ti->t[0] = p[j+1];
	break;
      }

  while ((i < len) && ((j + 2) < pnum))
    {
      if (nx > i)
	{
	  ti->t[i] = ti->t[i-1]*diff;
	  i++;
	}
      else
	{
	  j += 2;
	  while ((j + 2) < pnum)
	    if (!(dx = ((nx =ROUND(p[j+2])) - ROUND(p[j]))))
	      j += 2;
	    else
	      {
		diff = (float)pow(p[j+3]/p[j+1], 1.0F/dx);
		ti->t[i++] = p[j+1];
		break;
	      }
	}
    }

}


void tgen_harm(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)

{
  int i, j;
  float base, index;

  i = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
  base = 6.283185F/i;
  i--;

  while (i >= 0)
    {
      j = 0;
      index = i*base;
      while (++j < pnum)
	ti->t[i] += p[j]*(float)sin(j*index);
      i--;
    }
  ti->t[ti->len] = ti->t[0];

}

void tgen_harm_phase(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)

{
  int i, j;
  float base, index, k;

  i = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
  base = 6.283185F/i;
  i--;

  if (!(pnum % 2))
    epr(0,"control driver","tgen_harm_phase","Must have (f1, ph1) pairs.");

  while (i >= 0)
    {
      j = -1;
      k = 1.0F;
      index = i*base;
      while ((j += 2) < pnum)
	ti->t[i] += p[j]*(float)sin((k++)*index + p[j+1]);
      i--;
    }
  ti->t[ti->len] = ti->t[0];

}


void tgen_lineseg(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p) 

{
  int i, j, len, dx, nx;
  float diff;

  if (pnum < 5)
    epr(0,"control driver","tgen_lineseg","Too few parameters");

  if (!(pnum % 2))
    epr(0,"control driver","tgen_lineseg","Last parameter is an xval");
  
  if ((len = ROUND(p[1])) != 0)
    epr(0,"control driver","tgen_lineseg","x0 != 0");
  i = 3;

  while (i < pnum)
    {
      j = ROUND(p[i]);
      if (j < len)
	epr(0,"control driver","tgen_lineseg","Decreasing xval in sequence");
      len = j;
      i += 2;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  i = j = 1;
  while ((j + 2) < pnum)
    if (!(dx = ((nx = ROUND(p[j+2])) - ROUND(p[j]))))
      j += 2;
    else
      {
	diff = (p[j+3]-p[j+1])/dx;
	ti->t[ti->len] = ti->t[0] = p[j+1];
	break;
      }

  while ((i < len) && ((j + 2) < pnum))
    {
      if (nx > i)
	{
	  ti->t[i] = ti->t[i-1] + diff;
	  i++;
	}
      else
	{
	  j += 2;
	  while ((j + 2) < pnum)
	    if (!(dx = ((nx =ROUND(p[j+2])) - ROUND(p[j]))))
	      j += 2;
	    else
	      {
		diff = (p[j+3]-p[j+1])/dx;
		ti->t[i++] = p[j+1];
		break;
	      }
	}
    }

}

void tgen_periodic(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int i, j;
  float base, index;

  i = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
  base = 6.283185F/i;
  i--;

  if ((pnum-1) % 3)
    epr(0,"control driver","tgen_periodic","Must have (p1, f1, ph1) triplets.");

  while (i >= 0)
    {
      j = -2;
      index = i*base;
      while ((j += 3) < pnum)
	ti->t[i] += p[j+1]*(float)sin(p[j]*index + p[j+2]);
      i--;
    }
  ti->t[ti->len] = ti->t[0];

}


void tgen_polynomial(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p) 

{
  float scale, yval, ynth;
  int i, j;

  if (pnum < 4)
    epr(0,"control driver","tgen_polynomial","Too few parameters");

  tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    
  scale = (1.0F/p[0])*(p[2]-p[1]);

  for (i=0; i < ti->len; i++)
    {
      yval = p[1] + scale*(p[0] - i);
      ynth = 1.0F;
      ti->t[i] = p[3];
      j = 3;
      while (++j < pnum)
	{
	  ynth *= yval;
	  ti->t[i] += p[j]*ynth;
	}
    }

  ti->t[ti->len] = ti->t[0];

}

#define RAN_UNIFORM 1
#define RAN_LINRAMP 2
#define RAN_EXPO    3
#define RAN_GAUSS   4
#define RAN_PPROC   5

void tgen_random(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int dist, i, j;
  float c1, x, y;

  if (pnum < 3)
    epr(0,"control driver","tgen_random","Too few parameters");

  tgen_init(ENGINE_PTR_COMMA ti, pnum, p);

  dist = ROUND(p[1]);
    
  switch(dist) {
  case RAN_UNIFORM:
  case RAN_LINRAMP:
    if (pnum < 4)
      epr(0,"control driver","tgen_random","Too few parameters");
    c1 = p[3] - p[2];
    if ((dist == 2) && !(c1))
      epr(0,"control driver","tgen_random","p1 == p2 (dist 2)");
    break;
  case RAN_GAUSS:
    if (pnum < 4)
      epr(0,"control driver","tgen_random","Too few parameters");
    if (p[3] <= 0.0F)
      epr(0,"control driver","tgen_random","p2 <= 0 (dist 4)");
    c1 = (float)sqrt(2.0*p[3]);
    break;
  case RAN_EXPO:
  case RAN_PPROC:
    break;
  default:
    epr(0,"control driver","tgen_random","Distribution not 1,2,3,4 or 5.");
  }

  j = 0; i = 0;
  while (i < ti->len)
    {
      switch(dist) {
      case RAN_UNIFORM:
        ti->t[i] = c1*RMULT*((float)rand()) + p[2];
        break;
      case RAN_LINRAMP:
        x = RMULT*((float)rand());
        y = RMULT*((float)rand());
        if (x > y)
          ti->t[i] =  c1*x + p[2];
        else
          ti->t[i] =  c1*y + p[2];
        break;
      case RAN_EXPO:
        ti->t[i] = -p[2]*(float)log(RMULT*((float)rand())+1e-45F);
        break;
      case RAN_GAUSS:
        ti->t[i] = p[2]+c1*
	  (float)sqrt(-2.0F*(float)log(RMULT*((float)rand())+1e-45F))
	  *(float)cos(6.283185F*RMULT*((float)rand()));
        break;
      case RAN_PPROC:
        ti->t[i] = 0;
        if (j == 0)
	  {
	    j = ROUND(-p[2]*(float)log(RMULT*((float)rand())+1e-45F))+1;
	    if (i != 0)
	      ti->t[i] = 1.0F;
	  }
	j--;
        break;
      }
      i++;
    }
  ti->t[ti->len] = ti->t[0];

}

#undef RAN_UNIFORM 
#undef RAN_LINRAMP 
#undef RAN_EXPO    
#undef RAN_GAUSS   
#undef RAN_PPROC   


#define SAMP_SR        5
#define SAMP_LOOPSTART 4
#define SAMP_LOOPEND   3
#define SAMP_BASEFREQ  2
#define SAMP_LLMEM     1
#define SAMP_DATABLOCK 5

void tgen_sample(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int i;
  double intdummy;

  if (pnum < SAMP_DATABLOCK + 1)
    epr(0,"control driver","tgen_sample","Zero-length sample");

  ti->len = pnum - SAMP_DATABLOCK;
  ti->lenf = (float)(pnum - SAMP_DATABLOCK);
  ti->oconst = ti->lenf*EV(ATIME);

  if (ti->llmem)
    free(ti->t); 
  ti->t = p;

  ti->stamp = EV(scorebeats);

  if (p[pnum - SAMP_SR] > 0)
    {
      ti->sr = p[pnum - SAMP_SR];
      if (p[pnum - SAMP_SR] == EV(ARATE))
	{
	  ti->dint = 1;
	  ti->dfrac = 0;
	}
      else
	{
	  ti->dfrac = 4294967296.0*
	    modf(((double)p[pnum - SAMP_SR])/((double)EV(ARATE)), &intdummy);
	  ti->dint = intdummy;
	}
    }
  else
    {
      ti->sr = EV(ARATE);
      ti->dint = 1;
      ti->dfrac = 0;
    }

#if (INTERP_TYPE == INTERP_SINC)

  if (p[pnum - SAMP_SR] > 0)
    {
      if (EV(ARATE) >= p[pnum - SAMP_SR])
	{
	  ti->sffl = 1.0F;
	  ti->sfui = 0x00010000;
	  ti->dsincr = SINC_PILEN;
	}
      else
	{
	  if ((EV(ARATE)*SINC_UPMAX) >= p[pnum - SAMP_SR])
	    ti->sffl = (EV(ARATE)/ti->sr);
	  else
	    ti->sffl = (1.0F/SINC_UPMAX);
	  ti->sfui = ((float)(pow(2,16)))*ti->sffl + 0.5F;
	  ti->dsincr = (SINC_PILEN*ti->sfui) >> 16;
	}
    }
  else
    {
      ti->sffl = 1.0F;
      ti->sfui = 0x00010000;
      ti->dsincr = SINC_PILEN;
    }

#endif

  if (p[pnum - SAMP_BASEFREQ] > 0)
    {
      ti->base = p[pnum - SAMP_BASEFREQ];
    }
  else
    {  
      ti->base = 0;
    }

  if (p[pnum - SAMP_LOOPSTART] > 0)
    {
      ti->start = p[pnum - SAMP_LOOPSTART];
    }
  else
    {
      ti->start = 0;
    }

  if (p[pnum - SAMP_LOOPEND] > 0)
    {
      ti->tend = ti->end = p[pnum - SAMP_LOOPEND];
      if (ti->end==0)
	{
	  ti->tend = ti->len - 1;
	}
    }
  else
    {
      ti->end = 0;
      ti->tend = i - 1;
    }
  ti->llmem = (int)(p[pnum - SAMP_LLMEM]);
  ti->t[ti->len] = ti->t[0];
}

#undef SAMP_SR 
#undef SAMP_LOOPSTART 
#undef SAMP_LOOPEND 
#undef SAMP_BASEFREQ 
#undef SAMP_DATABLOCK 


void tgen_spline(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int i, j, len, endpoint, newpoint;
  float Q, R, S, a, b, c, d, k1, k2, xf;

  if (pnum < 6)
    epr(0,"control driver","tgen_spline","Too few parameters");

  if ((pnum - 3) % 3)
    epr(0,"control driver","tgen_spline",
	"Parmeter list must end with (x, y)");
    
  if ((len = ROUND(p[1])) != 0)
    epr(0,"control driver","tgen_cubicseg","x0 != 0");
  i = 4;

  while (i < pnum)
    {
      j = ROUND(p[i]);
      if (j < len)
	epr(0,"control driver","tgen_spline",
	    "Consecutive x values decreasing.");
      p[i] = (float)(len = j);
      i += 3;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  endpoint = i = 0;
  xf = 0.0F;

  for (j = 1; (j+4) < pnum; j += 3)
    {

      if ((newpoint = p[j+3]) != endpoint)
	{
	  endpoint = newpoint;

	  if (j == 1)
	    k1 = 0.0F;
	  else
	    k1 = p[j-1];
	  if ((j+5) == pnum)
	    k2 = 0;
	  else
	    k2 = p[j+2];

	  Q = 1.0F/(p[j] - p[j+3]);
	  R = Q*(p[j]*p[j] - p[j+3]*p[j+3]) - 2.0F*p[j];
	  if (R == 0.0F)
	    epr(0,"control driver","tgen_splice",
		"No spline solution for these parameters");
	  R =1.0F/R;
	  S = Q*(p[j]*p[j] - p[j+3]*p[j+3]) - 2.0F*p[j+3];
	  if (S == 0.0F)
	    epr(0,"control driver","tgen_splice",
		"No spline solution for these parameters");
	  S =1.0F/S;
	  
	  a = Q*(p[j]*p[j]*p[j]-p[j+3]*p[j+3]*p[j+3])*(R-S);
	  a += -3.0F*(R*p[j]*p[j] - S*p[j+3]*p[j+3]);
	  if (a==0.0F)
	    epr(0,"control driver","tgen_spline",
		"No spline solution for these parameters");
	  a =1.0F/a;
	  a *= Q*(p[j+1]-p[j+4])*(R-S) - R*k1 + S*k2;
	  b = Q*(k1-k2)*0.5F;
	  b += - 1.5F*a*Q*(p[j]*p[j] - p[j+3]*p[j+3]);
	  c = k1 - 3.0F*a*p[j]*p[j] -2.0F*b*p[j];
	  d=p[j+1]-a*p[j]*p[j]*p[j]-b*p[j]*p[j]-c*p[j];

	  while (i < endpoint)
	    {
	      ti->t[i] =  xf*xf*xf*a + xf*xf*b + xf*c + d;
	      xf = (++i);
	    }
	}
    }
  ti->t[ti->len] = ti->t[0];

}

void tgen_step(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int i, j, len, nx;

  if (pnum < 4)
    epr(0,"control driver","tgen_step","Too few parameters");

  if (pnum % 2)
    epr(0,"control driver","tgen_step","Last parameter is a yval");
  
  if ((len = ROUND(p[1])) != 0)
    epr(0,"control driver","tgen_step","x0 != 0");
  i = 3;

  while (i < pnum)
    {
      j = ROUND(p[i]);
      if (j < len)
	epr(0,"control driver","tgen_step","Decreasing xval in sequence");
      len = j;
      i += 2;
    }

  if (p[0] <= 0.0F)
    {
      p[0] = len;
      tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
    }
  else
    {
      j = tgen_init(ENGINE_PTR_COMMA ti, pnum, p);
      len = (j < len) ? j : len; 
    }

  i = 0; 
  j = 2;
  nx = ROUND(p[3]);
  while (i < len)
    {
      if (nx > i)
	{
	  ti->t[i] = p[j];
	  i++;
	}
      else
	{
	  if ((j += 2) == pnum)
	    break;
	  nx = ROUND(p[j+1]);
	}
    }
  ti->t[ti->len] = ti->t[0];
}


double tgen_bessel(double x)

{
  double sum,a;

  sum = a = 1.0;
  a *= x*x; sum += a*2.5e-1;
  a *= x*x; sum += a*1.5625e-2;
  a *= x*x; sum += a*4.340278e-4;
  a *= x*x; sum += a*6.781684e-6;
  a *= x*x; sum += a*6.781684e-8; 
  a *= x*x; sum += a*4.709503e-10;
  a *= x*x; sum += a*2.402808e-12; 
  a *= x*x; sum += a*9.385967e-15;
  a *= x*x; sum += a*2.896903e-17;
  a *= x*x; sum += a*7.242258e-20;

  return sum;
}


#define WIN_HAMMING  1
#define WIN_HANNING  2
#define WIN_BARTLETT 3
#define WIN_GAUSSIAN 4
#define WIN_KAISER   5
#define WIN_BOXCAR   6

void tgen_window(ENGINE_PTR_DECLARE_COMMA tableinfo * ti, int pnum, float * p)  

{
  int type,i;
  float c1, c2;
  double d1, d2, d3;

  if (pnum < 2)
    epr(0,"control driver","tgen_window","Too few parameters");

  tgen_init(ENGINE_PTR_COMMA ti, pnum, p);

  type = ROUND(p[1]);
    
  switch(type) {
  case WIN_HAMMING:
  case WIN_HANNING:
    c1 = 6.283185F/(p[0]-1.0F);
    break;
  case WIN_BARTLETT:
    c1 = 2.0F/(p[0]-1.0F);
    c2 = 1.0F/c1;
    break;
  case WIN_GAUSSIAN:
    c1 = - 18.0F/(p[0]*p[0]);
    c2 = - 0.5F*p[0];
    break;
  case WIN_KAISER:     
    if (pnum < 3)
      epr(0,"control driver","tgen_window","Parameter p needed (type 5)");
    d1 = (p[0]-1.0)/2.0;
    d2 = d1*d1;
    d3 = 1.0/tgen_bessel(p[2]*d1);
    break;
  case WIN_BOXCAR:
    break;
  default:
    epr(0,"control driver","tgen_window","Window type not 1,2,3,4,5 or 6");
  }

  for (i=0; i < ti->len; i++)
    switch(type) {
    case WIN_HAMMING:
      ti->t[i] = 0.54F - 0.46F*(float)cos(c1*i);
      break;
    case WIN_HANNING:
      ti->t[i] = 0.5F*(1.0F - (float)cos(c1*i));
      break;
    case WIN_BARTLETT:
      ti->t[i] = 1.0F - c1*(float)fabs(i - c2);
      break;
    case WIN_GAUSSIAN:
      ti->t[i] = 
	(float)exp(c1*(c2+i)*(c2+i));
      break;
    case WIN_KAISER:
      ti->t[i] = (float)(d3*tgen_bessel(p[2]*sqrt(d2 - (i-d1)*(i-d1))));
      break;
    case WIN_BOXCAR:
      ti->t[i] = 1.0F;
      break;
    }
  ti->t[ti->len] = ti->t[0];

}

#undef WIN_HAMMING  
#undef WIN_HANNING  
#undef WIN_BARTLETT 
#undef WIN_GAUSSIAN 
#undef WIN_KAISER   
#undef WIN_BOXCAR   

/* ************************* */
/*  end of table generators  */
/* ************************* */
#define CSYS_CDRIVER_AUCONTROL
#define ASYS_OUTDRIVER_AUDIOUNIT
#define ASYS_INDRIVER_AUDIOUNIT
#define ASYS_HASOUTPUT
#define ASYS_HASINPUT
#define ASYS_KSYNC


/* component definitions generated by sfront (audio.c) */

#define ASYS_AUDIOUNIT_COMP_TYPE kAudioUnitType_Effect
#define ASYS_AUDIOUNIT_COMP_SUBTYPE 'spcf'
#define ASYS_AUDIOUNIT_COMP_MANU 'ucBe'

/* end of sfront component definitions */


/*
#    Sfront, a SAOL to C translator    
#    This file: audiounit audio driver for sfront
#
# Copyright (c) 1999-2008, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  Redistributions of source code must retain the above copyright
#  notice, this list of conditions and the following disclaimer.
#
#  Redistributions in binary form must reproduce the above copyright
#  notice, this list of conditions and the following disclaimer in the
#  documentation and/or other materials provided with the distribution.
#
#  Neither the name of the University of California, Berkeley nor the
#  names of its contributors may be used to endorse or promote products
#  derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#    Maintainer: John Lazzaro, lazzaro@cs.berkeley.edu
*/


/****************************************************************/
/****************************************************************/
/*           Apple AudioUnit audio driver for sfront            */
/****************************************************************/

/*~~~~~~~~~~~~~*/
/* debug level */
/*_____________*/

/*  Level 0:  No debugging messages            */
/*  Level 1:  Session setup and error messages */
/*  Level 2:  + All MIDI events                */
/*  Level 3:  + All Rendering calls            */

#if !defined(ASYS_AUDIOUNIT_DEBUG_LEVEL)
#define ASYS_AUDIOUNIT_DEBUG_LEVEL 0
#endif

/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* output clipping behavior */
/*__________________________*/

#define ASYS_AUDIOUNIT_OUTPUT_CLIPPING  0

/*~~~~~~~~~~~~~~~~~~~~~*/
/* aucontrol detection */
/*_____________________*/

#if defined(CSYS_CDRIVER_AUCONTROL) || defined(CSYS_CDRIVER_AUCONTROLM)
#define ASYS_AUDIOUNIT_HAS_AUCONTROL  1
#else
#define ASYS_AUDIOUNIT_HAS_AUCONTROL  0
#endif

#if defined(CSYS_CDRIVER_AUCONTROLM)
#define ASYS_AUDIOUNIT_MIDISUPPORT  1
#else
#define ASYS_AUDIOUNIT_MIDISUPPORT  0
#endif

/*~~~~~~~~~~~~~~~~~~~*/
/* reentrant defines */
/*___________________*/

#if (ENGINE_STYLE == ENGINE_REENTRANT)
#define MY_ENGINE_PTR        My->ENGINE_PTR
#define MY_ENGINE_PTR_COMMA  My->ENGINE_PTR,
#define MY_ENGINE_PTR_ASSIGNED_TO  My->ENGINE_PTR = 
#define MY_ENGINE_PTR_DESTROY_SEMICOLON  free(MY_ENGINE_PTR);
#define MY_ENGINE_PTR_IS_NULL (MY_ENGINE_PTR == NULL)
#define MY_ENGINE_PTR_IS_NOT_NULL (MY_ENGINE_PTR != NULL)
#define ENGINE_MEMSTATUS_SEMICOLON(x) \
        asysn_audiounit_memstatus(ENGINE_PTR, ENGINE_SIZE, x);
#else
#define MY_ENGINE_PTR 
#define MY_ENGINE_PTR_COMMA 
#define MY_ENGINE_PTR_ASSIGNED_TO  
#define MY_ENGINE_PTR_DESTROY_SEMICOLON 
#define MY_ENGINE_PTR_IS_NULL (0)
#define MY_ENGINE_PTR_IS_NOT_NULL (1)
#define ENGINE_MEMSTATUS_SEMICOLON(x) 
#endif

/*~~~~~~~~~~~~~~~~~*/
/* include headers */
/*_________________*/

#if !defined(__COREAUDIO_USE_FLAT_INCLUDES__)
#include <CoreServices/CoreServices.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioUnitUtilities.h>
#else
#include <ConditionalMacros.h>
#include <CoreServices.h>
#include <AudioUnit.h>
#include <AudioUnitUtilities.h>
#endif

/* 64-bit integer headers */

#include <stdint.h>

/* for control driver socket system */

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/uio.h>
#include <fcntl.h>

/* include files for memory paging */

#include <sys/mman.h>

/* include files for synchronization */

#include <libkern/OSAtomic.h>

/*~~~~~~~~~~~~~~~~~~*/
/* endian detection */
/*__________________*/

Float32 asysn_audiounit_float32_endian_test = -1;

#define ASYS_AUDIOUNIT_FLOAT32_BIGENDIAN \
        ((((char *)(&asysn_audiounit_float32_endian_test))[0]) != 0)
 
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*     audiounit defines     */
/*___________________________*/

/* after if if-tree runs, only one audio type should be 1 */

#if (!defined(ASYS_HASINPUT) && defined(ASYS_HASOUTPUT) && defined(ASYS_ACTIVE_O))
#define ASYS_AUDIOUNIT_MUSICDEVICE 1
#else
#define ASYS_AUDIOUNIT_MUSICDEVICE 0
#endif
                                     
#if (defined(ASYS_HASINPUT) && defined(ASYS_HASOUTPUT) && defined(ASYS_ACTIVE_IO))
#define ASYS_AUDIOUNIT_EFFECT 1
#else
#define ASYS_AUDIOUNIT_EFFECT 0
#endif


/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  The Wiretap Macro Calls */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/********************/
/* level one macros */
/********************/

#if (ASYS_AUDIOUNIT_DEBUG_LEVEL >= 1)
 
#define ASYS_AUDIOUNIT_WIRETAP(a, b) do {\
      asysn_audiounit_wiretap(a, b);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(a) do {\
      asysn_audiounit_wiretap_putstring(a);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_PRINT_COMPONENT_RESULT(a) do {\
      asysn_audiounit_wiretap_print_component_result(a);\
      } while (0)

#else

#define ASYS_AUDIOUNIT_WIRETAP(a, b) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(a) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_PRINT_COMPONENT_RESULT(a) do { } while (0)

#endif

/********************/
/* level two macros */
/********************/

#if (ASYS_AUDIOUNIT_DEBUG_LEVEL >= 2)

#define ASYS_AUDIOUNIT_WIRETAP_MIDIEVENT(a, b, c, d, e, f) do {\
      asysn_audiounit_wiretap_midievent(a, b, c, d, e, f);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_M(a, b) do {\
      asysn_audiounit_wiretap(a, b);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_GETPARAMETER(a, b, c, d, e, f) do {\
      asysn_audiounit_wiretap_getparameter(a, b, c, d, e, f);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_SETPARAMETER(a, b, c, d, e, f, g) do {\
      asysn_audiounit_wiretap_setparameter(a, b, c, d, e, f, g);\
      } while (0)
#else

#define ASYS_AUDIOUNIT_WIRETAP_MIDIEVENT(a, b, c, d, e, f) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_M(a, b) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_GETPARAMETER(a, b, c, d, e, f) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_SETPARAMETER(a, b, c, d, e, f, g) do { } while (0)

#endif

/**********************/
/* level three macros */
/**********************/

#if (ASYS_AUDIOUNIT_DEBUG_LEVEL >= 3)

#define ASYS_AUDIOUNIT_WIRETAP_RENDERER(a, b, c, d, e, f) do {\
      asysn_audiounit_wiretap_renderer(a, b, c, d, e, f);\
      } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_R(a, b) do {\
      asysn_audiounit_wiretap(a, b);\
      } while (0)

#else

#define ASYS_AUDIOUNIT_WIRETAP_RENDERER(a, b, c, d, e, f) do { } while (0)
#define ASYS_AUDIOUNIT_WIRETAP_R(a, b) do { } while (0)

#endif


/****************************************************************/
/*        Part One: Sfront asys_ dummy function calls           */
/****************************************************************/
        
#if (defined(ASYS_HASOUTPUT) && !defined(ASYS_HASINPUT))

/****************************************************************/
/*        sets up audio output for a given srate/channels       */
/****************************************************************/

int asys_osetup(int srate, int ochannels, int osample, 
		char * oname, int toption) { return ASYS_DONE; }

#endif

#if (!defined(ASYS_HASOUTPUT) && defined(ASYS_HASINPUT))

/****************************************************************/
/*        sets up audio input for a given srate/channels       */
/****************************************************************/

int asys_isetup(int srate, int ichannels, int isample, 
                char * iname, int toption) { return ASYS_DONE; }

#endif

#if (defined(ASYS_HASOUTPUT) && defined(ASYS_HASINPUT))

/****************************************************************/
/*   sets up audio input and output for a given srate/channels  */
/****************************************************************/

int asys_iosetup(int srate, int ichannels, int ochannels,
                 int isample, int osample, char * iname, 
		 char * oname, int toption) { return ASYS_DONE; }
#endif

#if (defined(ASYS_HASOUTPUT)&&(!defined(ASYS_HASINPUT)))

/****************************************************************/
/*                    shuts down audio output                   */
/****************************************************************/

void asys_oshutdown(void)  { }

#endif

#if (!defined(ASYS_HASOUTPUT)&&(defined(ASYS_HASINPUT)))

/****************************************************************/
/*              shuts down audio input device                   */
/****************************************************************/

void asys_ishutdown(void) { }

#endif

#if (defined(ASYS_HASOUTPUT)&&(defined(ASYS_HASINPUT)))

/****************************************************************/
/*              shuts down audio input and output device        */
/****************************************************************/

void asys_ioshutdown(void) { }

#endif

/****************************************************************/
/*        syntactically necessary, but will never be called     */
/****************************************************************/

void asys_main(void) { }

/*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   ksync function suite  */
/*-------------------------*/

#if defined(ASYS_KSYNC)

/***********************************************************/
/*         initializes k-rate boundaries sync              */
/***********************************************************/

void ksyncinit() { }

/***********************************************************/
/*         synchronizes on k-rate boundaries               */
/***********************************************************/

float ksync() {  return 0.0F; }  /* encodes "uses no CPU" */

#endif   /* ASYS_KSYNC */


/*************************************************************/
/*         Start of Part Two: The AudioUnit Component        */
/*************************************************************/

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  AudioUnit Property Constants  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#define ASYS_AUDIOUNIT_NULL    0
#define ASYS_AUDIOUNIT_MONO    1
#define ASYS_AUDIOUNIT_STEREO  2

#define ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS ASYS_AUDIOUNIT_STEREO

#if defined(ASYS_HASINPUT)
#define ASYS_AUDIOUNIT_INPUT_MAXCHANNELS  ASYS_AUDIOUNIT_STEREO
#define ASYS_AUDIOUNIT_SUPPORTED_FORMATS  4
#else
#define ASYS_AUDIOUNIT_INPUT_MAXCHANNELS  0
#define ASYS_AUDIOUNIT_SUPPORTED_FORMATS  2
#endif

/* static input and output channels, from SAOL program */

#define ASYS_AUDIOUNIT_OUTPUT_CHANNELS (ENDBUS_output_bus - BUS_output_bus)

#if defined(ASYS_HASINPUT)
#define ASYS_AUDIOUNIT_INPUT_CHANNELS (ENDBUS_input_bus - BUS_input_bus)
#else
#define ASYS_AUDIOUNIT_INPUT_CHANNELS 0
#endif

/* element limits */

#if (ASYS_AUDIOUNIT_EFFECT)

#define ASYS_AUDIOUNIT_ELEMENT_INPUTPREF   1  /* our default */
#define ASYS_AUDIOUNIT_ELEMENT_INPUTMAX    8  /* maximum we accept */

#define ASYS_AUDIOUNIT_ELEMENT_OUTPUTPREF  1  /* our default */
#define ASYS_AUDIOUNIT_ELEMENT_OUTPUTMAX   1  /* maximum we accept */

#define ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE   8  /* for I-bound data structures */

#else

#define ASYS_AUDIOUNIT_ELEMENT_INPUTPREF   0  /* our default */
#define ASYS_AUDIOUNIT_ELEMENT_INPUTMAX    0  /* maximum we accept */

#define ASYS_AUDIOUNIT_ELEMENT_OUTPUTPREF  1  /* our default */
#define ASYS_AUDIOUNIT_ELEMENT_OUTPUTMAX   1  /* maximum we accept */

#define ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE   1  /* for I-bound data structures */

#endif

/* maximum number of frames per slice */

#define ASYS_AUDIOUNIT_FRAMES_PER_SLICE  4096

/* in-place processing not supported */

#define ASYS_AUDIOUNIT_INPLACE_PROCESSING 0 

/* number of instruments -- only used by MusicDevice */

#define ASYS_AUDIOUNIT_INSTRUMENT_COUNT  1 

/* constants for property listeners */

#define ASYS_AUDIOUNIT_PROPLISTEN_PARAMETERINFO  0
#define ASYS_AUDIOUNIT_PROPLISTEN_PARAMETERLIST  1
#define ASYS_AUDIOUNIT_PROPLISTEN_CPULOAD        2
#define ASYS_AUDIOUNIT_PROPLISTEN_LATENCY        3
#define ASYS_AUDIOUNIT_PROPLISTEN_TAILTIME       4
#define ASYS_AUDIOUNIT_PROPLISTEN_FACTORYPRESETS 5
#define ASYS_AUDIOUNIT_PROPLISTEN_RENDERQUALITY  6
#define ASYS_AUDIOUNIT_PROPLISTEN_CURRENTPRESET  7
#define ASYS_AUDIOUNIT_PROPLISTEN_PRESENTPRESET  8
#define ASYS_AUDIOUNIT_PROPLISTEN_STREAMFROMDISK 9
#define ASYS_AUDIOUNIT_PROPLISTEN_MAXIMUMFRAMESPERSLICE 10
#define ASYS_AUDIOUNIT_PROPLISTEN_ARRAYSIZE     11   /* ++ for each property added */

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  reset/paging behaviors   */
/*___________________________*/

/* active reset does Uninitialize/Initialize for ResetSelect */

#define ASYS_AUDIOUNIT_PASSIVE_RESET 0
#define ASYS_AUDIOUNIT_ACTIVE_RESET  1
#define ASYS_AUDIOUNIT_RESET_TYPE ASYS_AUDIOUNIT_PASSIVE_RESET

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* istate-machine constants  */
/*___________________________*/

#define ASYS_AUDIOUNIT_ISTATE_ABSENT           0 
#define ASYS_AUDIOUNIT_ISTATE_INSTANCE         1
#define ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING 2
#define ASYS_AUDIOUNIT_ISTATE_FLUX             3
#define ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING     4
#define ASYS_AUDIOUNIT_ISTATE_ENGINE           5
#define ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING   6
#define ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER    7

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* istate-machine return values  */
/*_______________________________*/

#define ASYS_AUDIOUNIT_IRESPONSE_COMPLETE               0
#define ASYS_AUDIOUNIT_IRESPONSE_INITIALIZE             1
#define ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE           2
#define ASYS_AUDIOUNIT_IRESPONSE_CLOSE                  3
#define ASYS_AUDIOUNIT_IRESPONSE_RENDER                 4
#define ASYS_AUDIOUNIT_IRESPONSE_WIRE                   5
#define ASYS_AUDIOUNIT_IRESPONSE_REINITIALIZE           6
#define ASYS_AUDIOUNIT_IRESPONSE_REUNINITIALIZE         7
#define ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE_AND_CLOSE 8
#define ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL             9
#define ASYS_AUDIOUNIT_IRESPONSE_ERROR                 10

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   Control Driver Constants     */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* retry limit for socket writing */

#define ASYS_AUDIOUNIT_RETRY_MAX  256

/* bitfield constants for MIDIevent flags variable */

#define ASYS_AUDIOUNIT_MIDIFLAGS_WAITING 0x01u  /* queuing flag bit */

/* bitfield constants for SASLevent flags variable */

#define ASYS_AUDIOUNIT_SASLFLAGS_WAITING 0x01u  /* queuing flag bit */

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   SAOL Parameter Constants    */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#define ASYS_AUDIOUNIT_PARAMETER_DEFAULT_UNIT kAudioUnitParameterUnit_Generic
#define ASYS_AUDIOUNIT_PARAMETER_DEFAULT_MINIMUM 0.0F
#define ASYS_AUDIOUNIT_PARAMETER_DEFAULT_MAXIMUM 1.0F
#define ASYS_AUDIOUNIT_PARAMETER_DEFAULT_DEFAULT 0.5F

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*           TypeDefs             */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  typedef for a MIDI event   */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_MIDIevent {
  unsigned char cmd;
  unsigned char d0;
  unsigned char d1;
  unsigned char flags;
  int kcycleidx;
} asysn_audiounit_MIDIevent;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  typedef for a SASL event   */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_SASLevent {
  int index;
  Float32 value;
  unsigned char flags;
  int kcycleidx;
} asysn_audiounit_SASLevent;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  typedef for SAOL parameter */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_saolparam {
  Float32 value;
  int index;
  int use;
} asysn_audiounit_saolparam;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  typedef for a listener proc  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_proplisten {
  AudioUnitPropertyListenerProc lproc;
  void * lrefcon;
  struct asysn_audiounit_proplisten * next;
} asysn_audiounit_proplisten;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  typedef for a render notify proc  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_rendernotify {
  AURenderCallback nproc;
  void * nrefcon;
  struct asysn_audiounit_rendernotify * next;
} asysn_audiounit_rendernotify;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* typedef for instance state */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct asysn_audiounit_InstanceState {

  ComponentInstance component;  /* component instance ID */
  volatile int istate;          /* initialization state machine */
  ENGINE_PTR_DECLARE_SEMICOLON  /* engine pointer for the instance */

  /*~~~~~~~~~~~~~~~~~~~~~~~*/
  /* mirrored engine state */
  /*~~~~~~~~~~~~~~~~~~~~~~~*/

  UInt32 acycle;
  Float32 krate;
  volatile uint64_t acycleidx_kcycleidx;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* current value of properties */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  /* common to Effects and MusicDevices */

  AudioUnitConnection MakeConnection[ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE];
  Float32 CPULoad;
  OSSpinLock lock_CPULoad;
  AudioStreamBasicDescription OutputStreamFormat;
  OSType SRCAlgorithm;
  UInt32 InputElementCount;
  UInt32 OutputElementCount;
  Float64 Latency;
  AUChannelInfo SupportedNumChannels[ASYS_AUDIOUNIT_SUPPORTED_FORMATS];
  UInt32 MaximumFramesPerSlice;
  Float64 TailTime;
  UInt32 BypassEffect;
  UInt32 LastBypassEffect;
  OSStatus LastRenderError;
  AURenderCallbackStruct SetRenderCallback[ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE];
  asysn_audiounit_rendernotify * rendernotify;
  OSSpinLock lock_rendernotify;
  UInt32 RenderQuality;
  UInt32 InPlaceProcessing;
  AUPreset PresentPreset;
  OSSpinLock lock_PresentPreset;
  UInt32 OfflineRender;
  Float64 PresentationLatency;
  UInt32 StreamFromDisk;
  asysn_audiounit_proplisten * proplisteners[ASYS_AUDIOUNIT_PROPLISTEN_ARRAYSIZE];
  OSSpinLock lock_proplisteners;
  OSSpinLock lock_sampledelivery;  /* for MakeConnection and SetRenderCallback */
  
  /* exclusive to MusicDevices and MusicEffects (?) */

  UInt32 InstrumentCount;

  /* properties exclusive to Effects */

  AudioStreamBasicDescription InputStreamFormat[ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE];

  /*~~~~~~~~~~~~~~~~~~~~*/
  /* per-instance state */
  /*~~~~~~~~~~~~~~~~~~~~*/

  /* data output array: common to Effects and Music Devices */

  Float32 * mData_Output[ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS];

  /* rendering templates and data carriers: exclusive to Effects */

  AudioBufferList * AudioBufferTemplate;
  AudioBufferList * AudioBufferCarrier;

  /*~~~~~~~~~~~~~~~~~~~~~*/
  /* aup_ property state */
  /*~~~~~~~~~~~~~~~~~~~~~*/

  int num_saolparams;
  AudioUnitParameterID * parameterlist;
  AudioUnitParameterInfo * parameterinfo;
  int * pvs_size;
  char *** pvs_cstr;
  asysn_audiounit_saolparam * saolparam;

  int num_factory;
  Float32 ** factorypreset_values;
  AUPreset * factorypreset_info;
  CFMutableArrayRef factorypreset_array;

  /*~~~~~~~~~~~~~*/
  /* ksync state */
  /*~~~~~~~~~~~~~*/

  UInt64 ksync_timespent;  /* total time spent in one kpass */
  Float32 ksync_normalize;    /* unit translations: 1/ticks to Hz */
  Float32 ksync_last_cpuload;   /* used for missing data heuristic  */

  /*~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* state passed to engine */
  /*~~~~~~~~~~~~~~~~~~~~~~~~*/

  char componentname[32];

  char mpipeflag[32];
  char mpipevalue[32];

  char spipeflag[32];
  char spipevalue[32];

  char * argv[5];
  int argc;

  /*~~~~~~~~~~~~~~~~~*/
  /* aucontrol state */
  /*~~~~~~~~~~~~~~~~~*/

  int mpipepair[2];
  int spipepair[2];

} asysn_audiounit_InstanceState;


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* The page-management system  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int asysn_audiounit_opencount = 0;   /* number of audiounit instances */
OSSpinLock asysn_audiounit_lock_opencount;      /* protects opencount */

/********************************************/
/* set memory status of a region of memory  */
/********************************************/

void asysn_audiounit_memstatus(void * addr, size_t len, int advice)

{
  madvise((caddr_t) addr, len, advice);
  return;
}


/*~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~*/
/* The ksync system  */
/*~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~*/

/*********************************************/
/* startup initialization for ksync system   */
/*********************************************/

void asysn_audiounit_ksyncinit(asysn_audiounit_InstanceState * My)

{
  My->ksync_timespent = 0;

  OSSpinLockLock(&(My->lock_CPULoad));

  My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());
  if (My->CPULoad)
    My->ksync_normalize = My->ksync_normalize/My->CPULoad;

  OSSpinLockUnlock(&(My->lock_CPULoad));

  My->ksync_last_cpuload = 0.0F;   
}

/********************************************/
/* computes cpuload value at end of a kpass */
/********************************************/

Float32 asysn_audiounit_ksync(asysn_audiounit_InstanceState * My)

{
  UInt64 now = AudioGetCurrentHostTime();
  Float32 ret;

  ret = My->ksync_normalize*((Float32)(now - My->ksync_timespent));

  if (ret > 1.0F)
    ret = 1.0F;

  if (ret < 1.0F)
    My->ksync_last_cpuload = ret;
  else
    {
      ret = My->ksync_last_cpuload;
      My->ksync_last_cpuload = 1.0F;
    }

  My->ksync_timespent = now;

  return ret;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*    Fast Dispatch Functions     */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/***********************************************/
/* Helper Routine for MIDIEvent Fast Dispatch  */
/***********************************************/

ComponentResult asysn_audiounit_sendMIDIevent(asysn_audiounit_MIDIevent * MIDIevent,
					      asysn_audiounit_InstanceState * My)

{
  uint64_t acycleidx_kcycleidx;
  int32_t acycleidx, kcycleidx;
  int write_failed = 0;
  int retry = 0;
  int acount;

  if (!ASYS_AUDIOUNIT_HAS_AUCONTROL)
    return noErr;

  /* on entry, MIDIevent->kcycle holds offset of event (in samples) from the */
  /* start of the next buffer. This start time in the sfront engine is coded */
  /* by the tuple (kcycleidx, acycleidx).                                    */

  acycleidx_kcycleidx = My->acycleidx_kcycleidx;
  kcycleidx = (int32_t)(acycleidx_kcycleidx & 0x00000000FFFFFFFF);
  acycleidx = (int32_t)(acycleidx_kcycleidx >> 32);

  acount = acycleidx + MIDIevent->kcycleidx;
  MIDIevent->kcycleidx = kcycleidx;

  /* initialization of acycleidx_kcycleidx skirts zero-time condition issue */

  while (acount >= My->acycle)
    {
      acount -= My->acycle;
      MIDIevent->kcycleidx++;
    }

  while (write(My->mpipepair[1], MIDIevent, sizeof(asysn_audiounit_MIDIevent)) < 0)
    if ((errno != EINTR) || (++retry >= ASYS_AUDIOUNIT_RETRY_MAX))
      {
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING("\tSocket write error in sendMIDIevent\n\n");
	write_failed = 1;
	break;
      }
 
  return noErr;
}

/***********************************/
/* MIDIEvent Fast Dispatch Handler */
/***********************************/

ComponentResult asysn_audiounit_MyMIDIEventProc(void * inComponentStorage,
						UInt32 inStatus,
						UInt32 inData1,
						UInt32 inData2,
						UInt32 inOffsetSampleFrame)
{
  asysn_audiounit_MIDIevent MIDIevent;
  int result;

  if (!inComponentStorage)           /* avoid race condition */
    return noErr;

  MIDIevent.cmd = (unsigned char) inStatus;
  MIDIevent.d0 = (unsigned char)inData1;
  MIDIevent.d1 = (unsigned char)inData2;
  MIDIevent.flags = ASYS_AUDIOUNIT_MIDIFLAGS_WAITING;
  MIDIevent.kcycleidx = (int) inOffsetSampleFrame;

  result = asysn_audiounit_sendMIDIevent
    (&MIDIevent, (asysn_audiounit_InstanceState *) inComponentStorage);

  ASYS_AUDIOUNIT_WIRETAP_MIDIEVENT(inComponentStorage, inStatus, inData1, 
				   inData2, inOffsetSampleFrame, "Fast Dispatch");

  return result;
}

/**************************************************/
/* Helper Routine for GetParameter Fast Dispatch  */
/**************************************************/

ComponentResult asysn_audiounit_getSASLevent(AudioUnitParameterID inID, 
					     Float32 * outValue,
					     asysn_audiounit_InstanceState * My)

{
  if (inID >= My->num_saolparams)
    return kAudioUnitErr_InvalidParameter;

  *outValue = My->saolparam[inID].value;
  return noErr;
}

/**************************************/
/* GetParameter Fast Dispatch Handler */
/**************************************/

ComponentResult asysn_audiounit_MyGetParameterProc(void * inComponentStorage,
						   AudioUnitParameterID inID,
						   AudioUnitScope inScope,
						   AudioUnitElement inElement,
						   Float32 * outValue)
{
  int result;

  if (!inComponentStorage)           /* avoid race condition */
    return noErr;

  if (inScope != kAudioUnitScope_Global)  /* inScope */
    return kAudioUnitErr_InvalidScope;

  result = asysn_audiounit_getSASLevent
    (inID, outValue, (asysn_audiounit_InstanceState *) inComponentStorage);

  ASYS_AUDIOUNIT_WIRETAP_GETPARAMETER(inComponentStorage, inID, inScope, 
				      inElement, outValue, "Fast Dispatch");

  return result;
}

/**************************************************/
/* Helper Routine for SetParameter Fast Dispatch  */
/**************************************************/

ComponentResult asysn_audiounit_sendSASLevent(asysn_audiounit_SASLevent * SASLevent,
					      asysn_audiounit_InstanceState * My)

{
  uint64_t acycleidx_kcycleidx;
  int32_t acycleidx, kcycleidx;
  int write_failed = 0;
  int retry = 0;
  int acount;

  if (!ASYS_AUDIOUNIT_HAS_AUCONTROL)
    return noErr;

  /* on entry:                                                               */
  /*            SASLevent->index holds AudioUnit index (not SAOL index) and  */
  /*            has been range-checked.                                      */
  /*                                                                         */
  /*            SASLevent->value is range-checked.                           */
  /*                                                                         */
  /*            SASLevent->flags is not set.                                 */
  /*                                                                         */
  /*            SASLevent->kcycle holds offset of event (in samples) from    */
  /*            the start of the next buffer. This start time in the sfront  */
  /*            engine is coded by the tuple (kcycleidx, acycleidx).         */

  SASLevent->index = My->saolparam[SASLevent->index].index;
  SASLevent->flags = ASYS_AUDIOUNIT_SASLFLAGS_WAITING;

  acycleidx_kcycleidx = My->acycleidx_kcycleidx;
  kcycleidx = (int32_t)(acycleidx_kcycleidx & 0x00000000FFFFFFFF);
  acycleidx = (int32_t)(acycleidx_kcycleidx >> 32);

  acount = acycleidx + SASLevent->kcycleidx;
  SASLevent->kcycleidx = kcycleidx;

  /* initialization of acycleidx_kcycleidx skirts zero-time condition issue */

  while (acount >= My->acycle)
    {
      acount -= My->acycle;
      SASLevent->kcycleidx++;
    }

  while (write(My->spipepair[1], SASLevent, sizeof(asysn_audiounit_SASLevent)) < 0)
    if ((errno != EINTR) || (++retry >= ASYS_AUDIOUNIT_RETRY_MAX))
      {
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING("\tSocket write error in sendSASLevent\n\n");
	write_failed = 1;
	break;
      }
 
  return noErr;
}

/**************************************/
/* SetParameter Fast Dispatch Handler */
/**************************************/

ComponentResult asysn_audiounit_MySetParameterProc(void * inComponentStorage,
						   AudioUnitParameterID inID,
						   AudioUnitScope inScope,
						   AudioUnitElement inElement,
						   Float32 inValue,
						   UInt32 inBufferOffsetInFrames)
{
  asysn_audiounit_InstanceState * My;
  asysn_audiounit_SASLevent SASLevent;
  int result;

  My = ((asysn_audiounit_InstanceState *) inComponentStorage);

  if (!My)           /* avoid race condition */
    return noErr;

  if ((inID < 0) || (inID >= My->num_saolparams))  /* range-check ID */
    return kAudioUnitErr_InvalidParameter;

  if (inScope != kAudioUnitScope_Global)  /* inScope */
    return kAudioUnitErr_InvalidScope;

  if (inValue > My->parameterinfo[inID].maxValue)
    inValue = My->parameterinfo[inID].maxValue;

  if (inValue < My->parameterinfo[inID].minValue)
    inValue = My->parameterinfo[inID].minValue;

  SASLevent.index = inID;
  SASLevent.value = My->saolparam[inID].value = inValue;
  SASLevent.kcycleidx = (int)(inBufferOffsetInFrames);

  result = asysn_audiounit_sendSASLevent(&SASLevent, My);

  ASYS_AUDIOUNIT_WIRETAP_SETPARAMETER(My, inID, inScope, 
				      inElement, inValue, inBufferOffsetInFrames, 
				      "Fast Dispatch");

  return result;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* AUEventListenerNotify support  */
/*                                */
/* To let global variable writes  */
/* by SAOL programs be seen in UI */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/************************************************/
/* checks all variables, does update if needed  */
/************************************************/

void asysn_audiounit_eventlistenernotify(asysn_audiounit_InstanceState * My)

{
  ENGINE_PTR_DECLARE_SEMICOLON
  AudioUnitEvent event;
  Float32 value;
  int i;

  ENGINE_PTR_ASSIGNED_TO  MY_ENGINE_PTR;  /* for NG() reference */

  for (i = 0; i < My->num_saolparams; i++)
    if (My->saolparam[i].use & CSYS_WRITTEN)
      if ((value = NG(My->saolparam[i].index)) != My->saolparam[i].value)
	{
	  if (value > My->parameterinfo[i].maxValue)
	    NG(My->saolparam[i].index) = value = My->parameterinfo[i].maxValue;
	  if (value < My->parameterinfo[i].minValue)
	    NG(My->saolparam[i].index) = value = My->parameterinfo[i].minValue;
	  
	  if (value != My->saolparam[i].value)
	    {
	      My->saolparam[i].value = value;
	      event.mEventType = kAudioUnitEvent_ParameterValueChange; 
	      event.mArgument.mParameter.mAudioUnit = My->component; 
	      event.mArgument.mParameter.mParameterID = i;
	      event.mArgument.mParameter.mScope = kAudioUnitScope_Global; 
	      event.mArgument.mParameter.mElement = 0; 
	      OSMemoryBarrier();
	      AUEventListenerNotify(NULL, NULL, &event);
	    }
	}
}

/*------------------------------------------*/
/*  Specialized Renderers for Each AU Type  */
/*------------------------------------------*/

#if ASYS_AUDIOUNIT_MUSICDEVICE 

/************************************************************/
/* Helper Routine For Fast-Dispatch Renderer -- MusicDevice */
/************************************************************/

void asysn_audiounit_activeo_renderer(asysn_audiounit_InstanceState * My,
				      AudioBufferList * ioData)

{
  ENGINE_PTR_DECLARE_SEMICOLON
  UInt32 frames = ioData->mBuffers[0].mDataByteSize/sizeof(Float32);
  Float32 eincr, escale;
  UInt32 do_effect, do_bypass;
  Float32 * out_left, * out_right;
  Float32 left;
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
  Float32 right;
#endif

  ENGINE_PTR_ASSIGNED_TO  MY_ENGINE_PTR;


  if (My->BypassEffect == 0)
    {
      if (My->LastBypassEffect == 0)
	{
	  do_effect = 1;
	  do_bypass = 0;
	  eincr = escale = 0.0F;
	}
      else
	{
	  do_effect = 1;
	  do_bypass = 1;
	  eincr = escale = 1.0F/(frames + 1);
	  My->LastBypassEffect = 0;
	}
    }
  else
    {
      if (My->LastBypassEffect == 1)
	{
	  do_effect = 0;
	  do_bypass = 1;
	  eincr = escale = 0.0F;
	}
      else
	{
	  do_effect = 1;
	  do_bypass = 1;
	  eincr = - 1.0F/(frames + 1);
	  escale = 1.0F + eincr;
	  My->LastBypassEffect = 1;
	}
    }

  out_left = (Float32 *) (ioData->mBuffers[0].mData);

  if (ioData->mNumberBuffers > ASYS_AUDIOUNIT_MONO)
    out_right = (Float32 *) (ioData->mBuffers[1].mData);
  else
    out_right = NULL;

  while (frames)
    if (EV(acycleidx) < EV(ACYCLE))
      {
	memset(&(TB(0)), 0, ENDBUS*sizeof(ASYS_OTYPE));  /* set bus to silence */
	main_apass(ENGINE_PTR);                         /* compute samples    */

	if (do_effect)
	  {
	    left = TB(BUS_output_bus);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
	    if ((left > 1.0F) || (left < -1.0F))
	      left = (left > 0) ? 1.0F : -1.0F;
#endif	    
	    *out_left = left;
	    
	    if (out_right)
	      {
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS == ASYS_AUDIOUNIT_MONO)
		*out_right = left;
#else
		right = TB(BUS_output_bus + 1);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
		if ((right > 1.0F) || (right < -1.0F))
		  right = (right > 0) ? 1.0F : -1.0F;
#endif
		*out_right = right;
#endif
	      }
	    else
	      {
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
		right = TB(BUS_output_bus + 1);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
		if ((right > 1.0F) || (right < -1.0F))
		  right = (right > 0) ? 1.0F : -1.0F;
#endif
		*out_left += right;
		*out_left *= 0.5F;
#endif
	      }

	    if (do_bypass)
	      {
		*out_left *= escale;
		if (out_right)
		  *out_right *= escale;
		escale += eincr;
	      }
	  }

	if (do_bypass && !do_effect)
	  {
	    *out_left = 0.0F;

	    if (out_right)
	      *out_right = 0.0F;
	  }

	out_left++;

	if (out_right)
	  out_right++;

	EV(acycleidx)++; frames--;     /* update all positions */
      }
    else
      {
	EV(acycleidx) = 0;
	if (EV(pass) == APASS)
	  {
	    EV(cpuload) = asysn_audiounit_ksync(My);
  	    EV(kcycleidx)++;         /* we run forever; don't test against endkcycle */
	  }
	EV(pass) = IPASS;
	EV(scorebeats) = EV(scoremult)*(EV(kcycleidx) - EV(kbase)) + EV(scorebase);
	EV(absolutetime) = (EV(kcycleidx) - 1)*EV(KTIME);
	main_ipass(ENGINE_PTR);
	EV(pass) = KPASS;
	main_control(ENGINE_PTR);
	main_kpass(ENGINE_PTR);      /* we run forever; don't check return value */
	if (ASYS_AUDIOUNIT_PARAMETERS_WRITTEN)
	  asysn_audiounit_eventlistenernotify(My);
	EV(pass) = APASS;
      }

  My->acycleidx_kcycleidx = (((uint64_t) (EV(kcycleidx))) & 0x00000000FFFFFFFF) | 
    (((uint64_t) (EV(acycleidx))) << 32);
}

/*****************************************/
/* Fast-Dispatch Renderer -- MusicDevice */
/*****************************************/

ComponentResult asysn_audiounit_MyRenderer(void * inComponentStorage,
					   AudioUnitRenderActionFlags * ioActionFlags,
					   const AudioTimeStamp * inTimeStamp,
					   UInt32 inOutputBusNumber,
					   UInt32 inNumberFrames,
					   AudioBufferList * ioData)
{
  UInt32 error_code;
  char message[128];
  asysn_audiounit_rendernotify * notify;
  AudioUnitRenderActionFlags action_flags;
  asysn_audiounit_InstanceState * My = ((asysn_audiounit_InstanceState *) 
					inComponentStorage);
  int i, iresponse;

  if (inNumberFrames > ASYS_AUDIOUNIT_FRAMES_PER_SLICE)
    return kAudioUnitErr_TooManyFramesToProcess;

  if (!My)
    return kAudioUnitErr_Uninitialized;   /* avoid race condition */

  switch(iresponse = asysn_audiounit_realtime_enterstate(My)) {
  case ASYS_AUDIOUNIT_IRESPONSE_WIRE:
  case ASYS_AUDIOUNIT_IRESPONSE_RENDER:
    break;
  case ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL:
  case ASYS_AUDIOUNIT_IRESPONSE_ERROR:
  default:
    return kAudioUnitErr_Uninitialized;   /* avoid race condition, bugs */
  }

  /* supply return buffers (only!) if needed */

  for (i = 0; i < ioData->mNumberBuffers; i++)
    if (ioData->mBuffers[i].mData == NULL)
      ioData->mBuffers[i].mData = (void *) My->mData_Output[i];

  /* do PreRender notification */

  OSSpinLockLock(&(My->lock_rendernotify));

  notify = My->rendernotify;
  while (notify)
    {
      if (notify->nproc)
	{
	  action_flags = kAudioUnitRenderAction_PreRender;

	  if (error_code = ((* notify->nproc)
			    (notify->nrefcon, &action_flags, 
			     inTimeStamp, (AudioUnitElement) 0, inNumberFrames,
			     ioData)))
	    {
	      sprintf(message, "\nasysn_audiounit_MyRenderer"
		      "\n\tError: PreRender Rendernotify callback" 
		      " returned OSType %u\n", (unsigned int) error_code);
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	    }
	}
      notify = notify->next;
    }

  OSSpinLockUnlock(&(My->lock_rendernotify));

  /* the actual rendering done by the audio unit */

  if (iresponse == ASYS_AUDIOUNIT_IRESPONSE_RENDER)
    {
      My->ksync_timespent = AudioGetCurrentHostTime() - My->ksync_timespent;
      asysn_audiounit_activeo_renderer(My, ioData);
      My->ksync_timespent = AudioGetCurrentHostTime() - My->ksync_timespent;
    }
  else   /* for _WIRE, zero all buffers */
    for (i = 0; i < ioData->mNumberBuffers; i++)  
      memset(ioData->mBuffers[i].mData, 0, sizeof(Float32)*inNumberFrames);

  /* do PostRender notification */

  OSSpinLockLock(&(My->lock_rendernotify));

  notify = My->rendernotify;
  while (notify)
    {
      if (notify->nproc)
	{
	  action_flags = kAudioUnitRenderAction_PostRender;

	  if (error_code = ((* notify->nproc)
			    (notify->nrefcon, &action_flags, 
			     inTimeStamp, (AudioUnitElement) 0, inNumberFrames,
			     ioData)))
	    {
	      sprintf(message, "\nasysn_audiounit_MyRenderer"
		      "\n\tError: PostRender Rendernotify callback" 
		      " returned OSType %u\n", (unsigned int) error_code);
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	    }
	}
      notify = notify->next;
    }

  OSSpinLockUnlock(&(My->lock_rendernotify));

  asysn_audiounit_realtime_exitstate(My);

  return noErr;
}

#endif /* ASYS_AUDIOUNIT_MUSICDEVICE */


#if ASYS_AUDIOUNIT_EFFECT

/********************************************************/
/* Helper Routine For Fast-Dispatch Renderer -- Effects */
/********************************************************/

void asysn_audiounit_activeio_renderer(asysn_audiounit_InstanceState * My,
				       AudioBufferList * ioData)

{
  ENGINE_PTR_DECLARE_SEMICOLON
  UInt32 frames = ioData->mBuffers[0].mDataByteSize/sizeof(Float32);
  Float32 eincr, bincr, escale, bscale;
  UInt32 do_effect, do_bypass;
  Float32 * out_left, * out_right, * in_left, * in_right;
  Float32 left;
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
  Float32 right;
#endif

  ENGINE_PTR_ASSIGNED_TO  MY_ENGINE_PTR;

  if (My->BypassEffect == 0)
    {
      if (My->LastBypassEffect == 0)
	{
	  do_effect = 1;
	  do_bypass = 0;
	  eincr = bincr = bscale = escale = 0.0F;
	}
      else
	{
	  do_effect = 1;
	  do_bypass = 1;
	  eincr = escale = 1.0F/(frames + 1);
	  bincr = - eincr;
	  bscale = 1.0F - eincr;
	  My->LastBypassEffect = 0;
	}
    }
  else
    {
      if (My->LastBypassEffect == 1)
	{
	  do_effect = 0;
	  do_bypass = 1;
	  eincr = bincr = bscale = escale = 0.0F;
	}
      else
	{
	  do_effect = 1;
	  do_bypass = 1;
	  bincr = bscale = 1.0F/(frames + 1);
	  eincr = - bincr;
	  escale = 1.0F - bincr;
	  My->LastBypassEffect = 1;
	}
    }

  out_left = (Float32 *) (ioData->mBuffers[0].mData);

  if (ioData->mNumberBuffers > ASYS_AUDIOUNIT_MONO)
    out_right = (Float32 *) (ioData->mBuffers[1].mData);
  else
    out_right = NULL;

  in_left = (Float32 *) (My->AudioBufferCarrier->mBuffers[0].mData);

  if (My->AudioBufferCarrier->mNumberBuffers > ASYS_AUDIOUNIT_MONO)
    in_right = (Float32 *) (My->AudioBufferCarrier->mBuffers[1].mData);
  else
    in_right = NULL;

  while (frames)
    if (EV(acycleidx) < EV(ACYCLE))
      {
	memset(&(TB(0)), 0, ENDBUS*sizeof(ASYS_OTYPE));  /* set bus to silence */

	TB(BUS_input_bus) = *in_left;
 
	if (in_right)
	  {
#if (ASYS_AUDIOUNIT_INPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
	    TB(BUS_input_bus + 1) = *in_right;
#else
	    TB(BUS_input_bus) += *in_right;
	    TB(BUS_input_bus) *= 0.5F;
#endif
	  }
	else
	  {
#if (ASYS_AUDIOUNIT_INPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
	    TB(BUS_input_bus + 1) = *in_left;
#endif
	  }

	main_apass(ENGINE_PTR);                   /* compute samples    */

	if (do_effect)
	  {
	    left = TB(BUS_output_bus);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
	    if ((left > 1.0F) || (left < -1.0F))
	      left = (left > 0) ? 1.0F : -1.0F;
#endif
	    *out_left = left;

	    if (out_right)
	      {
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS == ASYS_AUDIOUNIT_MONO)
		*out_right = left;
#else
		right = TB(BUS_output_bus + 1);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
		if ((right > 1.0F) || (right < -1.0F))
		  right = (right > 0) ? 1.0F : -1.0F;
#endif
		*out_right = right;
#endif
	      }
	    else
	      {
#if (ASYS_AUDIOUNIT_OUTPUT_CHANNELS > ASYS_AUDIOUNIT_MONO)
		right = TB(BUS_output_bus + 1);

#if (ASYS_AUDIOUNIT_OUTPUT_CLIPPING)
		if ((right > 1.0F) || (right < -1.0F))
		  right = (right > 0) ? 1.0F : -1.0F;
#endif
		*out_left += right;
		*out_left *= 0.5F;
#endif
	      }
	  }

	if (do_bypass)
	  {
	    if (!do_effect)
	      *out_left = *in_left;
	    else
	      {
		if (out_right || !in_right)
		  *out_left = escale*(*out_left) + bscale*(*in_left);
		else
		  *out_left = escale*(*out_left) + bscale*0.5F*(*in_left);
	      }

	    if (out_right)
	      {
		if (!do_effect)
		  {
		    if (in_right)
		      *out_right = *in_right;
		    else 
		      *out_right = *in_left;
		  }
		else
		  {
		    if (in_right)
		      *out_right = escale*(*out_right) + bscale*(*in_right);
		    else 
		      *out_right = escale*(*out_right) + bscale*(*in_left);
		  }
	      }
	    else
	      {
		if (in_right)
		  {
		    if (!do_effect)
		      *out_left = 0.5F*(*out_left + *in_right);
		    else
		      *out_left = *out_left + bscale*0.5F*(*in_right);
		  }
	      }

	    if (do_effect)
	      {
		bscale += bincr;
		escale += eincr;
	      }
	  }

	in_left++;
	out_left++;

	if (in_right)
	  in_right++;

	if (out_right)
	  out_right++;

	EV(acycleidx)++; frames--;  /* updates */
      }
    else
      {
	EV(acycleidx) = 0;
	if (EV(pass) == APASS)
	  {      
	    EV(cpuload) = asysn_audiounit_ksync(My);
	    EV(kcycleidx)++;         /* we run forever; don't test against endkcycle */
	  }
	EV(pass) = IPASS;
	EV(scorebeats) = EV(scoremult)*(EV(kcycleidx) - EV(kbase)) + EV(scorebase);
	EV(absolutetime) = (EV(kcycleidx) - 1)*EV(KTIME);
	main_ipass(ENGINE_PTR);
	EV(pass) = KPASS;
	main_control(ENGINE_PTR);
	main_kpass(ENGINE_PTR);      /* we run forever; don't check return value */
	if (ASYS_AUDIOUNIT_PARAMETERS_WRITTEN)
	  asysn_audiounit_eventlistenernotify(My);
	EV(pass) = APASS;
      }

  My->acycleidx_kcycleidx = (((uint64_t) (EV(kcycleidx))) & 0x00000000FFFFFFFF) | 
    (((uint64_t) (EV(acycleidx))) << 32);
}

/**********************************************************/
/* Helper Routine For Fast-Dispatch Renderer -- Pass-thru */
/**********************************************************/

void asysn_audiounit_activeio_passthru(asysn_audiounit_InstanceState * My,
				       AudioBufferList * ioData)

{
  UInt32 frames = ioData->mBuffers[0].mDataByteSize/sizeof(Float32);
  Float32 * out_left, * out_right, * in_left, * in_right;

  out_left = (Float32 *) (ioData->mBuffers[0].mData);

  if (ioData->mNumberBuffers > ASYS_AUDIOUNIT_MONO)
    out_right = (Float32 *) (ioData->mBuffers[1].mData);
  else
    out_right = NULL;

  in_left = (Float32 *) (My->AudioBufferCarrier->mBuffers[0].mData);

  if (My->AudioBufferCarrier->mNumberBuffers > ASYS_AUDIOUNIT_MONO)
    in_right = (Float32 *) (My->AudioBufferCarrier->mBuffers[1].mData);
  else
    in_right = NULL;

  while (frames)
    {
      *out_left = *in_left;

      if (out_right)
	{
	  if (in_right)
	    *out_right = *in_right;
	  else 
	    *out_right = *in_left;
	}
      else
	if (in_right)
	  *out_left = 0.5F*(*out_left + *in_right);
      
      in_left++;
      out_left++;
      
      if (in_right)
	in_right++;
      
      if (out_right)
	out_right++;
      
      frames--;
    }
}


/*************************************/
/* Fast-Dispatch Renderer -- Effects */
/*************************************/

ComponentResult asysn_audiounit_MyRenderer(void * inComponentStorage,
					   AudioUnitRenderActionFlags * ioActionFlags,
					   const AudioTimeStamp * inTimeStamp,
					   UInt32 inOutputBusNumber,
					   UInt32 inNumberFrames,
					   AudioBufferList * ioData)
{
  UInt32 error_code;
  char message[128];  
  asysn_audiounit_rendernotify * notify;
  AudioUnitRenderActionFlags action_flags;
  AudioTimeStamp AudioTimeStampCarrier;
  asysn_audiounit_InstanceState * My = ((asysn_audiounit_InstanceState *) 
					inComponentStorage);
  ComponentResult return_value;
  int i, iresponse;

  if (inNumberFrames > ASYS_AUDIOUNIT_FRAMES_PER_SLICE)
    return kAudioUnitErr_TooManyFramesToProcess;

  if (!My)
    return kAudioUnitErr_Uninitialized;    /* avoid race condition */

  switch(iresponse = asysn_audiounit_realtime_enterstate(My)) {
  case ASYS_AUDIOUNIT_IRESPONSE_WIRE:
  case ASYS_AUDIOUNIT_IRESPONSE_RENDER:
    break;
  case ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL:
  case ASYS_AUDIOUNIT_IRESPONSE_ERROR:
  default:
    return kAudioUnitErr_Uninitialized;   /* avoid race condition, bugs */
  }

  return_value = noErr;
  action_flags = 0;

  /* prepare AudioTimeStampCarrier */

  memset(&AudioTimeStampCarrier, 0, sizeof(AudioTimeStamp));
  AudioTimeStampCarrier.mSampleTime = inTimeStamp->mSampleTime;
  AudioTimeStampCarrier.mFlags = kAudioTimeStampSampleTimeValid;

  /* prepare AudioBufferCarrier */

  memcpy(My->AudioBufferCarrier, My->AudioBufferTemplate, 
	 sizeof(UInt32) + My->AudioBufferTemplate->mNumberBuffers*sizeof(AudioBuffer));
 
  for (i = 0; i < My->AudioBufferCarrier->mNumberBuffers; i++)
    My->AudioBufferCarrier->mBuffers[i].mDataByteSize = 
      ioData->mBuffers[0].mDataByteSize;

  /* retrieve audio samples to process, log errors */ 
  /*   assumptions: 
     - a callback returning noErr returns the amount of data we asked for 
  */

  OSSpinLockLock(&(My->lock_sampledelivery));

  if (My->SetRenderCallback[0].inputProc)
    {
      if (error_code = ((* My->SetRenderCallback[0].inputProc)
			(My->SetRenderCallback[0].inputProcRefCon, &action_flags, 
			 &AudioTimeStampCarrier, (AudioUnitElement) 0, inNumberFrames,
			 My->AudioBufferCarrier)))
	{
	  sprintf(message, "\nMyRenderer"
		  "\n\tError: Render callback returned OSType %u\n", 
		  (unsigned int) error_code);
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	  return_value = kAudioUnitErr_NoConnection;
	}
    }
  else
    {
      if ((My->MakeConnection[0].sourceAudioUnit == 0) || 
	  (error_code = AudioUnitRender(My->MakeConnection[0].sourceAudioUnit, 
					&action_flags, &AudioTimeStampCarrier, 
					My->MakeConnection[0].sourceOutputNumber,
					inNumberFrames, My->AudioBufferCarrier)))
	{
	  if (My->MakeConnection[0].sourceAudioUnit == 0)
	    {
	      sprintf(message, "\nMyRenderer"
		      "\n\tError: No connection to use with AudioUnitRender\n");
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	      return_value = kAudioUnitErr_NoConnection;
	    }
	  else
	    {
	      sprintf(message, "\nMyRenderer\n\tError: AudioUnitRender returned:\n\t"); 
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	      ASYS_AUDIOUNIT_WIRETAP_PRINT_COMPONENT_RESULT(error_code);
	      sprintf(message, "\n"); 
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	      return_value = error_code;
	    }
	}
    }

  OSSpinLockUnlock(&(My->lock_sampledelivery));

  /* supply return buffers (only!) if needed */

  for (i = 0; i < ioData->mNumberBuffers; i++)  /* straight-wire, no gain */
    if (ioData->mBuffers[i].mData == NULL)
      ioData->mBuffers[i].mData = (void *) My->mData_Output[i];

  /* do PreRender notification */

  OSSpinLockLock(&(My->lock_rendernotify));

  notify = My->rendernotify;
  while (notify)
    {
      if (notify->nproc)
	{
	  action_flags = kAudioUnitRenderAction_PreRender;

	  if (error_code = ((* notify->nproc)
			    (notify->nrefcon, &action_flags, 
			     inTimeStamp, (AudioUnitElement) 0, inNumberFrames,
			     ioData)))
	    {
	      sprintf(message, "\nasysn_audiounit_MyRenderer"
		      "\n\tError: PreRender Rendernotify callback" 
		      " returned OSType %u\n", (unsigned int) error_code);
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	    }
	}
      notify = notify->next;
    }

  OSSpinLockUnlock(&(My->lock_rendernotify));

  /* the actual rendering done by the audio unit */

  if (return_value == noErr) 
    {
      if (iresponse == ASYS_AUDIOUNIT_IRESPONSE_RENDER)
	{
	  My->ksync_timespent = AudioGetCurrentHostTime() - My->ksync_timespent;
	  asysn_audiounit_activeio_renderer(My, ioData);
	  My->ksync_timespent = AudioGetCurrentHostTime() - My->ksync_timespent;
	}
      else
	asysn_audiounit_activeio_passthru(My, ioData);
    }
  else
    for (i = 0; i < ioData->mNumberBuffers; i++)  /* in case of error, silence */
      memset(ioData->mBuffers[i].mData, 0, ioData->mBuffers[i].mDataByteSize);

  /* do PostRender notification */

  OSSpinLockLock(&(My->lock_rendernotify));

  notify = My->rendernotify;
  while (notify)
    {
      if (notify->nproc)
	{
	  action_flags = kAudioUnitRenderAction_PostRender;

	  if (error_code = ((* notify->nproc)
			    (notify->nrefcon, &action_flags, 
			     inTimeStamp, (AudioUnitElement) 0, inNumberFrames,
			     ioData)))
	    {
	      sprintf(message, "\nasysn_audiounit_MyRenderer"
		      "\n\tError: PostRender Rendernotify callback" 
		      " returned OSType %u\n", (unsigned int) error_code);
	      ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);
	    }
	}
      notify = notify->next;
    }

  OSSpinLockUnlock(&(My->lock_rendernotify));

  asysn_audiounit_realtime_exitstate(My);

  return return_value;
}

#endif /* ASYS_AUDIOUNIT_EFFECT */

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Helper Functions for all Selector Calls    */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/****************************************************************/
/* Helper Function for Factory Presets: returns a preset name   */
/****************************************************************/

CFStringRef asysn_audiounit_factorypresetname(CFIndex idx) 

{
  CFStringRef ret = NULL;
  char * nptr, * val_name;
  int i, j;

  j = 0;

  for (i = 0; i < CSYS_GLOBALNUM; i++)
    if ((csys_global[i].type == CSYS_TABLE) && csys_global[i].name &&
	!strncmp("aup_factory_", csys_global[i].name, strlen("aup_factory_")) &&
	strlen(csys_global[i].name) > strlen("aup_factory_") && ((j++) == idx))
      {
	nptr = val_name = malloc
	  (strlen(&(csys_global[i].name[strlen("aup_factory_")])) + 1);

	strcpy(val_name, &(csys_global[i].name[strlen("aup_factory_")]));

	do {
	  if (*nptr == '_')
	    *nptr = ' ';
	} while (*(++nptr));

	ret = CFStringCreateWithCString(NULL, val_name, kCFStringEncodingASCII);
	break;
      }
      
  return ret;
}

/****************************************************************/
/* Helper Function for Factory Presets: returns a preset number */
/****************************************************************/

int asysn_audiounit_factorypresetnumber(CFStringRef cfstr) 

{
  char * nptr, * val_name, * match;
  int i, j;

  if (!cfstr)
    return -1;

  match = calloc(CFStringGetLength(cfstr) + 1, 1);
  CFStringGetCString(cfstr, match, CFStringGetLength(cfstr) + 1, 
		     kCFStringEncodingASCII);

  j = 0;

  for (i = 0; i < CSYS_GLOBALNUM; i++)
    if ((csys_global[i].type == CSYS_TABLE) && csys_global[i].name &&
	!strncmp("aup_factory_", csys_global[i].name, strlen("aup_factory_")) &&
	strlen(csys_global[i].name) > strlen("aup_factory_"))
      {
	nptr = val_name = malloc
	  (strlen(&(csys_global[i].name[strlen("aup_factory_")])) + 1);

	strcpy(val_name, &(csys_global[i].name[strlen("aup_factory_")]));

	do {
	  if (*nptr == '_')
	    *nptr = ' ';
	} while (*(++nptr));

	if (!strcmp(match, val_name))
	  return j;

	j++;
      }
      
  return -1;
}


/***********************************************************************/
/* Helper Function for GetProperty: Create ParameterValueStrings array */
/***********************************************************************/

CFMutableArrayRef asysn_audiounit_parametervaluestrings_create
(asysn_audiounit_InstanceState * My, AudioUnitParameterID i) 

{
  CFMutableArrayRef array;
  int j;

  if ((i < 0) || (i >= My->num_saolparams) || (My->pvs_size[i] == 0))
    return NULL;

  array = CFArrayCreateMutable(NULL, My->pvs_size[i], &kCFTypeArrayCallBacks);

  for (j = 0; j < My->pvs_size[i]; j++)
    CFArrayInsertValueAtIndex(array, j, CFStringCreateWithCString
			      (NULL, My->pvs_cstr[i][j], kCFStringEncodingASCII));

  return array;
}

/*****************************************************************/
/* Pre-engine-execution passes for initialize_parametersystem()  */
/*****************************************************************/

ComponentResult asysn_audiounit_earlypass_parametersystem
                        (asysn_audiounit_InstanceState * My) 

{
  int i, j, msize;
  char * nptr;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Pass 1 on csys_global[]: count parameters and factory presets  */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  for (i = 0; i < CSYS_GLOBALNUM; i++)
    if (csys_global[i].name &&
	(!strncmp(csys_global[i].name, "aup_", 4)) &&
	(strlen(csys_global[i].name) > 4))
      switch (csys_global[i].type) {
      case CSYS_IRATE:
      case CSYS_KRATE:
	if (csys_global[i].width != 1)
	  break;

	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	/* change sfront/src/writepre.c line 3206 */
	/* when the code below changes            */
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	/* skip over reserved names */

 	if ((nptr = strstr(&(csys_global[i].name[4]), "_idx")) && 
	    (strlen(nptr) > 6) && (nptr[4] >= '0') && (nptr[4] <= '9'))
	  break;

	if (strstr(&(csys_global[i].name[4]), "_unit_"))
	  break;

	if ((strstr(&(csys_global[i].name[4]), "_slider_squarelaw")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_cubic")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_squareroot")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_cuberoot")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_log")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_exp")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_linear")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider")) ||
	    (strstr(&(csys_global[i].name[4]), "_checkbox")) ||
	    (strstr(&(csys_global[i].name[4]), "_menu")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_number")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_checkbox")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_menu")))
	  break;

	My->num_saolparams++;
	break;
      case CSYS_TABLE:
	if (!strncmp("aup_factory_", csys_global[i].name, strlen("aup_factory_"))
	    && (strlen(csys_global[i].name) > strlen("aup_factory_")))
	  My->num_factory++;
	break;
      default:
	break;
      }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Create data structures for parameters, factory presets, etc  */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (My->num_saolparams == 0)
    return noErr;

  if (!(My->parameterlist = (AudioUnitParameterID *) 
	malloc(My->num_saolparams*sizeof(AudioUnitParameterID))))
    return kAudioUnitErr_FailedInitialization;

  for (i = 0; i < My->num_saolparams; i++)
    My->parameterlist[i] = i;

  if (!(My->parameterinfo = (AudioUnitParameterInfo *)
	calloc(My->num_saolparams, sizeof(AudioUnitParameterInfo))))
    return kAudioUnitErr_FailedInitialization;

  if (!(My->pvs_size = (int *) calloc(My->num_saolparams, sizeof(int))))
    return kAudioUnitErr_FailedInitialization;

  if (!(My->pvs_cstr = (char ***) calloc(My->num_saolparams, sizeof(char **))))
    return kAudioUnitErr_FailedInitialization;

  msize = My->num_saolparams*sizeof(asysn_audiounit_saolparam);
  if (!(My->saolparam = (asysn_audiounit_saolparam *) malloc(msize)))
    return kAudioUnitErr_FailedInitialization;
  asysn_audiounit_memstatus(My->saolparam, msize, MADV_WILLNEED);

  if (My->num_factory)
    {
      if (!(My->factorypreset_info = (AUPreset *) 
	    calloc(My->num_factory, sizeof(AUPreset))))
	return kAudioUnitErr_FailedInitialization;

      msize = My->num_factory*sizeof(Float32 *);
      if (!(My->factorypreset_values = (Float32 **) calloc(1, msize)))
	return kAudioUnitErr_FailedInitialization;
      asysn_audiounit_memstatus(My->factorypreset_values, msize, MADV_WILLNEED);

      for (i = 0; i < My->num_factory; i++)
	{
	  msize = My->num_saolparams*sizeof(Float32);
	  if (!(My->factorypreset_values[i] = (Float32 *) malloc(msize)))
	    return kAudioUnitErr_FailedInitialization;
	  asysn_audiounit_memstatus(My->factorypreset_values[i], msize, MADV_WILLNEED);
	}
    }

  /* todo: add parameter value strings data structures */

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Pass 2 on csys_global[]: Initialize parameter names  */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  j = 0;

  for (i = 0; i < CSYS_GLOBALNUM; i++)
    if (csys_global[i].name &&
	(!strncmp(csys_global[i].name, "aup_", 4)) &&
	(strlen(csys_global[i].name) > 4))
      switch (csys_global[i].type) {
      case CSYS_IRATE:
      case CSYS_KRATE:
	if (csys_global[i].width != 1)
	  break;

	/* skip over reserved names */

	if ((nptr = strstr(&(csys_global[i].name[4]), "_idx")) && 
	    (strlen(nptr) > 6) && (nptr[4] >= '0') && (nptr[4] <= '9'))
	  break;

	if (strstr(&(csys_global[i].name[4]), "_unit_"))
	  break;

	if ((strstr(&(csys_global[i].name[4]), "_slider_squarelaw")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_cubic")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_squareroot")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_cuberoot")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_log")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_exp")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider_linear")) ||
	    (strstr(&(csys_global[i].name[4]), "_slider")) ||
	    (strstr(&(csys_global[i].name[4]), "_checkbox")) ||
	    (strstr(&(csys_global[i].name[4]), "_menu")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_number")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_checkbox")) ||
	    (strstr(&(csys_global[i].name[4]), "_display_menu")))
	  break;

	/* Do all initialization that may be done early */

	strncpy(My->parameterinfo[j].name, &(csys_global[i].name[4]), 60);
	My->parameterinfo[j].name[59] = '\0';

	My->parameterinfo[j].unit = ASYS_AUDIOUNIT_PARAMETER_DEFAULT_UNIT;
	My->parameterinfo[j].minValue = ASYS_AUDIOUNIT_PARAMETER_DEFAULT_MINIMUM;
	My->parameterinfo[j].maxValue = ASYS_AUDIOUNIT_PARAMETER_DEFAULT_MAXIMUM;
	My->parameterinfo[j].defaultValue = ASYS_AUDIOUNIT_PARAMETER_DEFAULT_DEFAULT;

	My->parameterinfo[j].flags = kAudioUnitParameterFlag_HasCFNameString |
	  kAudioUnitParameterFlag_IsReadable | kAudioUnitParameterFlag_IsWritable |
	  kAudioUnitParameterFlag_CFNameRelease;

	My->saolparam[j].index = csys_global[i].index;
	My->saolparam[j].use = csys_global[i].use;

	j++;
	break;
      default:
	break;
      }

  return noErr;
}


/******************************************************************/
/* Parameter-value strings pass for initialize_parametersystem()  */
/******************************************************************/

ComponentResult asysn_audiounit_parametervaluestrings_parametersystem
                (asysn_audiounit_InstanceState * My) 

{
  int i, j, k, str_idx, len;
  char * nptr, * match;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Create parameter-value strings for Indexed parameters */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  for (j = 0; j < My->num_saolparams; j++)
    if (My->parameterinfo[j].unit == kAudioUnitParameterUnit_Indexed)
      {
	My->pvs_size[j] = (My->parameterinfo[j].maxValue 
			   - My->parameterinfo[j].minValue + 1);
	
	My->pvs_cstr[j] = (char **) calloc(My->pvs_size[j], sizeof(char *));

	if (!(My->pvs_cstr[j]))
	  return kAudioUnitErr_FailedInitialization;

	/* create array and default values */

	for (k = 0; k < My->pvs_size[j]; k++)
	  {
	    My->pvs_cstr[j][k] = (char *)calloc(strlen("4294967295") + 1, sizeof(char));

	    if (!(My->pvs_cstr[j][k]))
	      return kAudioUnitErr_FailedInitialization;

	    sprintf(My->pvs_cstr[j][k], "%u", k);
	  }

	/* overwrite default values with custom strings */

	len = strlen("aup_") + strlen(My->parameterinfo[j].name) + strlen("_idx");

	if (!(match = malloc(len + 1)))
	  return kAudioUnitErr_FailedInitialization;

	sprintf((nptr = match), "aup_%s_idx", My->parameterinfo[j].name); 

	do {
	  if (*nptr == ' ')
	    *nptr = '_';
	} while (*(++nptr));

	for (i = 0; i < CSYS_GLOBALNUM; i++)
	  if (csys_global[i].name && (csys_global[j].width == 1) &&
	      ((csys_global[i].type == CSYS_IRATE) || 
	       (csys_global[i].type == CSYS_KRATE)) &&
	      (!strncmp(csys_global[i].name, match, len)) && 
	      (strlen(csys_global[i].name) > len) &&
	      (sscanf(&(csys_global[i].name[len]), "%u_", &k) == 1) && 
	      (k >= 0) && (k <= My->pvs_size[j]))
	    {
	      str_idx = len;
	      while (csys_global[i].name[str_idx++] != '_');

	      if (csys_global[i].name[str_idx] != '\0')
		{
		  free(My->pvs_cstr[j][k]);

		  nptr = My->pvs_cstr[j][k] 
		    = malloc(strlen(&(csys_global[i].name[str_idx])) + 1);

		  if (!nptr)
		    return kAudioUnitErr_FailedInitialization;

		  strcpy(My->pvs_cstr[j][k], &(csys_global[i].name[str_idx]));

		  do {
		    if (*nptr == '_')
		      *nptr = ' ';
		  } while (*(++nptr));
		}
	    }

	free(match);
      }

  return noErr;
}

/*********************************************************/
/* Factory preset pass for initialize_parametersystem()  */
/*********************************************************/

ComponentResult asysn_audiounit_factorypreset_parametersystem
                (ENGINE_PTR_DECLARE_COMMA 
                 asysn_audiounit_InstanceState * My) 

{
  int i, j, k, idx;
  char * nptr, * match;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* initialize factory preset data structures */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (My->num_factory)
    {
      j = 0;
      My->factorypreset_array = CFArrayCreateMutable(NULL, My->num_factory, NULL);

      if (!(My->factorypreset_array))
	return kAudioUnitErr_FailedInitialization;

      for (i = 0; i < CSYS_GLOBALNUM; i++)
	if ((csys_global[i].type == CSYS_TABLE) && csys_global[i].name &&
	    !strncmp("aup_factory_", csys_global[i].name, strlen("aup_factory_")) &&
	    strlen(csys_global[i].name) > strlen("aup_factory_"))
	  {
	    nptr = match = malloc
	      (strlen(&(csys_global[i].name[strlen("aup_factory_")])) + 1);

	    if (!nptr)
	      return kAudioUnitErr_FailedInitialization;

	    strcpy(match, &(csys_global[i].name[strlen("aup_factory_")]));
	    do {
	      if (*nptr == '_')
		*nptr = ' ';
	    } while (*(++nptr));

	    My->factorypreset_info[j].presetNumber = j;
	    My->factorypreset_info[j].presetName = CFStringCreateWithCString
	      (NULL, match, kCFStringEncodingASCII);

	    if (!(My->factorypreset_info[j].presetName))
	      return kAudioUnitErr_FailedInitialization;

	    CFArrayAppendValue(My->factorypreset_array, &(My->factorypreset_info[j]));

	    free(match);

	    for (k = 0; k < My->num_saolparams; k++)
	      My->factorypreset_values[j][k] = My->parameterinfo[k].defaultValue;

	    if (EV(gtables)[idx = csys_global[i].index].t)
	      for (k = 0; (k < EV(gtables)[idx].len) && (k < My->num_saolparams); k++)
		{
		  My->factorypreset_values[j][k] = EV(gtables)[idx].t[k];
		  My->factorypreset_values[j][k]  = 
		    (My->factorypreset_values[j][k] < My->parameterinfo[k].minValue)
		    ? My->parameterinfo[k].minValue : My->factorypreset_values[j][k];
		  My->factorypreset_values[j][k]  = 
		    (My->factorypreset_values[j][k] > My->parameterinfo[k].maxValue)
		    ? My->parameterinfo[k].maxValue : My->factorypreset_values[j][k];
		}

	    j++;
	  }
    }

  return noErr;
}

/************************************************************/
/* Helper for Open Selector: Initializes parameter system.  */
/************************************************************/

ComponentResult asysn_audiounit_initialize_parametersystem
                   (asysn_audiounit_InstanceState * My) 

{
  int i, j, k, ret, match_len, unit_num;
  char * nptr, * kptr, * match;
  ENGINE_PTR_DECLARE_SEMICOLON

  if ((ASYS_AUDIOUNIT_HAS_AUCONTROL == 0) || (CSYS_GLOBALNUM == 0))
    return noErr;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Parameter initialization passes that precede engine start */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if ((ret = asysn_audiounit_earlypass_parametersystem(My)) != noErr)
    return ret;

  if (My->num_saolparams == 0)
    return noErr;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Execute engine to define parameters and factory presets  */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  ENGINE_PTR_ASSIGNED_TO  system_init(My->argc, My->argv, 
				      My->OutputStreamFormat.mSampleRate);

  if (ENGINE_PTR_IS_NULL)
    return kAudioUnitErr_FailedInitialization;

  effects_init(ENGINE_PTR);
  main_initpass(ENGINE_PTR);
  asysn_audiounit_ksyncinit(My);

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Use engine results to complete parameter data structures */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  for (j = 0; j < My->num_saolparams; j++)
    {
      match = malloc(strlen("aup_") + strlen(My->parameterinfo[j].name) + 
		     strlen("_pinfo") + 1);

      if (!match)
	{
	  shut_down(ENGINE_PTR);
	  return kAudioUnitErr_FailedInitialization;
	}

      sprintf(match, "aup_%s_pinfo", My->parameterinfo[j].name); 

      for (i = 0; i < CSYS_GLOBALNUM; i++)
	if ((csys_global[i].type == CSYS_TABLE) && csys_global[i].name &&
	    !strcmp(match, csys_global[i].name))
	  {
	    k = csys_global[i].index;

	    if (EV(gtables)[k].t && (EV(gtables)[k].len >= 3) && 
		(EV(gtables)[k].t[0] <= EV(gtables)[k].t[1]) && 
		(EV(gtables)[k].t[1] <= EV(gtables)[k].t[2]))
	      {
		My->parameterinfo[j].minValue = EV(gtables)[k].t[0];
		My->parameterinfo[j].defaultValue = EV(gtables)[k].t[1];
		My->parameterinfo[j].maxValue = EV(gtables)[k].t[2];

		if (My->parameterinfo[j].minValue > My->parameterinfo[j].maxValue)
		  {
		    My->parameterinfo[j].minValue = EV(gtables)[k].t[2];
		    My->parameterinfo[j].maxValue = EV(gtables)[k].t[0];
		  }

		if (My->parameterinfo[j].defaultValue > My->parameterinfo[j].maxValue)
		  My->parameterinfo[j].defaultValue = EV(gtables)[k].t[2];

		if (My->parameterinfo[j].defaultValue < My->parameterinfo[j].minValue)
		  My->parameterinfo[j].defaultValue = EV(gtables)[k].t[0];
	      }
	    break;
	  }

      free(match);

      match = malloc((match_len = strlen("aup_") + strlen(My->parameterinfo[j].name) + 
		      strlen("_unit_")) + 1);

      if (!match)
	{
	  shut_down(ENGINE_PTR);
	  return kAudioUnitErr_FailedInitialization;
	}

      sprintf(match, "aup_%s_unit_", My->parameterinfo[j].name); 

      for (i = 0; i < CSYS_GLOBALNUM; i++)
	if (((csys_global[i].type == CSYS_IRATE) || 
	     (csys_global[i].type == CSYS_KRATE)) && 
	    (csys_global[i].width == 1) &&
	    csys_global[i].name &&
	    !strncmp(match, csys_global[i].name, match_len))
	  {
	    if (strlen(csys_global[i].name) <= match_len)
	      continue;

	    nptr = &(csys_global[i].name[match_len]);

	    /* for now, only support custom units */

	    kptr = malloc(strlen(nptr) + 1);

	    if (!kptr)
	      {
		shut_down(ENGINE_PTR);
		return kAudioUnitErr_FailedInitialization;
	      }

	    strcpy(kptr, &(csys_global[i].name[match_len]));

	    nptr = kptr;
	    do {
	      if (*nptr == '_')
		*nptr = ' ';
	    } while (*(++nptr));

	    My->parameterinfo[j].unitName = CFStringCreateWithCString
	      (NULL, kptr, kCFStringEncodingASCII);

	    if (!(My->parameterinfo[j].unitName))
	      {
		shut_down(ENGINE_PTR);
		return kAudioUnitErr_FailedInitialization;
	      }

	    My->parameterinfo[j].unit = kAudioUnitParameterUnit_CustomUnit;

	    free(kptr);
	    break;
	  }

      free(match);

      match = malloc((match_len = strlen("aup_") + strlen(My->parameterinfo[j].name) + 
		      strlen("_")) + 1);

      if (!match)
	{
	  shut_down(ENGINE_PTR);
	  return kAudioUnitErr_FailedInitialization;
	}

      sprintf(match, "aup_%s_", My->parameterinfo[j].name); 

      for (i = 0; i < CSYS_GLOBALNUM; i++)

	if (((csys_global[i].type == CSYS_IRATE) || 
	     (csys_global[i].type == CSYS_KRATE)) && 
	    (csys_global[i].width == 1) &&
	    csys_global[i].name &&
	    !strncmp(match, csys_global[i].name, match_len))
	  {
	    if (strlen(csys_global[i].name) <= match_len)
	      continue;

	    nptr = &(csys_global[i].name[match_len]);

	    if (!strcmp(nptr, "slider_squarelaw"))
	      My->parameterinfo[j].flags |= 
		kAudioUnitParameterFlag_DisplaySquared;
	    else
	      if (!strcmp(nptr, "slider_cubic"))
		My->parameterinfo[j].flags |= 
		  kAudioUnitParameterFlag_DisplayCubed;
	      else
		if (!strcmp(nptr, "slider_squareroot"))
		  My->parameterinfo[j].flags |= 
		    kAudioUnitParameterFlag_DisplaySquareRoot;
		else
		  if (!strcmp(nptr, "slider_cuberoot"))
		    My->parameterinfo[j].flags |= 
		      kAudioUnitParameterFlag_DisplayCubeRoot;
		  else
		    if (!strcmp(nptr, "slider_log"))
		      My->parameterinfo[j].flags |= 
			kAudioUnitParameterFlag_DisplayLogarithmic;
		    else
		      if (!strcmp(nptr, "slider_exp"))
			My->parameterinfo[j].flags |= 
			  kAudioUnitParameterFlag_DisplayExponential;
		      else
			if (!strcmp(nptr, "menu") || !strcmp(nptr, "display_menu"))
			  {
			    if (!strcmp(nptr, "display_menu"))
			      My->parameterinfo[j].flags &= 
				~ kAudioUnitParameterFlag_IsWritable;

			    My->parameterinfo[j].unit = 
			      kAudioUnitParameterUnit_Indexed;
			    if (My->parameterinfo[j].unitName)
			      {
				CFRelease(My->parameterinfo[j].unitName);
				My->parameterinfo[j].unitName = NULL;
			      }
			  }
			else
			  if (!strcmp(nptr, "checkbox") || 
			      !strcmp(nptr, "display_checkbox"))
			    {
			      if (!strcmp(nptr, "display_checkbox"))
				My->parameterinfo[j].flags &= 
				  ~ kAudioUnitParameterFlag_IsWritable;

			      My->parameterinfo[j].unit = 
				kAudioUnitParameterUnit_Boolean;
			      if (My->parameterinfo[j].unitName)
				{
				  CFRelease(My->parameterinfo[j].unitName);
				  My->parameterinfo[j].unitName = NULL;
				}
			    }
			  else
			    if (!strcmp(nptr, "display_number"))
			      {
				My->parameterinfo[j].flags &= 
				  ~ kAudioUnitParameterFlag_IsWritable;
			      }
			    else
			      if ((strcmp(nptr, "slider_linear") != 0) &&
				  (strcmp(nptr, "slider") != 0))
				continue;
	    break;
	  }

      free(match);
      
      My->saolparam[j].value = My->parameterinfo[j].defaultValue;

      nptr = &(My->parameterinfo[j].name[0]);
      do {
	if (*nptr == '_')
	  *nptr = ' ';
      } while (*(++nptr));
      
      My->parameterinfo[j].cfNameString = CFStringCreateWithCString
	(NULL, My->parameterinfo[j].name, kCFStringEncodingASCII);

      if (!(My->parameterinfo[j].cfNameString))
	{
	  shut_down(ENGINE_PTR);
	  return kAudioUnitErr_FailedInitialization;
	}
    }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* Create parameter-value strings for Indexed parameters */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if ((ret = asysn_audiounit_parametervaluestrings_parametersystem(My)) != noErr)
    {
      shut_down(ENGINE_PTR);
      return ret;
    }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* initialize factory preset data structures */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if ((ret = asysn_audiounit_factorypreset_parametersystem
       (ENGINE_PTR_COMMA My)) != noErr)
    {
      shut_down(ENGINE_PTR);
      return ret;
    }

  /*~~~~~~~~~~~~~~~~~~~*/
  /* Shut down engine  */
  /*~~~~~~~~~~~~~~~~~~~*/

  shut_down(ENGINE_PTR);
  return noErr;
}


/************************************************************/
/* Helper for Open Selector: Initializes property variables */
/************************************************************/

ComponentResult asysn_audiounit_initialize_properties
(asysn_audiounit_InstanceState * My) 

{
  int i;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* the first set of parameters define the signal flow  */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  /* the preferred audiounit bus widths */

  My->InputElementCount = ASYS_AUDIOUNIT_ELEMENT_INPUTPREF;
  My->OutputElementCount = ASYS_AUDIOUNIT_ELEMENT_OUTPUTPREF;

  /* defines fixed channel widths. todo: turn into an array of formats */

  if (ASYS_AUDIOUNIT_EFFECT)
    {
      My->SupportedNumChannels[0].inChannels = ASYS_AUDIOUNIT_MONO;
      My->SupportedNumChannels[0].outChannels = ASYS_AUDIOUNIT_MONO;
      My->SupportedNumChannels[1].inChannels = ASYS_AUDIOUNIT_MONO;
      My->SupportedNumChannels[1].outChannels = ASYS_AUDIOUNIT_STEREO;
      My->SupportedNumChannels[2].inChannels = ASYS_AUDIOUNIT_STEREO;
      My->SupportedNumChannels[2].outChannels = ASYS_AUDIOUNIT_MONO;
      My->SupportedNumChannels[3].inChannels = ASYS_AUDIOUNIT_STEREO;
      My->SupportedNumChannels[3].outChannels = ASYS_AUDIOUNIT_STEREO;
    }
  else
    {
      My->SupportedNumChannels[0].inChannels = ASYS_AUDIOUNIT_NULL;
      My->SupportedNumChannels[0].outChannels = ASYS_AUDIOUNIT_MONO;
      My->SupportedNumChannels[1].inChannels = ASYS_AUDIOUNIT_NULL;
      My->SupportedNumChannels[1].outChannels = ASYS_AUDIOUNIT_STEREO;
    }


  /* maximum audio sample chunk -- augment with other variables to resize buffers */
  
  My->MaximumFramesPerSlice = ASYS_AUDIOUNIT_FRAMES_PER_SLICE;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* our desired input and output stream format */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  for (i=0; i < ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE; i++)
    {
      /* mChannelsPerFrame code mono, stereo, etc */

      My->InputStreamFormat[i].mChannelsPerFrame = ASYS_AUDIOUNIT_INPUT_CHANNELS;

      /* could also be kAudioStreamAnyRate, once we support setproperty over-ride */

      My->InputStreamFormat[i].mSampleRate = SAOL_SRATE;
  
      /* a set of non-interleaved mono channels -- do not touch these */

      My->InputStreamFormat[i].mFormatID = kAudioFormatLinearPCM;
      
      My->InputStreamFormat[i].mFormatFlags = (kAudioFormatFlagIsFloat | 
					    kAudioFormatFlagIsPacked | 
					    kAudioFormatFlagIsNonInterleaved);
      
      if (ASYS_AUDIOUNIT_FLOAT32_BIGENDIAN)
	My->InputStreamFormat[i].mFormatFlags |= kAudioFormatFlagIsBigEndian;

      My->InputStreamFormat[i].mBytesPerPacket = 4;
      My->InputStreamFormat[i].mBytesPerFrame = 4;
      My->InputStreamFormat[i].mFramesPerPacket = 1;
      My->InputStreamFormat[i].mBitsPerChannel = 32;
      
      My->InputStreamFormat[i].mReserved = 0;
    }

  /* mChannelsPerFrame code mono, stereo, etc */

  My->OutputStreamFormat.mChannelsPerFrame = ASYS_AUDIOUNIT_OUTPUT_CHANNELS;

  /* could also be kAudioStreamAnyRate, once we support setproperty over-ride */

  My->OutputStreamFormat.mSampleRate = SAOL_SRATE;  

  My->acycle = ((int)(SAOL_SRATE))/((int)(SAOL_KRATE));
  My->krate = SAOL_KRATE;
  My->acycleidx_kcycleidx = 1;     /* acycleidx = 0, kcycleidx = 1 */

  /* a set of non-interleaved mono channels -- do not touch these */

  My->OutputStreamFormat.mFormatID = kAudioFormatLinearPCM;
  
  My->OutputStreamFormat.mFormatFlags = (kAudioFormatFlagIsFloat | 
					 kAudioFormatFlagIsPacked | 
					 kAudioFormatFlagIsNonInterleaved);

  if (ASYS_AUDIOUNIT_FLOAT32_BIGENDIAN)
    My->OutputStreamFormat.mFormatFlags |= kAudioFormatFlagIsBigEndian;

  My->OutputStreamFormat.mBytesPerPacket = 4;
  My->OutputStreamFormat.mBytesPerFrame = 4;
  My->OutputStreamFormat.mFramesPerPacket = 1;
  My->OutputStreamFormat.mBitsPerChannel = 32;

  My->OutputStreamFormat.mReserved = 0;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* end of streamformat property */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  /* tracks if the audio unit is currently bypassed */

  My->BypassEffect = 0;
  My->LastBypassEffect = 0;

  /* set to 1 if we renderer returns the buffers use for SetRenderCallback */

  My->InPlaceProcessing = ASYS_AUDIOUNIT_INPLACE_PROCESSING;

  /* define the null preset */

  My->PresentPreset.presetName = CFStringCreateWithCString(NULL, "Untitled", 
							   kCFStringEncodingASCII);
  if (!(My->PresentPreset.presetName))
    return kAudioUnitErr_FailedInitialization;

  My->PresentPreset.presetNumber = -1;

  My->lock_PresentPreset = 0;

  /* number of instruments offered by the music device */

  if (ASYS_AUDIOUNIT_MUSICDEVICE)
    My->InstrumentCount = ASYS_AUDIOUNIT_INSTRUMENT_COUNT;
  else
    My->InstrumentCount = 0;

  /* 1 if the audio unit intends to stream samples off of the disk */

  My->StreamFromDisk = 0;

  /* the function pointer and parameters SetRenderCallback provides */

  for (i=0; i < ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE; i++)
    {
      My->SetRenderCallback[i].inputProc = NULL;
      My->SetRenderCallback[i].inputProcRefCon = NULL;
    }

  /* how the host tells us who is driving us */

  for (i=0; i < ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE; i++)
    {
      My->MakeConnection[i].sourceAudioUnit = 0;
      My->MakeConnection[i].sourceOutputNumber = 0;
      My->MakeConnection[i].destInputNumber = 0;
    }

  /* lock for both SetRenderCallback and MakeConnection */

  My->lock_sampledelivery = 0;

 /* the function pointer and parameters RenderNotify provides, and its lock */
 
  My->rendernotify = NULL;
  My->lock_rendernotify = 0;

  /* tracks last rendering error for the rendering proc - clear after returning it */

  My->LastRenderError = noErr;

  /* for now, assume AU algorithm has no latency */

  My->Latency = 0.0;

  /* tail time (in a reverb sense) for the audio unit */

  My->TailTime = 0.0;

  /* to be used for selecting sfront's sample-rate conversion algorithm */

  My->SRCAlgorithm = kAudioUnitSRCAlgorithm_Polyphase;

  /* request for render quality, using 0-127 scaled enum in AudioUnitProperties.h */

  My->RenderQuality = kRenderQuality_Max;

  /* the % of total CPU load the host allots to us, 0 codes no limit */
  
  My->CPULoad = 0.0;
  My->lock_CPULoad = 0;
  My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());

  /* a heads-up for off-line operation, sent by host */

  My->OfflineRender = 0;

  /* a heads-up for the presentation latency, sent by host */

  My->PresentationLatency = 0.0;

  /* listener callback lists for each supported property */

  for (i = 0; i < ASYS_AUDIOUNIT_PROPLISTEN_ARRAYSIZE; i++)
    My->proplisteners[i] = NULL;

  My->lock_proplisteners = 0;

  /*~~~~~~~~~~~~~~~~~~~~~*/
  /* aup_ property state */
  /*~~~~~~~~~~~~~~~~~~~~~*/

  My->num_saolparams = 0;
  My->parameterlist = NULL;
  My->parameterinfo = NULL;
  My->pvs_size = NULL;
  My->pvs_cstr = NULL;
  My->saolparam = NULL;
  My->num_factory = 0;
  My->factorypreset_values = NULL;
  My->factorypreset_info = NULL;
  My->factorypreset_array = NULL;

  return noErr;
}

/************************************************************/
/* Helper for Open Selector: Memstatus for global variables */
/************************************************************/

void asysn_audiounit_globalvars_memstatus(int advice)

{
  int msize, i, j;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* global arrays used by main_ipass */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#if (defined(CSYS_MAXEXTCHAN) && (CSYS_MAXEXTCHAN > 0))
  msize = CSYS_MAXEXTCHAN*sizeof(int);
  asysn_audiounit_memstatus(cme_preset, msize, advice);
#endif

#if (defined(CSYS_MAXSASLINSTR) && (CSYS_MAXSASLINSTR > 0))
  msize = CSYS_MAXSASLINSTR*sizeof(int);
  asysn_audiounit_memstatus(csys_instrtablesize, msize, advice);
  asysn_audiounit_memstatus(csys_instrvarsize, msize, advice);
#endif

  /*~~~~~~~~~~~~~~~~~~~~~*/
  /* global table arrays */
  /*~~~~~~~~~~~~~~~~~~~~~*/

#if (defined(CSYS_TABLE_CATALOG_SIZE) && (CSYS_TABLE_CATALOG_SIZE > 0))
  for (i = 0; i < CSYS_TABLE_CATALOG_SIZE; i++)
    {
      msize = csys_table_catalog[i].num*sizeof(float);
      asysn_audiounit_memstatus(csys_table_catalog[i].t, msize, advice);
    }
#endif 

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* reflection interface variables */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#if (defined(CSYS_GLOBALNUM) && (CSYS_GLOBALNUM > 0))
  msize = CSYS_GLOBALNUM*sizeof(csys_varstruct);
  asysn_audiounit_memstatus(csys_global, msize, advice);
#endif

#if (defined(CSYS_TARGETNUM) && (CSYS_TARGETNUM > 0))
  msize = CSYS_TARGETNUM*sizeof(csys_targetstruct);
  asysn_audiounit_memstatus(csys_target, msize, advice);
#endif

#if (defined(CSYS_LABELNUM) && (CSYS_LABELNUM > 0))
  msize = CSYS_LABELNUM*sizeof(csys_labelstruct);
  asysn_audiounit_memstatus(csys_labels, msize, advice);
#endif

#if (defined(CSYS_PRESETNUM) && (CSYS_PRESETNUM > 0))
  msize = CSYS_PRESETNUM*sizeof(csys_presetstruct);
  asysn_audiounit_memstatus(csys_presets, msize, advice);
#endif

#if (defined(CSYS_SAMPLENUM) && (CSYS_SAMPLENUM > 0))
  msize = CSYS_SAMPLENUM*sizeof(csys_samplestruct);
  asysn_audiounit_memstatus(csys_samples, msize, advice);
#endif

#if (defined(CSYS_BUSNUM) && (CSYS_BUSNUM > 0))
  msize = CSYS_BUSNUM*sizeof(csys_busstruct);
  asysn_audiounit_memstatus(csys_bus, msize, advice);
#endif

#if (defined(CSYS_ROUTENUM) && (CSYS_ROUTENUM > 0))
  msize = CSYS_ROUTENUM*sizeof(csys_routestruct);
  asysn_audiounit_memstatus(csys_route, msize, advice);
#endif

#if (defined(CSYS_SENDNUM) && (CSYS_SENDNUM > 0))
  msize = CSYS_SENDNUM*sizeof(csys_sendstruct);
  asysn_audiounit_memstatus(csys_send, msize, advice);
  for (i = 0; i < CSYS_SENDNUM; i++)
    if (csys_send[i].bus)
      {
	msize = csys_send[i].nbus*sizeof(int);
	asysn_audiounit_memstatus(csys_send[i].bus, msize, advice);
      }
#endif

#if (defined(CSYS_INSTRNUM) && (CSYS_INSTRNUM > 0))
  msize = CSYS_INSTRNUM*sizeof(csys_instrstruct);
  asysn_audiounit_memstatus(csys_instr, msize, advice);
  for (i = 0; i < CSYS_INSTRNUM; i++)
    {
      if (csys_instr[i].name)
	{
	  msize = (strlen(csys_instr[i].name) + 1)*sizeof(char);
	  asysn_audiounit_memstatus(csys_instr[i].name, msize, advice);
	}
      if (csys_instr[i].vars)
	{
	  msize = csys_instr[i].numvars*sizeof(csys_varstruct);
	  asysn_audiounit_memstatus(csys_instr[i].vars, msize, advice);
	  for (j = 0; j < csys_instr[i].numvars; j++)
	    if (csys_instr[i].vars[j].name)
	      {
		msize = (strlen(csys_instr[i].vars[j].name) + 1)*sizeof(char);
		asysn_audiounit_memstatus(csys_instr[i].vars[j].name, msize, advice);
	      }
	}
    }
#endif

} 

/************************************************************/
/* Helper for Open Selector: Recover from a failed open.    */
/************************************************************/

ComponentResult asysn_audiounit_instance_cleanup
(ComponentParameters * p, asysn_audiounit_InstanceState * My)

{
  int i, k, msize;

  if (My)
    {
      if (My->PresentPreset.presetName)
	CFRelease(My->PresentPreset.presetName);

      if (My->AudioBufferCarrier)
	{
	  msize = sizeof(UInt32) + sizeof(AudioBuffer)*ASYS_AUDIOUNIT_INPUT_MAXCHANNELS;
	  asysn_audiounit_memstatus(My->AudioBufferCarrier, msize, MADV_FREE);
	  free(My->AudioBufferCarrier);
	}

      if (My->AudioBufferTemplate)
	{
	  msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;
	  for(i = 0; i < ASYS_AUDIOUNIT_INPUT_MAXCHANNELS; i++)
	    if (My->AudioBufferTemplate->mBuffers[i].mData)
	      {
		asysn_audiounit_memstatus(My->AudioBufferTemplate->mBuffers[i].mData, 
					  msize, MADV_FREE);
		free(My->AudioBufferTemplate->mBuffers[i].mData);
	      }
	  msize = sizeof(UInt32) + sizeof(AudioBuffer)*ASYS_AUDIOUNIT_INPUT_MAXCHANNELS;
	  asysn_audiounit_memstatus(My->AudioBufferTemplate, msize, MADV_FREE);
	  free(My->AudioBufferTemplate);
	}

      for(i = 0; i < ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS; i++)
	if (My->mData_Output[i])
	  {
	    msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;
	    asysn_audiounit_memstatus(My->mData_Output[i], msize, MADV_FREE);
	    free(My->mData_Output[i]);
	  }

      if (My->mpipepair[0])
	close(My->mpipepair[0]);
      if (My->mpipepair[1])
	close(My->mpipepair[1]);
      if (My->spipepair[0])
	close(My->spipepair[0]);
      if (My->spipepair[1])
	close(My->spipepair[1]);

      if (My->parameterlist)
	free(My->parameterlist);

      if (My->parameterinfo)
	{
	  for (i = 0; i < My->num_saolparams; i++)
	    {
	      if (My->parameterinfo[i].cfNameString)
		CFRelease(My->parameterinfo[i].cfNameString);
	      if (My->parameterinfo[i].unitName)
		CFRelease(My->parameterinfo[i].unitName);
	    }
	  free(My->parameterinfo);
	}

      if (My->pvs_size)
	{
	  if (My->pvs_cstr)
	    {
	      for (i = 0; i < My->num_saolparams; i++)
		if (My->pvs_cstr[i])
		  {
		    for (k = 0; k < My->pvs_size[i]; k++)
		      if (My->pvs_cstr[i][k])
			free(My->pvs_cstr[i][k]);
		    free(My->pvs_cstr[i]);
		  }
	      free(My->pvs_cstr);
	    }
	  free(My->pvs_size);
	}

      if (My->saolparam)
	{
	  msize = My->num_saolparams*sizeof(asysn_audiounit_saolparam);
	  asysn_audiounit_memstatus(My->saolparam, msize, MADV_FREE);
	  free(My->saolparam);
	}

      if (My->factorypreset_info)
	{
	  for (i = 0; i < My->num_factory; i++)
	    if (My->factorypreset_info[i].presetName)
	      CFRelease(My->factorypreset_info[i].presetName);
	  free(My->factorypreset_info);
	}

      if (My->factorypreset_values)
	{
	  msize = My->num_saolparams*sizeof(Float32);
	  for (i = 0; i < My->num_factory; i++)
	    if (My->factorypreset_values[i])
	      {
		asysn_audiounit_memstatus(My->factorypreset_values[i], msize, MADV_FREE);
		free(My->factorypreset_values[i]);
	      }

	  msize = My->num_factory*sizeof(Float32 *);
	  asysn_audiounit_memstatus(My->factorypreset_values, msize, MADV_FREE);
	  free(My->factorypreset_values);
	}

      if (My->factorypreset_array)
	CFRelease(My->factorypreset_array);

      msize = sizeof(asysn_audiounit_InstanceState);
      asysn_audiounit_memstatus(My, msize, MADV_FREE);
      free(My);
    }

  SetComponentInstanceStorage((ComponentInstance) p->params[0], (Handle) NULL); 
  return kAudioUnitErr_FailedInitialization;
}

/**********************************************************************/
/* Helper for Initialize Selector: Send all parameters to SAOL engine */
/**********************************************************************/

ComponentResult asysn_audiounit_engine_parameter_update
(asysn_audiounit_InstanceState * My) 

{
  asysn_audiounit_SASLevent SASLevent;
  int result = noErr;
  int i;
  
  for (i = 0; i < My->num_saolparams; i++)
    {
      SASLevent.index = i;
      SASLevent.value = My->saolparam[i].value;
      SASLevent.kcycleidx = 0;
      result |= asysn_audiounit_sendSASLevent(&SASLevent, My);    
    }

  return result;
}

/***************************************************************/
/* Helper for Uninitialize Selector: empty MIDI and SASL pipes */
/***************************************************************/

void asysn_audiounit_engine_emptypipes
(asysn_audiounit_InstanceState * My) 
{
  asysn_audiounit_MIDIevent mdummy;
  asysn_audiounit_SASLevent sdummy;
  int retry = 0;

  if (ASYS_AUDIOUNIT_MIDISUPPORT && My->mpipepair[0])
    do { 
      if (retry++ > ASYS_AUDIOUNIT_RETRY_MAX)
	break;
    } while ((read(My->mpipepair[0], &mdummy, 
		   sizeof(asysn_audiounit_MIDIevent)) > 0) || (errno == EINTR));

  retry == 0;

  if (My->spipepair[0])
    do {
      if (retry++ > ASYS_AUDIOUNIT_RETRY_MAX)
	break;
    } while ((read(My->spipepair[0], &sdummy, 
		   sizeof(asysn_audiounit_SASLevent)) > 0) || (errno == EINTR));
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Helper Functions for the GetProperty Selector Helper for ClassInfo  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/***********************************************/
/* Big-endian int append to a CFMutableDataRef */
/***********************************************/

void asysn_audiounit_classinfo_pdata_intwrite(CFMutableDataRef pdata, SInt32 value)

{
  unsigned char p[4];
  
  p[0] = 0x000000FF & (value >> 24);
  p[1] = 0x000000FF & (value >> 16);
  p[2] = 0x000000FF & (value >> 8);
  p[3] = 0x000000FF & (value);

  CFDataAppendBytes(pdata, (UInt8 *) p, 4);
}

/***************************************************/
/*  Big-endian float append to a CFMutableDataRef  */
/*   Assumes int and float share an enddian-ness   */   
/***************************************************/

void asysn_audiounit_classinfo_pdata_floatwrite(CFMutableDataRef pdata, Float32 fval)

{
  union { int i; float f ; } u;
  unsigned char p[4];

  u.f = fval;
  
  p[0] = 0x000000FF & (u.i >> 24);
  p[1] = 0x000000FF & (u.i >> 16);
  p[2] = 0x000000FF & (u.i >> 8);
  p[3] = 0x000000FF & (u.i);

  CFDataAppendBytes(pdata, (UInt8 *) p, 4);
}

/************************************/
/* Adds integer item to a ClassInfo */
/************************************/

int asysn_audiounit_classinfo_addint(CFMutableDictionaryRef ClassInfo, 
				     char * ckey, SInt32 value)

{
  CFStringRef key;
  CFNumberRef num;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      num = CFNumberCreate(NULL, kCFNumberSInt32Type, &value);
      if (num)
	{
	  CFDictionarySetValue(ClassInfo, key, num);
	  CFRelease(num);
	  errcode = 0;
	}
      CFRelease(key);
    }

  return errcode;
}

/************************************/
/* Adds float item to a ClassInfo */
/************************************/

int asysn_audiounit_classinfo_addfloat(CFMutableDictionaryRef ClassInfo, 
				       char * ckey, Float32 fval)

{
  CFStringRef key;
  CFNumberRef num;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      num = CFNumberCreate(NULL, kCFNumberFloatType, &fval);
      if (num)
	{
	  CFDictionarySetValue(ClassInfo, key, num);
	  CFRelease(num);
	  errcode = 0;
	}
      CFRelease(key);
    }

  return errcode;
}

/*************************************/
/* Adds CFString item to a ClassInfo */
/*************************************/

int asysn_audiounit_classinfo_addcfstr(CFMutableDictionaryRef ClassInfo, 
				       char * ckey, CFStringRef cfstr)

{
  CFStringRef key;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      CFDictionarySetValue(ClassInfo, key, cfstr);
      CFRelease(key);
      errcode = 0;
    }

  return errcode;
}

/************************************************************************/
/* Helper for GetProperty Selector: Create ClassInfo for use by AU host */
/************************************************************************/

CFMutableDictionaryRef asysn_audiounit_classinfo_create
                                           (asysn_audiounit_InstanceState * My) 

{
  CFMutableDictionaryRef ClassInfo;
  CFMutableDataRef pdata;
  CFStringRef key;
  int i;

  ClassInfo = CFDictionaryCreateMutable
    (NULL, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);

  /* component definitions adapted from volume.r */

  asysn_audiounit_classinfo_addint(ClassInfo, kAUPresetVersionKey, 
				   0);  /* version 0 */
  asysn_audiounit_classinfo_addint(ClassInfo, kAUPresetTypeKey, 
				   ASYS_AUDIOUNIT_COMP_TYPE);
  asysn_audiounit_classinfo_addint(ClassInfo, kAUPresetSubtypeKey, 
				   ASYS_AUDIOUNIT_COMP_SUBTYPE);
  asysn_audiounit_classinfo_addint(ClassInfo, kAUPresetManufacturerKey, 
				   ASYS_AUDIOUNIT_COMP_MANU);

  /* parameter data */

  if ((pdata = CFDataCreateMutable(NULL, 0)) && My->num_saolparams)
    {
      asysn_audiounit_classinfo_pdata_intwrite(pdata, 
  			              kAudioUnitScope_Global);     /* global scope */
      asysn_audiounit_classinfo_pdata_intwrite(pdata, 
				      0);                       /* element (bus) 0 */
      asysn_audiounit_classinfo_pdata_intwrite(pdata, 
				      My->num_saolparams); /* number of parameters */

      for (i = 0; i < My->num_saolparams; i++)
	{
	  /* parameter index */

	  asysn_audiounit_classinfo_pdata_intwrite(pdata, i);

	  /* value */

	  asysn_audiounit_classinfo_pdata_floatwrite
	    (pdata, 
	     (My->parameterinfo[i].flags & kAudioUnitParameterFlag_IsWritable) ?
	     My->saolparam[i].value : My->parameterinfo[i].defaultValue);
	}
    }

  if (pdata)
    {
      key = CFStringCreateWithCString(NULL, kAUPresetDataKey, kCFStringEncodingASCII);
      if (key)
	{
	  CFDictionarySetValue(ClassInfo, key, pdata);
	  CFRelease(key);
	}
      CFRelease(pdata);
    }

  asysn_audiounit_classinfo_addcfstr(ClassInfo, kAUPresetNameKey, 
				     My->PresentPreset.presetName);
  asysn_audiounit_classinfo_addint(ClassInfo, kAUPresetRenderQualityKey, 
				   (SInt32) (My->RenderQuality));
  asysn_audiounit_classinfo_addfloat(ClassInfo, kAUPresetCPULoadKey, 
				     My->CPULoad);

  /* add kAUPresetElementNameKey here if named elements/buses are supported */

  return ClassInfo;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Helper Functions for the SetProperty Selector Helper for ClassInfo  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/**************************************/
/* Read integer item from a ClassInfo */
/**************************************/

int asysn_audiounit_classinfo_readint(CFMutableDictionaryRef ClassInfo, 
				      char * ckey, SInt32 * value)

{
  CFStringRef key;
  CFNumberRef num;
  SInt32 newval;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      num = CFDictionaryGetValue(ClassInfo, key);
      if (num && CFNumberGetValue(num, kCFNumberSInt32Type, &newval))
	{
	  errcode = 0;
	  *value = newval;
	}
      CFRelease(key);
    }
  return errcode;
}

/**************************************/
/*  Read float item from a ClassInfo  */
/**************************************/

int asysn_audiounit_classinfo_readfloat(CFMutableDictionaryRef ClassInfo, 
					char * ckey, Float32 * value)

{
  CFStringRef key;
  CFNumberRef num;
  Float32 newval;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      num = CFDictionaryGetValue(ClassInfo, key);
      if (num && CFNumberGetValue(num, kCFNumberFloat32Type, &newval))
	{
	  errcode = 0;
	  *value = newval;
	}
      CFRelease(key);
    }
  return errcode;
}

/***************************************/
/* Read CFString item from a ClassInfo */
/***************************************/

int asysn_audiounit_classinfo_readcfstr(CFMutableDictionaryRef ClassInfo, 
					char * ckey, CFStringRef * cfstr)

{
  CFStringRef key;
  CFNumberRef num;
  CFStringRef newstr;
  int errcode = -1;

  key = CFStringCreateWithCString(NULL, ckey, kCFStringEncodingASCII);
  if (key)
    {
      if (newstr = CFDictionaryGetValue(ClassInfo, key))
	{
	  errcode = 0;
	  CFRetain(newstr);
	  if (*cfstr)
	    CFRelease(*cfstr);
	  *cfstr = newstr;
	}
      CFRelease(key);
    }
  return errcode;
}

/***********************************************/
/* Big-endian int read from a CFMutableDataRef */
/***********************************************/

SInt32 asysn_audiounit_classinfo_pdata_intread(CFMutableDataRef pdata, UInt8 ** p)

{
  SInt32 value;
  
  value = ((*p)[0] << 24) | ((*p)[1] << 16) | ((*p)[2] << 8) | ((*p)[3]);
  (*p) += 4;
  return value;
}

/*************************************************/
/* Big-endian float read from a CFMutableDataRef */
/*************************************************/

Float32 asysn_audiounit_classinfo_pdata_floatread(CFMutableDataRef pdata, UInt8 ** p)

{
  union { SInt32 i; Float32 f ; } u;

  u.i = ((*p)[0] << 24) | ((*p)[1] << 16) | ((*p)[2] << 8) | ((*p)[3]);
  (*p) += 4;
  return u.f;
}

/****************************************/
/* Read parameter data from a ClassInfo */
/****************************************/

int asysn_audiounit_classinfo_readpdata(asysn_audiounit_InstanceState * My, 
					CFMutableDictionaryRef ClassInfo)

{
  int errcode = -1;
  CFStringRef key;
  CFMutableDataRef pdata;
  UInt8 * p, * pmax;
  UInt32 scope, bus, count, idx;
  Float32 fval;

  key = CFStringCreateWithCString(NULL, kAUPresetDataKey, kCFStringEncodingASCII);

  if (key && (pdata = (CFMutableDataRef) CFDictionaryGetValue(ClassInfo, key)))
    {
      p = (UInt8 *) CFDataGetBytePtr(pdata);
      pmax = p + CFDataGetLength(pdata);

      do {
	if (p == pmax)
	  {
	    errcode = 0; 
	    break;
	  }
	if ((pmax - p) >= 12)
	  {
	    scope = asysn_audiounit_classinfo_pdata_intread(pdata, &p);
	    bus = asysn_audiounit_classinfo_pdata_intread(pdata, &p);
	    count = asysn_audiounit_classinfo_pdata_intread(pdata, &p);
	    if ((pmax - p) >= count*(sizeof(SInt32) + sizeof(Float32)))
	      if ((scope == kAudioUnitScope_Global) && !bus)
		while (count--)
		  {
		    idx = asysn_audiounit_classinfo_pdata_intread(pdata, &p);
		    fval = asysn_audiounit_classinfo_pdata_floatread(pdata, &p);
		    if (idx < My->num_saolparams)
		      {
			if (fval > My->parameterinfo[idx].maxValue)
			  fval = My->parameterinfo[idx].maxValue;
			
			if (fval < My->parameterinfo[idx].minValue)
			  fval = My->parameterinfo[idx].minValue;

			My->saolparam[idx].value = fval;
		      }
		  }
	      else
		p += count*(sizeof(SInt32) + sizeof(Float32));
	    else
	      break;
	  }
	else
	  break;
      } while (1);
      
      CFRelease(key);
    }

  return errcode;
}

/************************************************************************/
/* Helper for SetProperty Selector: Reads ClassInfo returned by AU host */
/************************************************************************/

void asysn_audiounit_classinfo_read(asysn_audiounit_InstanceState * My, 
				    CFMutableDictionaryRef ClassInfo) 

{
  asysn_audiounit_classinfo_readpdata(My, ClassInfo);

  asysn_audiounit_engine_parameter_update(My); 

  asysn_audiounit_classinfo_readint(ClassInfo, kAUPresetRenderQualityKey, 
				    (SInt32 *) &(My->RenderQuality));


  OSSpinLockLock(&(My->lock_CPULoad));

  asysn_audiounit_classinfo_readfloat(ClassInfo, kAUPresetCPULoadKey, 
				      &(My->CPULoad));
  My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());
  if (My->CPULoad)
    My->ksync_normalize = My->ksync_normalize/My->CPULoad;

  OSSpinLockUnlock(&(My->lock_CPULoad));


  OSSpinLockLock(&(My->lock_PresentPreset));

  asysn_audiounit_classinfo_readcfstr(ClassInfo, kAUPresetNameKey, 
				      &(My->PresentPreset.presetName));

  My->PresentPreset.presetNumber = 
    asysn_audiounit_factorypresetnumber(My->PresentPreset.presetName); 

  OSSpinLockUnlock(&(My->lock_PresentPreset));


  /* add kAUPresetElementNameKey here if named elements/buses are supported */
}

/********************************************************************/
/* Helper Function for SetProperty: PresentProperty Factory Presets */
/********************************************************************/

void asysn_audiounit_presentproperty_setfactory
(asysn_audiounit_InstanceState * My) 

{
  int j, k;

  j = My->PresentPreset.presetNumber;

  if ((j < 0) || (j > My->num_factory))
    return;

  for (k = 0; k < My->num_saolparams; k++)
    My->saolparam[k].value = My->factorypreset_values[j][k];

  asysn_audiounit_engine_parameter_update(My); 
}

/*********************************************************************/
/* Helper for Property Selectors: returns 1 for supported properties */
/*********************************************************************/

int asysn_audiounit_supported_property(AudioUnitPropertyID id, AudioUnitScope scope) 

{
  switch (id) {
  case kAudioUnitProperty_ClassInfo: /* 0 */
    return 1;
  case kAudioUnitProperty_MakeConnection: /* 1 */
    return 1;
  case kAudioUnitProperty_SampleRate: /* 2 */
    switch (scope) {
    case kAudioUnitScope_Input:
      return (ASYS_AUDIOUNIT_EFFECT) ? 1 : 0;
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Output:
      return 1;
    default:
      return 0;
    }
  case kAudioUnitProperty_ParameterList: /* 3 */
    return  (scope == kAudioUnitScope_Global);
  case kAudioUnitProperty_ParameterInfo: /* 4 */ 
    return 1;
  case kAudioUnitProperty_FastDispatch:  /* 5 */
    return 1;
  case kAudioUnitProperty_CPULoad: /* 6 */ 
    return (scope == kAudioUnitScope_Global) ? 
      ASYS_AUDIOUNIT_SUPPORT_PROPERTY_CPULOAD : 0; 
  case kAudioUnitProperty_StreamFormat: /* 8 */ 
    switch (scope) {
    case kAudioUnitScope_Input:
      return (ASYS_AUDIOUNIT_EFFECT) ? 1 : 0;
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Output:
      return 1;
    default:
      return 0;
    }
  case kAudioUnitProperty_SRCAlgorithm: /* 9 */
    return 1;
  case kAudioUnitProperty_ElementCount: /* 11, also kAudioUnitProperty_BusCount */
    return 1;
  case kAudioUnitProperty_Latency: /* 12 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_SupportedNumChannels: /* 13 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_MaximumFramesPerSlice: /* 14 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_ParameterValueStrings: /* 16 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_TailTime: /* 20 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_BypassEffect: /* 21 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_LastRenderError: /* 22 */
    return (scope == kAudioUnitScope_Global) ? 1 : 0; 
  case kAudioUnitProperty_SetRenderCallback:  /* 23 */
    return 1;
  case kAudioUnitProperty_FactoryPresets: /* 24 */
    return 1;
  case kAudioUnitProperty_RenderQuality: /* 26 */
    return (scope == kAudioUnitScope_Global) ? 
      ASYS_AUDIOUNIT_SUPPORT_PROPERTY_RENDERQUALITY : 0; 
  case kAudioUnitProperty_InPlaceProcessing: /* 29 */
    return 1;
  case kAudioUnitProperty_CocoaUI: /* 31 */
#if defined(ASYS_AUDIOUNIT_VIEW_BUNDLECF)
    return 1;
#else
    return 0;
#endif
  case kAudioUnitProperty_OfflineRender: /* 37 */
    return 1;
  case kAudioUnitProperty_PresentPreset: /* 36 */
    return 1;
  case kAudioUnitProperty_PresentationLatency: /* 40 */
    return 1;
  case kMusicDeviceProperty_InstrumentCount: /* 1000 */
    switch (scope) {
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Input:
      return ASYS_AUDIOUNIT_MUSICDEVICE ? 1 : 0;
    default:
      return 0;
    }
  case kMusicDeviceProperty_StreamFromDisk: /* 1011 */
    return ASYS_AUDIOUNIT_SUPPORT_PROPERTY_STREAMFROMDISK;
  default:
    return 0;
  }
}

/*******************************************************************/
/* Helper for property Selectors: returns property size (in bytes) */
/*******************************************************************/

unsigned int asysn_audiounit_property_size(AudioUnitPropertyID id, 
					   asysn_audiounit_InstanceState * My)

{
  switch (id) {
  case kAudioUnitProperty_ClassInfo: /* 0 */
    return sizeof(CFPropertyListRef);   
  case kAudioUnitProperty_MakeConnection: /* 1 */
    return sizeof(AudioUnitConnection);   
  case kAudioUnitProperty_SampleRate: /* 2 */
    return sizeof(Float64);   
  case kAudioUnitProperty_ParameterList: /* 3 */
    return My->num_saolparams*sizeof(AudioUnitParameterID);   
  case kAudioUnitProperty_ParameterInfo: /* 4 */ 
    return sizeof(AudioUnitParameterInfo);   
  case kAudioUnitProperty_FastDispatch:  /* 5 */
    return sizeof(void *);   
  case kAudioUnitProperty_CPULoad: /* 6 */ 
    return sizeof(Float32);   
  case kAudioUnitProperty_StreamFormat: /* 8 */ 
    return sizeof(AudioStreamBasicDescription);   
  case kAudioUnitProperty_SRCAlgorithm: /* 9 */
    return sizeof(OSType);   
  case kAudioUnitProperty_ElementCount: /* 11, also kAudioUnitProperty_BusCount */
    return sizeof(UInt32);   
  case kAudioUnitProperty_Latency: /* 12 */
    return sizeof(Float64);   
  case kAudioUnitProperty_SupportedNumChannels: /* 13 */
    return ASYS_AUDIOUNIT_SUPPORTED_FORMATS*sizeof(AUChannelInfo);   
  case kAudioUnitProperty_MaximumFramesPerSlice: /* 14 */
    return sizeof(UInt32);   
  case kAudioUnitProperty_ParameterValueStrings: /* 16 */
    return sizeof(CFArrayRef);
  case kAudioUnitProperty_TailTime: /* 20 */
    return sizeof(Float64);   
  case kAudioUnitProperty_BypassEffect: /* 21 */
    return sizeof(UInt32);   
  case kAudioUnitProperty_LastRenderError: /* 22 */
    return sizeof(OSStatus);   
  case kAudioUnitProperty_SetRenderCallback:  /* 23 */
    return sizeof(AURenderCallbackStruct);   
  case kAudioUnitProperty_FactoryPresets: /* 24 */
    return sizeof(CFArrayRef);
  case kAudioUnitProperty_RenderQuality: /* 26 */
    return sizeof(UInt32);   
  case kAudioUnitProperty_InPlaceProcessing: /* 29 */
    return sizeof(UInt32);   
  case kAudioUnitProperty_CocoaUI: /* 31 */
#if defined(ASYS_AUDIOUNIT_VIEW_BUNDLECF)
    return sizeof(AudioUnitCocoaViewInfo);
#else
    break;
#endif
  case kAudioUnitProperty_PresentPreset: /* 36 */
    return sizeof(AUPreset);
  case kAudioUnitProperty_OfflineRender: /* 37 */
    return sizeof(UInt32);   
  case kAudioUnitProperty_PresentationLatency: /* 40 */
    return sizeof(Float64);   
  case kMusicDeviceProperty_InstrumentCount: /* 1000 */
    return sizeof(UInt32);
  case kMusicDeviceProperty_StreamFromDisk: /* 1011 */
    return sizeof(UInt32);   
  default:  /* no entries here -- all cases above should specify datasize */
    break;
  }

  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING("WARNING: asysn_audiounit_property_size()" 
				   "called on an unsupported property\n");
  return 0;
}

/***************************************************************/
/* Helper for Property Listener: returns proplisteners[] index */
/***************************************************************/

int asysn_audiounit_proplisteners_index(AudioUnitPropertyID id)

{
  int property = -1;

  switch (id) {
  case kAudioUnitProperty_ParameterList: /* 3 */ 
    property = ASYS_AUDIOUNIT_PROPLISTEN_PARAMETERLIST;
    break;
  case kAudioUnitProperty_ParameterInfo: /* 4 */ 
    property = ASYS_AUDIOUNIT_PROPLISTEN_PARAMETERINFO;
    break;
  case kAudioUnitProperty_CPULoad: /* 6 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_CPULOAD;
    break;
  case kAudioUnitProperty_Latency: /* 12 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_LATENCY;
    break;
  case kAudioUnitProperty_MaximumFramesPerSlice: /* 14 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_MAXIMUMFRAMESPERSLICE;
    break;
  case kAudioUnitProperty_TailTime: /* 20 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_TAILTIME;
    break;
  case kAudioUnitProperty_FactoryPresets: /* 24 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_FACTORYPRESETS;
    break;
  case kAudioUnitProperty_RenderQuality: /* 26 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_RENDERQUALITY;
    break;
  case kAudioUnitProperty_CurrentPreset: /* 28 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_CURRENTPRESET;
    break;
  case kAudioUnitProperty_PresentPreset: /* 36 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_PRESENTPRESET;
    break;
  case kMusicDeviceProperty_StreamFromDisk: /* 1011 */
    property = ASYS_AUDIOUNIT_PROPLISTEN_STREAMFROMDISK;
    break;
  }

  return property;
}

/***************************************************************/
/* Helper for properties: informs listeners of a status change */
/***************************************************************/

void asysn_audiounit_proplisteners_update(
                             asysn_audiounit_InstanceState * My,
			     AudioUnitPropertyID id, 
			     AudioUnitScope scope,
			     AudioUnitElement element)

{
  int idx = asysn_audiounit_proplisteners_index(id);
  asysn_audiounit_proplisten * lp;

  if (idx < 0)
    return;

  OSSpinLockLock(&(My->lock_proplisteners));

  for (lp = My->proplisteners[idx]; lp != NULL; lp = lp->next)
    (*(lp->lproc))(lp->lrefcon, My->component, id, scope, element);

  /* some apps are buggy and need a double-send ... */

  for (lp = My->proplisteners[idx]; lp != NULL; lp = lp->next)
    (*(lp->lproc))(lp->lrefcon, My->component, id, scope, element);

  OSSpinLockUnlock(&(My->lock_proplisteners));

  return;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  The istate state transition functions    */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/***************************************************************/
/* State transition diagram on entrance to InitializeSelect:   */
/*                                                             */
/*           ABSENT: Return _MY_IS_NULL.                       */
/*         INSTANCE: Set FLUX, return _INITIALIZE.             */
/* INSTANCE_PASSING: Set FLUX_PASSING, return _INITIALIZE.     */
/*             FLUX: Spin to INSTANCE* or ENGINE*.             */
/*     FLUX_PASSING: Spin to INSTANCE* or ENGINE*.             */
/*           ENGINE: Return _REINITIALIZE.                     */
/*   ENGINE_PASSING: Return _REINITIALIZE.                     */
/*    ENGINE_RENDER: Return _REINITIALIZE.                     */
/***************************************************************/

int asysn_audiounit_initializeselect_enterstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ret = ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL;
	break; 
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_INITIALIZE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_INITIALIZE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	OSMemoryBarrier();
	continue;          /* spin to _INSTANCE or _ENGINE */
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	ret = ASYS_AUDIOUNIT_IRESPONSE_REINITIALIZE;
	break;
      default:
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - initializeselect_entrance (default)\n\n");
	break; 
      }
    } while (0);

  return ret;
}

/***************************************************************/
/* State transition diagram on exit from InitializeSelect:     */
/*                                                             */
/*                                                             */
/*           ABSENT: Return _ERROR.                            */
/*         INSTANCE: Return _ERROR.                            */
/* INSTANCE_PASSING: Return _ERROR.                            */
/*             FLUX: Set ENGINE, return _COMPLETE.             */
/*     FLUX_PASSING: Set ENGINE_PASSING, return _COMPLETE.     */
/*           ENGINE: Return _ERROR.                            */
/*   ENGINE_PASSING: Return _ERROR.                            */
/*    ENGINE_RENDER: Return _ERROR.                            */
/*                                                             */
/***************************************************************/

int asysn_audiounit_initializeselect_exitstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - initializeselect_exit (absent)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - initializeselect_exit (instance*)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - initializeselect_exit (engine*)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      default:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - initializeselect_exit (default)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      }
    } while (0);
  
  return ret;
}


/********************************************************************/
/* State transition diagram on entrance to UninitializeSelect:      */
/*                                                                  */
/*           ABSENT: Return _MY_IS_NULL.                            */
/*         INSTANCE: Return _REUNINITIALIZE.                        */ 
/* INSTANCE_PASSING: Return _REUNINITIALIZE.                        */ 
/*             FLUX: Spin to INSTANCE* or ENGINE*.                  */
/*     FLUX_PASSING: Spin to INSTANCE* or ENGINE*.                  */
/*           ENGINE: Set FLUX, return _UNINITIALIZE                 */
/*   ENGINE_PASSING: Spin to ENGINE.                                */
/*    ENGINE_RENDER: Spin to ENGINE.                                */
/********************************************************************/

int asysn_audiounit_uninitializeselect_enterstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ret = ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL;
	break; 
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	ret = ASYS_AUDIOUNIT_IRESPONSE_REUNINITIALIZE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	OSMemoryBarrier();
	continue;          /* spin to _INSTANCE or _ENGINE */
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	OSMemoryBarrier();
	continue;           /* spin to _ENGINE */
      default:
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - uninitializeselect_entrance\n\n");
	break; 
      }
    } while (0);

  return ret;
}


/***************************************************************/
/* State transition diagram on exit from UninitializeSelect:   */
/*                                                             */
/*           ABSENT: Return _ERROR.                            */
/*         INSTANCE: Return _ERROR.                            */
/* INSTANCE_PASSING: Return _ERROR.                            */
/*             FLUX: Set INSTANCE, return _COMPLETE.           */
/*     FLUX_PASSING: Set INSTANCE_PASSING, return _COMPLETE.   */
/*           ENGINE: Return _ERROR.                            */
/*   ENGINE_PASSING: Return _ERROR.                            */
/*    ENGINE_RENDER: Return _ERROR.                            */
/*                                                             */
/***************************************************************/

int asysn_audiounit_uninitializeselect_exitstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - uninitializeselect_exit (absent)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - uninitializeselect_exit (instance*)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - uninitializeselect_exit (engine*)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      default:
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - uninitializeselect_exit (default)\n\n");
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	break;
      }
    } while (0);
  
  return ret;
}

/*****************************************************************/
/* State transition diagram on entrance to CloseSelect:          */
/*                                                               */
/*           ABSENT: Return _MY_IS_NULL.                         */
/*         INSTANCE: Set ABSENT, return _CLOSE.                  */
/* INSTANCE_PASSING: Spin to INSTANCE                            */
/*             FLUX: Spin to INSTANCE or ENGINE                  */
/*     FLUX_PASSING: Spin to INSTANCE or ENGINE                  */
/*           ENGINE: Set ABSENT, return UNINITIALIZE_AND_CLOSE   */
/*   ENGINE_PASSING: Spin to ENGINE                              */
/*    ENGINE_RENDER: Spin to ENGINE                              */
/*****************************************************************/

int asysn_audiounit_closeselect_enterstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ret = ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL;
	break; 
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     ASYS_AUDIOUNIT_ISTATE_ABSENT, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_CLOSE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	OSMemoryBarrier();
	continue;          /* spin to _INSTANCE */
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	OSMemoryBarrier();
	continue;          /* spin to _INSTANCE or _ENGINE */
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     ASYS_AUDIOUNIT_ISTATE_ABSENT, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE_AND_CLOSE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	OSMemoryBarrier();
	continue;          /* spin to _ENGINE */
	break;
      default:
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - closeselect_enter\n\n");
	break; 
      }
    } while (0);
  
  return ret;
}

/******************************************************************/
/* State transition diagram on entrance to real-time thread:      */
/*                                                                */
/*           ABSENT: Return _MY_IS_NULL (abnormal)                */
/*         INSTANCE: Set INSTANCE_PASSING, return _WIRE           */
/* INSTANCE_PASSING: Set INSTANCE, return _WIRE (abnormal)        */
/*             FLUX: Set FLUX_PASSING, return _WIRE               */
/*     FLUX_PASSING: Set FLUX, return _WIRE (abnormal)            */
/*           ENGINE: Set ENGINE_RENDER, return _RENDER            */
/*   ENGINE_PASSING: Set ENGINE_RENDER, return _RENDER (abnormal) */
/*    ENGINE_RENDER: Return _RENDER (abnormal)                    */
/******************************************************************/

int asysn_audiounit_realtime_enterstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ret = ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (absent)\n\n");
	break; 
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_WIRE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_WIRE;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (instance_passing)\n\n");
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_WIRE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_WIRE;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (flux_passing)\n\n");
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_RENDER;
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_RENDER;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (engine_passing)\n\n");
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_RENDER;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (engine_render)\n\n");
	break;
      default:
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_entrance (default)\n\n");
	break; 
      }
    } while (0);

  return ret;
}

/************************************************************/
/* State transition diagram on exit from real-time thread:  */
/*                                                          */
/*           ABSENT: Return _MY_IS_NULL (abnormal)          */
/*         INSTANCE: Return _COMPLETE (abnormal)            */
/* INSTANCE_PASSING: Set INSTANCE, return _COMPLETE         */
/*             FLUX: Return _COMPLETE (abnormal)            */
/*     FLUX_PASSING: Set FLUX, return _COMPLETE             */
/*           ENGINE: Return _COMPLETE (abnormal)            */
/*   ENGINE_PASSING: Set ENGINE, return _COMPLETE           */
/*    ENGINE_RENDER: Set ENGINE, return _COMPLETE           */
/************************************************************/

int asysn_audiounit_realtime_exitstate(asysn_audiounit_InstanceState * My)

{
  int ret;

  do 
    {
      switch(My->istate) {
      case ASYS_AUDIOUNIT_ISTATE_ABSENT:
	ret = ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_exit (absent)\n\n");
	break; 
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE:
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_exit (instance)\n\n");
	break;
      case ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_INSTANCE_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_INSTANCE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX:
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_exit (flux)\n\n");
	break;
      case ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_FLUX_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_FLUX, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE:
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_exit (engine)\n\n");
	break;
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE_PASSING, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break; 
      case ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER:
	if (!OSAtomicCompareAndSwap32Barrier
	    (ASYS_AUDIOUNIT_ISTATE_ENGINE_RENDER, 
	     ASYS_AUDIOUNIT_ISTATE_ENGINE, 
	     (int32_t *) &(My->istate)))
	  continue;
	ret = ASYS_AUDIOUNIT_IRESPONSE_COMPLETE;
	break;
      default:
	ret = ASYS_AUDIOUNIT_IRESPONSE_ERROR;
	ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	  ("\tBad transition - realtime_exit (default)\n\n");
	break; 
      }
    } while (0);

  return ret;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  The Selector Calls    */
/*~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~*/

/**************************************/
/*          Selector: Open            */
/*                                    */
/*  params[0]: ComponentInstance ci   */
/*                                    */
/**************************************/

ComponentResult asysn_audiounit_kComponentOpenSelect(ComponentParameters * p, 
						     asysn_audiounit_InstanceState * My)
{ 
  int socket_flags;
  int i, msize;

  msize = sizeof(asysn_audiounit_InstanceState);
  if (!(My = (asysn_audiounit_InstanceState *) calloc(1, msize)))
    return asysn_audiounit_instance_cleanup(p, My);
  asysn_audiounit_memstatus(My, msize, MADV_WILLNEED);

  My->component = (ComponentInstance) (p->params[0]);
  sprintf(My->componentname, "c%u", (unsigned int) My->component);
  My->argv[0] = My->componentname;

  if (asysn_audiounit_initialize_properties(My) != noErr)
    return asysn_audiounit_instance_cleanup(p, My);

  if (ASYS_AUDIOUNIT_EFFECT)
    {
      msize = (sizeof(UInt32) + sizeof(AudioBuffer)*ASYS_AUDIOUNIT_INPUT_MAXCHANNELS);

      if (!(My->AudioBufferCarrier = calloc(msize, 1)))
	return asysn_audiounit_instance_cleanup(p, My);

      if (!(My->AudioBufferTemplate = calloc(msize, 1)))
	return asysn_audiounit_instance_cleanup(p, My);

      asysn_audiounit_memstatus(My->AudioBufferCarrier, msize, MADV_WILLNEED);
      asysn_audiounit_memstatus(My->AudioBufferTemplate, msize, MADV_WILLNEED);
      
      My->AudioBufferTemplate->mNumberBuffers = ASYS_AUDIOUNIT_INPUT_CHANNELS;

      msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;

      for(i = 0; i < ASYS_AUDIOUNIT_INPUT_MAXCHANNELS; i++)
	{
	  My->AudioBufferTemplate->mBuffers[i].mNumberChannels = 1;

	  if (!(My->AudioBufferTemplate->mBuffers[i].mData = malloc(msize)))
	    return asysn_audiounit_instance_cleanup(p, My);

	  asysn_audiounit_memstatus(My->AudioBufferTemplate->mBuffers[i].mData, 
				    msize, MADV_WILLNEED);
	}
    }

  msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;

  for(i = 0; i < ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS; i++)
    {
      if (!(My->mData_Output[i] = (Float32 *) malloc(msize)))
	return asysn_audiounit_instance_cleanup(p, My);
      asysn_audiounit_memstatus(My->mData_Output[i], msize, MADV_WILLNEED);
    }

  if (ASYS_AUDIOUNIT_HAS_AUCONTROL)
    {
      /*~~~~~~~~~~~~~~~~~~~~*/
      /* set up MIDI socket */
      /*--------------------*/

      if (socketpair(AF_UNIX, SOCK_DGRAM, 0, My->mpipepair))
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during MIDI socketpair() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	} 

      if ((socket_flags = fcntl(My->mpipepair[0], F_GETFL, 0)) == -1)
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during mpipepair[0] F_GETFL fcntl() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	}

      if (fcntl(My->mpipepair[0], F_SETFL, socket_flags | O_NONBLOCK) == -1)
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during mpipepair[0] F_SETFL fcntl() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	}

      /*~~~~~~~~~~~~~~~~~~~~*/
      /* set up SASL socket */
      /*--------------------*/

      if (socketpair(AF_UNIX, SOCK_DGRAM, 0, My->spipepair))
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during SASL socketpair() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	} 

      if ((socket_flags = fcntl(My->spipepair[0], F_GETFL, 0)) == -1)
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during spipepair[0] F_GETFL fcntl() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	}

      if (fcntl(My->spipepair[0], F_SETFL, socket_flags | O_NONBLOCK) == -1)
	{
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\n\tIn OpenSelect: Error during spipepair[0] F_SETFL fcntl() call.\n");
	  return asysn_audiounit_instance_cleanup(p, My);
	}

      /*~~~~~~~~~~~~~~~~~~~~~~*/
      /* set up argc and argv */
      /*----------------------*/

      My->argc = 5;

      My->argv[1] = My->mpipeflag;
      My->argv[2] = My->mpipevalue;

      sprintf(My->mpipeflag, "-asys_audiounit_mpipe");
      sprintf(My->mpipevalue, "%i", My->mpipepair[0]);

      My->argv[3] = My->spipeflag;
      My->argv[4] = My->spipevalue;

      sprintf(My->spipeflag, "-asys_audiounit_spipe");
      sprintf(My->spipevalue, "%i", My->spipepair[0]);
      }
  else
    {
      My->argc = 1;

      My->argv[1] = My->argv[2] = NULL;
      My->mpipepair[0] = My->mpipepair[1] = 0;

      My->argv[3] = My->argv[4] = NULL;
      My->spipepair[0] = My->spipepair[1] = 0;
    }

  if (asysn_audiounit_initialize_parametersystem(My) != noErr)
    return asysn_audiounit_instance_cleanup(p, My);

  OSSpinLockLock(&asysn_audiounit_lock_opencount);

  if (!(asysn_audiounit_opencount++))
    asysn_audiounit_globalvars_memstatus(MADV_WILLNEED);

  OSSpinLockUnlock(&asysn_audiounit_lock_opencount);

  My->istate = ASYS_AUDIOUNIT_ISTATE_INSTANCE;

  SetComponentInstanceStorage((ComponentInstance) p->params[0], (Handle) My); 

  return noErr; 
}

/**************************************/
/*           Selector: Close          */
/*                                    */
/*  params[0]: ComponentInstance ci   */
/*                                    */
/**************************************/

ComponentResult asysn_audiounit_kComponentCloseSelect(ComponentParameters * p, 
					    asysn_audiounit_InstanceState * My)
{ 
  ENGINE_PTR_DECLARE_SEMICOLON
  asysn_audiounit_proplisten * remove;
  asysn_audiounit_rendernotify * notify;
  int i, k, msize, ret;

  ret = asysn_audiounit_closeselect_enterstate(My);

  SetComponentInstanceStorage((ComponentInstance) p->params[0], (Handle) NULL); 

  switch (ret) {
  case ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE_AND_CLOSE:
    ENGINE_PTR_ASSIGNED_TO  MY_ENGINE_PTR;
    ENGINE_MEMSTATUS_SEMICOLON(MADV_FREE)
    shut_down(ENGINE_PTR);
    MY_ENGINE_PTR_ASSIGNED_TO  NULL;
    break;
  case ASYS_AUDIOUNIT_IRESPONSE_CLOSE:
    break;
  case ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL:
  case ASYS_AUDIOUNIT_IRESPONSE_ERROR:
  default:
    return noErr;
  }

  if (ASYS_AUDIOUNIT_HAS_AUCONTROL)
    {
      close(My->mpipepair[0]);
      close(My->mpipepair[1]);
      close(My->spipepair[0]);
      close(My->spipepair[1]);

      if (My->num_saolparams)
	{
	  free(My->parameterlist);
	  for (i = 0; i < My->num_saolparams; i++)
	    {
	      if (My->parameterinfo[i].cfNameString)
		CFRelease(My->parameterinfo[i].cfNameString);
	      if (My->parameterinfo[i].unitName)
		CFRelease(My->parameterinfo[i].unitName);
	    }
	  free(My->parameterinfo);

	  for (i = 0; i < My->num_saolparams; i++)
	    if (My->pvs_size[i])
	      {
		for (k = 0; k < My->pvs_size[i]; k++)
		  free(My->pvs_cstr[i][k]);
		free(My->pvs_cstr[i]);
	      }
	  free(My->pvs_size);
	  free(My->pvs_cstr);

	  msize = My->num_saolparams*sizeof(asysn_audiounit_saolparam);
	  asysn_audiounit_memstatus(My->saolparam, msize, MADV_FREE);
	  free(My->saolparam);
	}

      if (My->num_factory)
	{
	  for (i = 0; i < My->num_factory; i++)
	    {
	      if (My->factorypreset_values[i])
		{
		  msize = My->num_saolparams*sizeof(Float32);
		  asysn_audiounit_memstatus(My->factorypreset_values[i], 
					    msize, MADV_FREE);
		  free(My->factorypreset_values[i]);
		}

	      if (My->factorypreset_info[i].presetName)
		CFRelease(My->factorypreset_info[i].presetName);
	    }

	  msize = My->num_factory*sizeof(Float32 *);
	  asysn_audiounit_memstatus(My->factorypreset_values, msize, MADV_FREE);
	  free(My->factorypreset_values);

	  free(My->factorypreset_info);
	  CFRelease(My->factorypreset_array);
	}
    }


  OSSpinLockLock(&(My->lock_PresentPreset));

  CFRelease(My->PresentPreset.presetName);

  OSSpinLockUnlock(&(My->lock_PresentPreset));


  OSSpinLockLock(&(My->lock_proplisteners));

  for (i = 0; i < ASYS_AUDIOUNIT_PROPLISTEN_ARRAYSIZE; i++)
    while (My->proplisteners[i] != NULL)
      {
	remove = My->proplisteners[i];
	My->proplisteners[i] = My->proplisteners[i]->next;
	msize = sizeof(asysn_audiounit_proplisten);
	asysn_audiounit_memstatus(remove, msize, MADV_FREE);
	free(remove);
      }

  OSSpinLockUnlock(&(My->lock_proplisteners));


  OSSpinLockLock(&(My->lock_rendernotify));

  while (My->rendernotify)
    {
      notify = My->rendernotify;
      My->rendernotify = notify->next;
      msize = sizeof(asysn_audiounit_rendernotify);
      asysn_audiounit_memstatus(notify, msize, MADV_FREE);
      free(notify);
    }

  OSSpinLockUnlock(&(My->lock_rendernotify));


  if (ASYS_AUDIOUNIT_EFFECT)
    {
      msize = (sizeof(UInt32) + sizeof(AudioBuffer)*ASYS_AUDIOUNIT_INPUT_MAXCHANNELS);
      asysn_audiounit_memstatus(My->AudioBufferCarrier, msize, MADV_FREE);
      free(My->AudioBufferCarrier);

      msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;
      for(i = 0; i < ASYS_AUDIOUNIT_INPUT_MAXCHANNELS; i++)
	{
	  asysn_audiounit_memstatus(My->AudioBufferTemplate->mBuffers[i].mData, 
				    msize, MADV_FREE);
	  free(My->AudioBufferTemplate->mBuffers[i].mData);
	}

      msize = (sizeof(UInt32) + sizeof(AudioBuffer)*ASYS_AUDIOUNIT_INPUT_MAXCHANNELS);
      asysn_audiounit_memstatus(My->AudioBufferTemplate, msize, MADV_FREE);
      free(My->AudioBufferTemplate);
    }

  for(i = 0; i < ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS; i++)
    {
      msize = sizeof(Float32)*ASYS_AUDIOUNIT_FRAMES_PER_SLICE;
      asysn_audiounit_memstatus(My->mData_Output[i], msize, MADV_FREE);
      free(My->mData_Output[i]);
    }

  msize = sizeof(asysn_audiounit_InstanceState);
  asysn_audiounit_memstatus(My, msize, MADV_FREE);
  free(My);

  OSSpinLockLock(&asysn_audiounit_lock_opencount);

  if (!(--asysn_audiounit_opencount))
    asysn_audiounit_globalvars_memstatus(MADV_FREE);

  if (asysn_audiounit_opencount < 0)   /* should never be true */
    asysn_audiounit_opencount = 0;

  OSSpinLockUnlock(&asysn_audiounit_lock_opencount);

  return noErr; 
}

/************************/
/* Selector: Initialize */
/*                      */
/************************/

ComponentResult asysn_audiounit_kAudioUnitInitializeSelect(ComponentParameters * p, 
  					        asysn_audiounit_InstanceState * My)
{ 
  ENGINE_PTR_DECLARE_SEMICOLON

  switch (asysn_audiounit_initializeselect_enterstate(My)) {
    case ASYS_AUDIOUNIT_IRESPONSE_INITIALIZE:
      break;
    case ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL:
    case ASYS_AUDIOUNIT_IRESPONSE_REINITIALIZE:
    default:
      return noErr;
    }

  ENGINE_PTR_ASSIGNED_TO  system_init(My->argc, My->argv, 
				      My->OutputStreamFormat.mSampleRate);
  
  MY_ENGINE_PTR_ASSIGNED_TO  ENGINE_PTR;
  
  if (ENGINE_PTR_IS_NULL)
    return kAudioUnitErr_FailedInitialization;
  
  ENGINE_MEMSTATUS_SEMICOLON(MADV_WILLNEED)
    
  effects_init(ENGINE_PTR);
  main_initpass(ENGINE_PTR);
  asysn_audiounit_ksyncinit(My);
  
  EV(kcycleidx) = EV(kbase);
  EV(acycleidx) = EV(ACYCLE) + 1;   /* +1 denotes zero-time condition */
  EV(pass) = IPASS;
  
  My->acycle = EV(ACYCLE);

  OSSpinLockLock(&(My->lock_CPULoad));

  My->krate = EV(KRATE);

  My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());
  if (My->CPULoad)
    My->ksync_normalize = My->ksync_normalize/My->CPULoad;

  OSSpinLockUnlock(&(My->lock_CPULoad));

  My->acycleidx_kcycleidx = 1;     /* acycleidx = 0, kcycleidx = 1 */

  if (ASYS_AUDIOUNIT_HAS_AUCONTROL)
    asysn_audiounit_engine_parameter_update(My);
  
  asysn_audiounit_initializeselect_exitstate(My);

  return noErr;
}

/**************************/
/* Selector: Uninitialize */
/*                        */
/**************************/

ComponentResult asysn_audiounit_kAudioUnitUninitializeSelect(ComponentParameters * p, 
						   asysn_audiounit_InstanceState * My)
{ 
  int j;
  ENGINE_PTR_DECLARE_SEMICOLON

  switch (asysn_audiounit_uninitializeselect_enterstate(My)) {
    case ASYS_AUDIOUNIT_IRESPONSE_UNINITIALIZE:
      break;
    case ASYS_AUDIOUNIT_IRESPONSE_MY_IS_NULL:
    case ASYS_AUDIOUNIT_IRESPONSE_REUNINITIALIZE:
    default:
      return noErr;
    }

  ENGINE_PTR_ASSIGNED_TO  MY_ENGINE_PTR;
  ENGINE_MEMSTATUS_SEMICOLON(MADV_FREE)
  shut_down(ENGINE_PTR);
  MY_ENGINE_PTR_ASSIGNED_TO  NULL;

  My->acycle = ((int)(SAOL_SRATE))/((int)(SAOL_KRATE));

  OSSpinLockLock(&(My->lock_CPULoad));

  My->krate = SAOL_KRATE;

  My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());
  if (My->CPULoad)
    My->ksync_normalize = My->ksync_normalize/My->CPULoad;

  OSSpinLockUnlock(&(My->lock_CPULoad));

  My->acycleidx_kcycleidx = 1;   /* acycleidx = 0, kcycleidx = 1 */

  asysn_audiounit_uninitializeselect_exitstate(My);
  asysn_audiounit_engine_emptypipes(My);

  return noErr; 
}

/********************************************/
/*        Selector: GetPropertyInfo         */
/*                                          */
/*  params[0]: Boolean * outWritable;       */
/*  params[1]: UInt32 * outDataSize;        */
/*  params[2]: AudioUnitElement inElement;  */
/*  params[3]: AudioUnitScope inScope;      */
/*  params[4]: AudioUnitPropertyID inID;    */
/*                                          */
/********************************************/

ComponentResult asysn_audiounit_kAudioUnitGetPropertyInfoSelect
(ComponentParameters * p, asysn_audiounit_InstanceState * My)

{ 
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* return early for invalid properties */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (!asysn_audiounit_supported_property((AudioUnitPropertyID)(p->params[4]),
			  (AudioUnitScope)(p->params[3])))
    return kAudioUnitErr_InvalidProperty; 

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* provide datasize if requested by a non-null pointer */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (p->params[1])
    *((UInt32 *)(p->params[1])) = asysn_audiounit_property_size(p->params[4], My);

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* provide writable status if requested by a non-null pointer */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (p->params[0])
    {
      switch (p->params[4] /* property ID */) {
      case kAudioUnitProperty_ParameterList: /* 3 */
      case kAudioUnitProperty_ParameterInfo: /* 4 */ 
      case kAudioUnitProperty_FastDispatch:  /* 5 */
	*((Boolean *)(p->params[0])) = 0;   
	break;
      case kAudioUnitProperty_ElementCount: /* 11, also kAudioUnitProperty_BusCount */
	/* for hosts that require AUs to support side-chaining */
	*((Boolean *)(p->params[0])) = (p->params[3] == kAudioUnitScope_Input);   
	break;
      case kAudioUnitProperty_SupportedNumChannels: /* 13 */
      case kAudioUnitProperty_ParameterValueStrings: /* 16 */
      case kAudioUnitProperty_FactoryPresets: /* 24 */
      case kAudioUnitProperty_CocoaUI: /* 31 */
	*((Boolean *)(p->params[0])) = 0;   
	break;
      default:
	*((Boolean *)(p->params[0])) = 1;   
	break;
      }
    }

  return noErr;
}

/********************************************/
/*         Selector: GetProperty            */
/*                                          */
/*  params[0]: UInt32 * ioDataSize;         */
/*  params[1]: void * outData;              */
/*  params[2]: AudioUnitElement inElement;  */
/*  params[3]: AudioUnitScope inScope;      */
/*  params[4]: AudioUnitPropertyID inID;    */
/*                                          */
/********************************************/

ComponentResult asysn_audiounit_kAudioUnitGetPropertySelect(ComponentParameters * p, 
					       asysn_audiounit_InstanceState * My)
{ 
  ComponentResult result = noErr;
  CFBundleRef au_bundle = NULL;
  CFURLRef view_bundle_url = NULL;
  UInt32 size;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* return early for invalid properties */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (!asysn_audiounit_supported_property((AudioUnitPropertyID)(p->params[4]),
			  (AudioUnitScope)(p->params[3])))
    return kAudioUnitErr_InvalidProperty; 

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* check size of property storage */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (*((UInt32 *)(p->params[0])) != 
      (size = asysn_audiounit_property_size(p->params[4], My)))
    if (*((UInt32 *)(p->params[0])) > size)
      *((UInt32 *)(p->params[0])) = size;
    else
      return kAudioUnitErr_InvalidPropertyValue;

  /*~~~~~~~~~~~~~~~~~~~~~~~*/
  /* assign property value */
  /*~~~~~~~~~~~~~~~~~~~~~~~*/

  switch (p->params[4] /* property ID */) {
  case kAudioUnitProperty_ClassInfo: /* 0 */   
    *((CFPropertyListRef *)((void *)(p->params[1]))) = 
      asysn_audiounit_classinfo_create(My);
    break;
  case kAudioUnitProperty_MakeConnection: /* 1 */
    if ((p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    OSSpinLockLock(&(My->lock_sampledelivery));
    ((AudioUnitConnection *)((void *)(p->params[1])))->sourceAudioUnit
      = My->MakeConnection[p->params[2]].sourceAudioUnit;
    ((AudioUnitConnection *)((void *)(p->params[1])))->sourceOutputNumber
      = My->MakeConnection[p->params[2]].sourceOutputNumber;
    ((AudioUnitConnection *)((void *)(p->params[1])))->destInputNumber
      = My->MakeConnection[p->params[2]].destInputNumber;
    OSSpinLockUnlock(&(My->lock_sampledelivery));
    break;
  case kAudioUnitProperty_SampleRate: /* 2 */
    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:
    if (!(ASYS_AUDIOUNIT_EFFECT) || 
	(p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    *((Float64 *)((void *)(p->params[1]))) = 
      My->InputStreamFormat[p->params[2]].mSampleRate;
    break;
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Output:
      *((Float64 *)((void *)(p->params[1]))) = My->OutputStreamFormat.mSampleRate;
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_ParameterList: /* 3 */
    memcpy(((AudioUnitParameterID *)((void *)(p->params[1]))), My->parameterlist,
	   sizeof(AudioUnitParameterID)*My->num_saolparams);
    break;
  case kAudioUnitProperty_ParameterInfo: /* 4 */
    if (p->params[2] < My->num_saolparams)  /* p->param[2] requests ParameterID. */
      {
	if (My->parameterinfo[p->params[2]].cfNameString)
	  CFRetain(My->parameterinfo[p->params[2]].cfNameString);
	if (My->parameterinfo[p->params[2]].unitName)
	  CFRetain(My->parameterinfo[p->params[2]].unitName);
	memcpy(((AudioUnitParameterInfo *)((void *)(p->params[1]))), 
	       &(My->parameterinfo[p->params[2]]), sizeof(AudioUnitParameterInfo));
      }
    else
      result = kAudioUnitErr_InvalidProperty; 
    break;
  case kAudioUnitProperty_FastDispatch:  /* 5 */
    switch (p->params[2] /* inElement */ )  {
    case kAudioUnitRenderSelect:
      *((AudioUnitRenderProc *)(p->params[1])) = asysn_audiounit_MyRenderer;
      break;
    case kMusicDeviceMIDIEventSelect:
      if (ASYS_AUDIOUNIT_MIDISUPPORT)
	*((MusicDeviceMIDIEventProc *)(p->params[1])) = asysn_audiounit_MyMIDIEventProc;
      else
	result = kAudioUnitErr_InvalidProperty; 
      break;
    case kAudioUnitGetParameterSelect:
      *((AudioUnitGetParameterProc *)(p->params[1])) = 
	asysn_audiounit_MyGetParameterProc;
      break;
    case kAudioUnitSetParameterSelect:
      *((AudioUnitSetParameterProc *)(p->params[1])) = 
	asysn_audiounit_MySetParameterProc;
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_CPULoad: /* 6 */ 
    *((Float32 *)((void *)(p->params[1]))) = My->CPULoad;
    break;
  case kAudioUnitProperty_StreamFormat: /* 8 */ 
    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:
    if (!(ASYS_AUDIOUNIT_EFFECT) || 
	(p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame
      = My->InputStreamFormat[p->params[2]].mChannelsPerFrame;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mSampleRate
      = My->InputStreamFormat[p->params[2]].mSampleRate;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFormatID
      = My->InputStreamFormat[p->params[2]].mFormatID;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFormatFlags
      = My->InputStreamFormat[p->params[2]].mFormatFlags;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBytesPerPacket
      = My->InputStreamFormat[p->params[2]].mBytesPerPacket;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBytesPerFrame
      = My->InputStreamFormat[p->params[2]].mBytesPerFrame;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFramesPerPacket
      = My->InputStreamFormat[p->params[2]].mFramesPerPacket;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBitsPerChannel
      = My->InputStreamFormat[p->params[2]].mBitsPerChannel;
    ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mReserved
      = My->InputStreamFormat[p->params[2]].mReserved;
      break;
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Output:
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame
	= My->OutputStreamFormat.mChannelsPerFrame;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mSampleRate
	= My->OutputStreamFormat.mSampleRate;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFormatID
	= My->OutputStreamFormat.mFormatID;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFormatFlags
	= My->OutputStreamFormat.mFormatFlags;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBytesPerPacket
	= My->OutputStreamFormat.mBytesPerPacket;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBytesPerFrame
	= My->OutputStreamFormat.mBytesPerFrame;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFramesPerPacket
	= My->OutputStreamFormat.mFramesPerPacket;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mBitsPerChannel
	= My->OutputStreamFormat.mBitsPerChannel;
      ((AudioStreamBasicDescription *)((void *)(p->params[1])))->mReserved
	= My->OutputStreamFormat.mReserved;
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_SRCAlgorithm: /* 9 */
    *((OSType *)((void *)(p->params[1]))) = My->SRCAlgorithm;
    break;
  case kAudioUnitProperty_ElementCount: /* 11, also kAudioUnitProperty_BusCount */
    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:
      *((UInt32 *)((void *)(p->params[1]))) = My->InputElementCount;
      break;
    case kAudioUnitScope_Global:
    case kAudioUnitScope_Output:
      *((UInt32 *)((void *)(p->params[1]))) = My->OutputElementCount;
      break;
    case kAudioUnitScope_Part:
    case kAudioUnitScope_Group:
      *((UInt32 *)((void *)(p->params[1]))) = 0;
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_Latency: /* 12 */
    *((Float64 *)((void *)(p->params[1]))) = My->Latency;
    break;
  case kAudioUnitProperty_SupportedNumChannels: /* 13 */
    memcpy(((AudioUnitParameterID *)((void *)(p->params[1]))), My->SupportedNumChannels,
	   ASYS_AUDIOUNIT_SUPPORTED_FORMATS*sizeof(AUChannelInfo));
    break;
  case kAudioUnitProperty_MaximumFramesPerSlice: /* 14 */
    *((UInt32 *)((void *)(p->params[1]))) = My->MaximumFramesPerSlice;
    break;
  case kAudioUnitProperty_ParameterValueStrings: /* 16 */
    if ((*((CFMutableArrayRef *)((void *)(p->params[1]))) = 
	 asysn_audiounit_parametervaluestrings_create(My, p->params[2])) == NULL)
      result = kAudioUnitErr_InvalidProperty; 
    break;
  case kAudioUnitProperty_TailTime: /* 20 */
    *((Float64 *)((void *)(p->params[1]))) = My->TailTime;
    break;
  case kAudioUnitProperty_BypassEffect: /* 21 */
    *((UInt32 *)((void *)(p->params[1]))) = My->BypassEffect;
    break;
  case kAudioUnitProperty_LastRenderError: /* 22 */
    *((OSStatus *)((void *)(p->params[1]))) = My->LastRenderError;
    My->LastRenderError = noErr;
    break;
  case kAudioUnitProperty_SetRenderCallback:  /* 23 */
    if ((p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    OSSpinLockLock(&(My->lock_sampledelivery));
    ((AURenderCallbackStruct *)((void *)(p->params[1])))->inputProc
      = My->SetRenderCallback[p->params[2]].inputProc;
    ((AURenderCallbackStruct *)((void *)(p->params[1])))->inputProcRefCon
      = My->SetRenderCallback[p->params[2]].inputProcRefCon;
    OSSpinLockUnlock(&(My->lock_sampledelivery));
    break;
  case kAudioUnitProperty_FactoryPresets: /* 24 */
    if (My->num_factory)
      {
	CFRetain(My->factorypreset_array);
	*((CFPropertyListRef *)((void *)(p->params[1]))) = My->factorypreset_array;
      }
    else
      result = kAudioUnitErr_InvalidProperty; 
    break;
  case kAudioUnitProperty_RenderQuality: /* 26 */
    *((UInt32 *)((void *)(p->params[1]))) = My->RenderQuality;
    break;
  case kAudioUnitProperty_InPlaceProcessing: /* 29 */
    *((UInt32 *)((void *)(p->params[1]))) = My->InPlaceProcessing;
    break;
  case kAudioUnitProperty_CocoaUI: /* 31 */
#if defined(ASYS_AUDIOUNIT_VIEW_BUNDLECF)

    if ((au_bundle = CFBundleGetBundleWithIdentifier(ASYS_AUDIOUNIT_AU_BUNDLECF)))
      CFRetain(au_bundle);
    else
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }

    view_bundle_url = CFBundleCopyResourceURL(au_bundle,
					      ASYS_AUDIOUNIT_VIEW_BUNDLECF,
					      CFSTR("bundle"), NULL);
    CFRelease(au_bundle);

    if (!view_bundle_url)
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }

    ((AudioUnitCocoaViewInfo *)(p->params[1]))->mCocoaAUViewBundleLocation =
      view_bundle_url;

    ((AudioUnitCocoaViewInfo *)(p->params[1]))->mCocoaAUViewClass[0] =
      CFStringCreateCopy(NULL, ASYS_AUDIOUNIT_VIEW_BASECLASSCF);
#else
    result = kAudioUnitErr_InvalidProperty; 
#endif
    break;
  case kAudioUnitProperty_PresentPreset: /* 36 */
    OSSpinLockLock(&(My->lock_PresentPreset));
    ((AUPreset *)(p->params[1]))->presetNumber = My->PresentPreset.presetNumber;
    ((AUPreset *)(p->params[1]))->presetName = My->PresentPreset.presetName;
    CFRetain(My->PresentPreset.presetName);
    OSSpinLockUnlock(&(My->lock_PresentPreset));
    break;
  case kAudioUnitProperty_OfflineRender: /* 37 */
    *((UInt32 *)((void *)(p->params[1]))) = My->OfflineRender;
    break;
  case kAudioUnitProperty_PresentationLatency: /* 40 */
    *((Float64 *)((void *)(p->params[1]))) = My->PresentationLatency;
    break;
  case kMusicDeviceProperty_InstrumentCount: /* 1000 */
    if (ASYS_AUDIOUNIT_MUSICDEVICE)
      *((UInt32 *)((void *)(p->params[1]))) = My->InstrumentCount;
    else
      result = kAudioUnitErr_InvalidProperty; 
    break;
  case kMusicDeviceProperty_StreamFromDisk: /* 1011 */
    *((UInt32 *)((void *)(p->params[1]))) = My->StreamFromDisk;
    break;
  default:  /* should never happen */
    result = kAudioUnitErr_InvalidProperty; 
    break;
  }

  return result;
}

/********************************************/
/*         Selector: SetProperty            */
/*                                          */
/*  params[0]: UInt32 inDataSize;           */
/*  params[1]: const void * inData;         */
/*  params[2]: AudioUnitElement inElement;  */
/*  params[3]: AudioUnitScope inScope;      */
/*  params[4]: AudioUnitPropertyID inID;    */
/*                                          */
/********************************************/

ComponentResult asysn_audiounit_kAudioUnitSetPropertySelect(ComponentParameters * p, 
					       asysn_audiounit_InstanceState * My)
{ 
  ComponentResult result = noErr; 
  CFStringRef key;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* return early for invalid properties */
  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

  if (!asysn_audiounit_supported_property((AudioUnitPropertyID)(p->params[4]),
			  (AudioUnitScope)(p->params[3])))
    return kAudioUnitErr_InvalidProperty; 

  switch (p->params[4] /* property ID */) {
  case kAudioUnitProperty_ClassInfo: /* 0 */  
    asysn_audiounit_classinfo_read(
                        My, *((CFMutableDictionaryRef *)((void *)(p->params[1])))); 
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_PresentPreset,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_RenderQuality,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_CPULoad,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_MakeConnection: /* 1 */
    if ((p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    OSSpinLockLock(&(My->lock_sampledelivery));
    My->MakeConnection[p->params[2]].sourceAudioUnit = 
      ((AudioUnitConnection *)((void *)(p->params[1])))->sourceAudioUnit;
    My->MakeConnection[p->params[2]].sourceOutputNumber = 
      ((AudioUnitConnection *)((void *)(p->params[1])))->sourceOutputNumber;
    My->MakeConnection[p->params[2]].destInputNumber = 
      ((AudioUnitConnection *)((void *)(p->params[1])))->destInputNumber;
    OSSpinLockUnlock(&(My->lock_sampledelivery));
    break;
  case kAudioUnitProperty_SampleRate: /* 2 */
    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:
      if (!(ASYS_AUDIOUNIT_EFFECT) || 
	  (p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	  (p->params[2] < 0))
	{
	  result = kAudioUnitErr_InvalidProperty; 
	  break;
	}
	My->InputStreamFormat[p->params[2]].mSampleRate = 
	  *((Float64 *)((void *)(p->params[1]))); 
      break;
    case kAudioUnitScope_Output:
      My->OutputStreamFormat.mSampleRate = *((Float64 *)((void *)(p->params[1])));
      break;
    case kAudioUnitScope_Global:
      if ((ASYS_AUDIOUNIT_EFFECT) && 
	  (p->params[2] < ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	  (p->params[2] >= 0))
	My->InputStreamFormat[p->params[2]].mSampleRate =
	  *((Float64 *)((void *)(p->params[1]))); 
      My->OutputStreamFormat.mSampleRate = *((Float64 *)((void *)(p->params[1])));
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_ParameterList: /* 3 */
    result = kAudioUnitErr_InvalidProperty; 
    break;    /* hosts not permitted to change Parameter List */
  case kAudioUnitProperty_ParameterInfo: /* 4 */ 
    result = kAudioUnitErr_InvalidProperty; 
    break;    /* hosts not permitted to change Parameter Info */
  case kAudioUnitProperty_FastDispatch:  /* 5 */
    result = kAudioUnitErr_InvalidProperty; 
    break;    /* hosts not permitted to set the dispatch routines */
  case kAudioUnitProperty_CPULoad: /* 6 */

    OSSpinLockLock(&(My->lock_CPULoad));

    My->CPULoad = *((Float32 *)((void *)(p->params[1])));
    My->ksync_normalize = (Float32) (My->krate/AudioGetHostClockFrequency());
    if (My->CPULoad)
      My->ksync_normalize = My->ksync_normalize/My->CPULoad;

    OSSpinLockUnlock(&(My->lock_CPULoad));

    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_CPULoad,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_StreamFormat: /* 8 */ 

    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    /* reject non-native endian-ness */
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    
    if (ASYS_AUDIOUNIT_FLOAT32_BIGENDIAN != 
	(((((AudioStreamBasicDescription *)((void *)(p->params[1])))->mFormatFlags)
	  & kAudioFormatFlagIsBigEndian) != 0))
      {
	result = kAudioUnitErr_InvalidPropertyValue; 
	break;
      }

    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:

      if (!(ASYS_AUDIOUNIT_EFFECT) || 
	  (p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	  (p->params[2] < 0))
	{
	  result = kAudioUnitErr_InvalidProperty; 
	  break;
	}

      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
      /* if the AU uses an input, accept if it doesn't exceed channel limit */
      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

      if (((AudioStreamBasicDescription *)
	   ((void *)(p->params[1])))->mChannelsPerFrame &&
	  (((AudioStreamBasicDescription *)
	    ((void *)(p->params[1])))->mChannelsPerFrame
	   <= ASYS_AUDIOUNIT_INPUT_MAXCHANNELS))
	{
	  if (p->params[2] == 0)
	    My->AudioBufferTemplate->mNumberBuffers = 
	      ((AudioStreamBasicDescription *)
	       ((void *)(p->params[1])))->mChannelsPerFrame;
	  
	  My->InputStreamFormat[p->params[2]].mChannelsPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mChannelsPerFrame;
	  My->InputStreamFormat[p->params[2]].mSampleRate = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mSampleRate;
	  My->InputStreamFormat[p->params[2]].mFormatID = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatID;
	  My->InputStreamFormat[p->params[2]].mFormatFlags = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatFlags;
	  My->InputStreamFormat[p->params[2]].mBytesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerPacket;
	  My->InputStreamFormat[p->params[2]].mBytesPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerFrame;
	  My->InputStreamFormat[p->params[2]].mFramesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFramesPerPacket;
	  My->InputStreamFormat[p->params[2]].mBitsPerChannel = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBitsPerChannel;
	  My->InputStreamFormat[p->params[2]].mReserved = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mReserved;
	}
      else
	result = kAudioUnitErr_InvalidProperty; 
      break;
    case kAudioUnitScope_Output:

      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
      /* if number of output channels meet our limits, accept */
      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

      if (((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame  
	  && 
	  (((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame 
	   <= ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS))
	{
	  My->OutputStreamFormat.mChannelsPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mChannelsPerFrame;
	  My->OutputStreamFormat.mSampleRate = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mSampleRate;
	  My->OutputStreamFormat.mFormatID = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatID;
	  My->OutputStreamFormat.mFormatFlags = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatFlags;
	  My->OutputStreamFormat.mBytesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerPacket;
	  My->OutputStreamFormat.mBytesPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerFrame;
	  My->OutputStreamFormat.mFramesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFramesPerPacket;
	  My->OutputStreamFormat.mBitsPerChannel = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBitsPerChannel;
	  My->OutputStreamFormat.mReserved = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mReserved;
	}
      else
	result = kAudioUnitErr_InvalidPropertyValue; 
      break;
    case kAudioUnitScope_Global:

      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
      /* if number of output channels meet our limits, accept */
      /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

      if (((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame  
	  && 
	  (((AudioStreamBasicDescription *)((void *)(p->params[1])))->mChannelsPerFrame 
	   <= ASYS_AUDIOUNIT_OUTPUT_MAXCHANNELS))
	{
	  My->OutputStreamFormat.mChannelsPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mChannelsPerFrame;
	  My->OutputStreamFormat.mSampleRate = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mSampleRate;
	  My->OutputStreamFormat.mFormatID = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatID;
	  My->OutputStreamFormat.mFormatFlags = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFormatFlags;
	  My->OutputStreamFormat.mBytesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerPacket;
	  My->OutputStreamFormat.mBytesPerFrame = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBytesPerFrame;
	  My->OutputStreamFormat.mFramesPerPacket = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mFramesPerPacket;
	  My->OutputStreamFormat.mBitsPerChannel = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mBitsPerChannel;
	  My->OutputStreamFormat.mReserved = 
	    ((AudioStreamBasicDescription *)
	     ((void *)(p->params[1])))->mReserved;
	}
      else
	{
	  result = kAudioUnitErr_InvalidPropertyValue; 
	  break;
	}

      if ((ASYS_AUDIOUNIT_EFFECT) && 
	  (p->params[2] < ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	  (p->params[2] >= 0))
	{

	  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	  /* if the AU uses an input, accept if it doesn't exceed channel limit */
	  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
	  
	  if (((AudioStreamBasicDescription *)
	       ((void *)(p->params[1])))->mChannelsPerFrame &&
	      (((AudioStreamBasicDescription *)
		((void *)(p->params[1])))->mChannelsPerFrame
	       <= ASYS_AUDIOUNIT_INPUT_MAXCHANNELS)) 
	    {
	      if (p->params[2] == 0)
		My->AudioBufferTemplate->mNumberBuffers = 
		  ((AudioStreamBasicDescription *)
		   ((void *)(p->params[1])))->mChannelsPerFrame;
	      
	      My->InputStreamFormat[p->params[2]].mChannelsPerFrame = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mChannelsPerFrame;
	      My->InputStreamFormat[p->params[2]].mSampleRate = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mSampleRate;
	      My->InputStreamFormat[p->params[2]].mFormatID = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mFormatID;
	      My->InputStreamFormat[p->params[2]].mFormatFlags = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mFormatFlags;
	      My->InputStreamFormat[p->params[2]].mBytesPerPacket = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mBytesPerPacket;
	      My->InputStreamFormat[p->params[2]].mBytesPerFrame = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mBytesPerFrame;
	      My->InputStreamFormat[p->params[2]].mFramesPerPacket = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mFramesPerPacket;
	      My->InputStreamFormat[p->params[2]].mBitsPerChannel = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mBitsPerChannel;
	      My->InputStreamFormat[p->params[2]].mReserved = 
		((AudioStreamBasicDescription *)
		 ((void *)(p->params[1])))->mReserved;
	    }
	  else
	    result = kAudioUnitErr_InvalidProperty; 
	}
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_SRCAlgorithm: /* 9 */
    My->SRCAlgorithm = *((OSType *)((void *)(p->params[1])));
    break;
  case kAudioUnitProperty_ElementCount: /* 11, also kAudioUnitProperty_BusCount */
    switch (p->params[3] /* inScope */ ) {
    case kAudioUnitScope_Input:
      if (*((UInt32 *)((void *)(p->params[1]))) <= ASYS_AUDIOUNIT_ELEMENT_INPUTMAX)
	My->InputElementCount = *((UInt32 *)((void *)(p->params[1])));
      else
	result = kAudioUnitErr_InvalidPropertyValue; 
      break;
    case kAudioUnitScope_Output:
      if (*((UInt32 *)((void *)(p->params[1]))) <= ASYS_AUDIOUNIT_ELEMENT_OUTPUTMAX)
	My->OutputElementCount = *((UInt32 *)((void *)(p->params[1])));
      else
	result = kAudioUnitErr_InvalidPropertyValue; 
      break;
    case kAudioUnitScope_Global:
      if (*((UInt32 *)((void *)(p->params[1]))) <= ASYS_AUDIOUNIT_ELEMENT_OUTPUTMAX)
	My->OutputElementCount = *((UInt32 *)((void *)(p->params[1])));
      else
	result = kAudioUnitErr_InvalidPropertyValue; 
      if (ASYS_AUDIOUNIT_EFFECT)
	{
	  if (*((UInt32 *)((void *)(p->params[1]))) <= ASYS_AUDIOUNIT_ELEMENT_INPUTMAX)
	    My->InputElementCount = *((UInt32 *)((void *)(p->params[1])));
	  else
	    result = kAudioUnitErr_InvalidPropertyValue;
	}
      break;
    case kAudioUnitScope_Part:
    case kAudioUnitScope_Group:
      break;
    default:
      result = kAudioUnitErr_InvalidProperty; 
      break;
    }
    break;
  case kAudioUnitProperty_Latency: /* 12 */
    My->Latency = *((Float64 *)((void *)(p->params[1])));
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_Latency,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_SupportedNumChannels: /* 13 */
    result = kAudioUnitErr_InvalidPropertyValue;
    break;
  case kAudioUnitProperty_MaximumFramesPerSlice: /* 14 */
    if (*((UInt32 *)((void *)(p->params[1]))) > ASYS_AUDIOUNIT_FRAMES_PER_SLICE)
      result = kAudioUnitErr_InvalidPropertyValue;
    else
      {
	My->MaximumFramesPerSlice = *((UInt32 *)((void *)(p->params[1])));
	asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_MaximumFramesPerSlice,
					     kAudioUnitScope_Global, (AudioUnitElement) 0);
      }
    break;
  case kAudioUnitProperty_ParameterValueStrings: /* 16 */
    result = kAudioUnitErr_InvalidProperty; 
    break;    /* hosts not permitted to set ParameterValueStrings */
  case kAudioUnitProperty_TailTime: /* 20 */
    My->TailTime = *((Float64 *)((void *)(p->params[1])));
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_TailTime,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_BypassEffect: /* 21 */
    My->BypassEffect = *((UInt32 *)((void *)(p->params[1])));
    break;
  case kAudioUnitProperty_LastRenderError: /* 22 */
    My->LastRenderError = *((OSStatus *)((void *)(p->params[1])));
    break;
  case kAudioUnitProperty_SetRenderCallback:  /* 23 */
    if ((p->params[2] >= ASYS_AUDIOUNIT_ELEMENT_ARRAYSIZE) ||
	(p->params[2] < 0))
      {
	result = kAudioUnitErr_InvalidProperty; 
	break;
      }
    OSSpinLockLock(&(My->lock_sampledelivery));
    memcpy(&(My->SetRenderCallback[p->params[2]]), 
	   (AURenderCallbackStruct *)(p->params[1]),
	   sizeof(AURenderCallbackStruct));
    OSSpinLockUnlock(&(My->lock_sampledelivery));
    break;
  case kAudioUnitProperty_FactoryPresets: /* 24 */
    result = kAudioUnitErr_InvalidProperty; 
    break;    /* hosts not permitted to change Factory Presets */
  case kAudioUnitProperty_RenderQuality: /* 26 */
    My->RenderQuality = *((UInt32 *)((void *)(p->params[1])));
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_RenderQuality,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_InPlaceProcessing: /* 29 */
    if (*((UInt32 *)((void *)(p->params[1]))) != ASYS_AUDIOUNIT_INPLACE_PROCESSING)
      result = kAudioUnitErr_InvalidPropertyValue;
    break;
  case kAudioUnitProperty_PresentPreset: /* 36 */
    OSSpinLockLock(&(My->lock_PresentPreset));
    CFRelease(My->PresentPreset.presetName);
    My->PresentPreset.presetNumber = ((AUPreset *)(p->params[1]))->presetNumber;
    My->PresentPreset.presetName = ((AUPreset *)(p->params[1]))->presetName;
    CFRetain(My->PresentPreset.presetName);
    OSSpinLockUnlock(&(My->lock_PresentPreset));
    asysn_audiounit_presentproperty_setfactory(My);
    asysn_audiounit_proplisteners_update(My, kAudioUnitProperty_PresentPreset,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  case kAudioUnitProperty_OfflineRender: /* 37 */
    My->OfflineRender = *((UInt32 *)((void *)(p->params[1]))); 
    break;
  case kAudioUnitProperty_PresentationLatency: /* 40 */
    My->PresentationLatency = *((Float64 *)((void *)(p->params[1]))); 
    break;
  case kMusicDeviceProperty_InstrumentCount: /* 1000 */
    if (ASYS_AUDIOUNIT_EFFECT)
      result = kAudioUnitErr_InvalidPropertyValue;
    else
      if (*((UInt32 *)((void *)(p->params[1]))) != ASYS_AUDIOUNIT_INSTRUMENT_COUNT)
	result = kAudioUnitErr_InvalidPropertyValue;
    break;
  case kMusicDeviceProperty_StreamFromDisk: /* 1011 */
    My->StreamFromDisk = *((UInt32 *)((void *)(p->params[1]))); 
    asysn_audiounit_proplisteners_update(My, kMusicDeviceProperty_StreamFromDisk,
					 kAudioUnitScope_Global, (AudioUnitElement) 0);
    break;
  default:   /* should never happen */
    result = kAudioUnitErr_InvalidProperty; 
    break;
  }

  return result;
}

/**********************************************************/
/*                 Selector: Render                       */
/*                                                        */
/* params[0]: AudioBufferList * ioData;                   */
/* params[1]: UInt32 inNumberFrames;                      */
/* params[2]: UInt32 inOutputBusNumber;                   */
/* params[3]: const AudioTimeStamp * inTimeStamp;         */
/* params[4]: AudioUnitRenderActionFlags * ioActionFlags; */
/*                                                        */
/**********************************************************/

ComponentResult asysn_audiounit_kAudioUnitRenderSelect(ComponentParameters * p, 
					  asysn_audiounit_InstanceState * My)
{ 
  return (asysn_audiounit_MyRenderer((void *) My, 
				     (AudioUnitRenderActionFlags *)(p->params[4]),
				     (const AudioTimeStamp *)(p->params[3]),
				     (UInt32)(p->params[2]), (UInt32)(p->params[1]),
				     (AudioBufferList *)(p->params[0])));
}

/*******************/
/* Selector: Reset */
/*                 */
/*******************/

ComponentResult asysn_audiounit_kAudioUnitResetSelect(ComponentParameters * p, 
					 asysn_audiounit_InstanceState * My)
{ 

#if (ASYS_AUDIOUNIT_RESET_TYPE == ASYS_AUDIOUNIT_ACTIVE_RESET)
  asysn_audiounit_kAudioUnitUninitializeSelect(p, My);
  asysn_audiounit_kAudioUnitInitializeSelect(p, My);
#endif

  return noErr; 
}

/*****************************************************/
/* Selector: Add Property Listener                   */
/*                                                   */
/*  params[0]: void * inProcRefCon;                  */
/*  params[1]: AudioUnitPropertyListenerProc inProc; */
/*  params[2]: AudioUnitPropertyID inID;             */
/*                                                   */
/* The caller passes in a function inProc to call    */
/* the AU changes the property, and an object        */ 
/* inProcRefCon to pass to the function.             */
/*****************************************************/

ComponentResult asysn_audiounit_kAudioUnitAddPropertyListenerSelect(
                                           ComponentParameters * p, 
					   asysn_audiounit_InstanceState * My)
{
  asysn_audiounit_proplisten * lp;
  AudioUnitPropertyListenerProc lproc;
  void * lrefcon;
  int idx, found, msize;

  if ((idx = asysn_audiounit_proplisteners_index(p->params[2])) < 0)
    return noErr; 
  
  lrefcon = (void *)(p->params[0]);
  lproc = (AudioUnitPropertyListenerProc)(p->params[1]);

  if (!lproc)
    return noErr; 

  found = 0;

  OSSpinLockLock(&(My->lock_proplisteners));

  for (lp = My->proplisteners[idx]; lp != NULL; lp = lp->next)
    if ((lp->lrefcon == lrefcon) && (lp->lproc == lproc))
      {
	found = 1;
	break;
      }

  if (!found)
    {
      msize = sizeof(asysn_audiounit_proplisten);
      if (lp = malloc(msize))
	{
	  asysn_audiounit_memstatus(lp, msize, MADV_WILLNEED);
	  lp->lproc = lproc;
	  lp->lrefcon = lrefcon;
	  lp->next = My->proplisteners[idx];
	  My->proplisteners[idx] = lp;
	}
    }

  OSSpinLockUnlock(&(My->lock_proplisteners));

  /* end of critical section */
  
  return noErr; 
}

/*****************************************************/
/* Selector: Remove Property Listener                */
/*                                                   */
/*  params[0]: AudioUnitPropertyListenerProc inProc; */
/*  params[1]: AudioUnitPropertyID inID;             */
/*                                                   */
/* Instructs the AudioUnit to stop calling the       */
/* inProc for the specified property inID.           */
/*****************************************************/

ComponentResult asysn_audiounit_kAudioUnitRemovePropertyListenerSelect(
                                           ComponentParameters * p, 
					   asysn_audiounit_InstanceState * My)
{
  asysn_audiounit_proplisten * lp, * remove, ** trailer;
  AudioUnitPropertyListenerProc lproc;
  int idx, msize;

  if ((idx = asysn_audiounit_proplisteners_index(p->params[1])) < 0)
    return noErr; 

  lproc = (AudioUnitPropertyListenerProc)(p->params[0]);

  OSSpinLockLock(&(My->lock_proplisteners));

  lp = My->proplisteners[idx];
  trailer = &(My->proplisteners[idx]);

  while (lp != NULL)
    if (lp->lproc == lproc)
      {
	remove = lp;
	lp = (*trailer) = lp->next;
	msize = sizeof(asysn_audiounit_proplisten);
	asysn_audiounit_memstatus(remove, msize, MADV_FREE);
 	free(remove);
      }
    else
      {
	trailer = &(lp->next);
	lp = lp->next;
      }

  OSSpinLockUnlock(&(My->lock_proplisteners));

  return noErr; 
}

/*****************************************************/
/* Selector: Remove Property Listener With User Data */
/*                                                   */
/*  params[0]: void * inProcRefCon;                  */
/*  params[1]: AudioUnitPropertyListenerProc inProc; */
/*  params[2]: AudioUnitPropertyID inID;             */
/*                                                   */
/* Instructs the AudioUnit to stop calling the       */
/* inProc for the specified property inID.           */
/*****************************************************/

ComponentResult asysn_audiounit_kAudioUnitRemovePropertyListenerWithUserDataSelect(
                                           ComponentParameters * p, 
					   asysn_audiounit_InstanceState * My)
{
  asysn_audiounit_proplisten * lp, * remove, ** trailer;
  AudioUnitPropertyListenerProc lproc;
  void * lrefcon;
  int idx, msize;

  if ((idx = asysn_audiounit_proplisteners_index(p->params[2])) < 0)
    return noErr; 

  lrefcon = (void *)(p->params[0]);
  lproc = (AudioUnitPropertyListenerProc)(p->params[1]);

  OSSpinLockLock(&(My->lock_proplisteners));

  lp = My->proplisteners[idx];
  trailer = &(My->proplisteners[idx]);

  while (lp != NULL)
    if ((lp->lproc == lproc) && (lp->lrefcon == lrefcon))
      {
	remove = lp;
	lp = (*trailer) = lp->next;
	msize = sizeof(asysn_audiounit_proplisten);
	asysn_audiounit_memstatus(remove, msize, MADV_FREE);
 	free(remove);
      }
    else
      {
	trailer = &(lp->next);
	lp = lp->next;
      }

  OSSpinLockUnlock(&(My->lock_proplisteners));

  return noErr; 
}

/*************************************/
/*  Selector: Add Render Notify      */
/*                                   */
/*  params[0]: void * inProcRefCon;  */
/*  params[1]: ProcPtr inProc;       */
/*                                   */
/*************************************/

ComponentResult asysn_audiounit_kAudioUnitAddRenderNotifySelect
(ComponentParameters * p, asysn_audiounit_InstanceState * My)
{ 
  asysn_audiounit_rendernotify * notify;
  int msize;

  msize = sizeof(asysn_audiounit_rendernotify);
  if (!(notify = malloc(msize)))
    return noErr;
  asysn_audiounit_memstatus(notify, msize, MADV_WILLNEED);

  notify->nproc = (AURenderCallback)(p->params[1]);
  notify->nrefcon = (void *)(p->params[0]);

  OSSpinLockLock(&(My->lock_rendernotify));

  notify->next = My->rendernotify;
  My->rendernotify = notify;

  OSSpinLockUnlock(&(My->lock_rendernotify));

  return noErr;
}

/*************************************/
/*  Selector: Remove Render Notify   */
/*                                   */
/*  params[0]: void * inProcRefCon;  */
/*  params[1]: ProcPtr inProc;       */
/*                                   */
/*************************************/

ComponentResult asysn_audiounit_kAudioUnitRemoveRenderNotifySelect
(ComponentParameters * p, asysn_audiounit_InstanceState * My)
{ 
  asysn_audiounit_rendernotify * notify, * remove, ** trailer;
  int msize;

  OSSpinLockLock(&(My->lock_rendernotify));

  notify = My->rendernotify;
  trailer = &(My->rendernotify);

  while (notify != NULL)
    if ((notify->nproc == (AURenderCallback)(p->params[1])) &&
	(notify->nrefcon == (void *)(p->params[0])))
      {
	remove = notify;
	notify = (*trailer) = notify->next;

	msize = sizeof(asysn_audiounit_rendernotify);
	asysn_audiounit_memstatus(remove, msize, MADV_FREE);
	free(remove);
      }
    else
      {
	trailer = &(notify->next);
	notify = notify->next;
      }

  OSSpinLockUnlock(&(My->lock_rendernotify));

  return noErr;
}


/******************************************/
/*     Selector: GetParameterSelect       */
/*                                        */
/* params[0]: Float32 * outValue          */
/* params[1]: AudioUnitElement inElement  */
/* params[2]: AudioUnitScope inScope      */
/* params[3]: AudioUnitParameterID inID   */
/*                                        */
/******************************************/

ComponentResult asysn_audiounit_kAudioUnitGetParameterSelect
(ComponentParameters * p, asysn_audiounit_InstanceState * My)
{ 
  if (((AudioUnitScope)(p->params[2])) != kAudioUnitScope_Global)  /* inScope */
    return kAudioUnitErr_InvalidScope;

  return asysn_audiounit_getSASLevent(((AudioUnitParameterID)(p->params[3])) /* inID */,
				      ((Float32 *)(p->params[0])) /* outValue */, My);
}


/*********************************************/
/*     Selector: SetParameterSelect          */
/*                                           */
/* params[0]: UInt32 inBufferOffsetInFrames  */
/* params[1]: Float32 inValue                */
/* params[2]: AudioUnitElement inElement     */
/* params[3]: AudioUnitScope inScope;        */
/* params[4]: AudioUnitParameterID inID;     */
/*                                           */
/*********************************************/

ComponentResult asysn_audiounit_kAudioUnitSetParameterSelect
(ComponentParameters * p, asysn_audiounit_InstanceState * My)
{
  asysn_audiounit_SASLevent SASLevent;
  int result;
  Float32 inValue;
  int inID;

  if (((AudioUnitScope)(p->params[3])) != kAudioUnitScope_Global)  /* inScope */
    return kAudioUnitErr_InvalidScope;

  inID = (AudioUnitParameterID)(p->params[4]);

  if ((inID < 0) || (inID >= My->num_saolparams))  /* range-check ID */
    return kAudioUnitErr_InvalidParameter;

  inValue = *((Float32 *)((unsigned char *)(&(p->params[1]))));

  if (inValue > My->parameterinfo[inID].maxValue)
    inValue = My->parameterinfo[inID].maxValue;

  if (inValue < My->parameterinfo[inID].minValue)
    inValue = My->parameterinfo[inID].minValue;

  SASLevent.index = inID;
  SASLevent.value = My->saolparam[inID].value = inValue;
  SASLevent.kcycleidx = (int)((UInt32)(p->params[0]));

  return asysn_audiounit_sendSASLevent(&SASLevent, My);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* The AudioUnit Entry Function */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/************************************************************/
/*   The function Component Manager calls to access the AU  */
/************************************************************/

extern ComponentResult asysn_audiounit_entry(ComponentParameters * p, Handle * obj) 
{ 
  asysn_audiounit_MIDIevent MIDIevent;
  ComponentResult result;

  if (obj == NULL) 
    switch (p->what) {
    case kComponentOpenSelect:
    case kComponentVersionSelect:
    case kComponentCanDoSelect:
      break;
    default:
    return noErr;  /* avoid race condition */
    }

  switch (p->what) { 
  case kComponentOpenSelect:
    result = asysn_audiounit_kComponentOpenSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kComponentCloseSelect:
    result = asysn_audiounit_kComponentCloseSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitInitializeSelect:
    result = asysn_audiounit_kAudioUnitInitializeSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitUninitializeSelect:
    result = asysn_audiounit_kAudioUnitUninitializeSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitGetPropertyInfoSelect:
    result = asysn_audiounit_kAudioUnitGetPropertyInfoSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitGetPropertySelect:
    result = asysn_audiounit_kAudioUnitGetPropertySelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitSetPropertySelect:
    result = asysn_audiounit_kAudioUnitSetPropertySelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitRenderSelect:
    result = asysn_audiounit_kAudioUnitRenderSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitResetSelect:
    result = asysn_audiounit_kAudioUnitResetSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitAddPropertyListenerSelect:
    result = asysn_audiounit_kAudioUnitAddPropertyListenerSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitRemovePropertyListenerSelect:
    result = asysn_audiounit_kAudioUnitRemovePropertyListenerSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitRemovePropertyListenerWithUserDataSelect:
    result = asysn_audiounit_kAudioUnitRemovePropertyListenerWithUserDataSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitAddRenderNotifySelect:
    result = asysn_audiounit_kAudioUnitAddRenderNotifySelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitRemoveRenderNotifySelect:
    result = asysn_audiounit_kAudioUnitRemoveRenderNotifySelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitGetParameterSelect:
    result = asysn_audiounit_kAudioUnitGetParameterSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitSetParameterSelect:
    result = asysn_audiounit_kAudioUnitSetParameterSelect
      (p, (asysn_audiounit_InstanceState *) obj);
    break;
  case kAudioUnitScheduleParametersSelect:
    result = noErr;
    break;
  case kMusicDeviceMIDIEventSelect:
    if (ASYS_AUDIOUNIT_MIDISUPPORT)
      {
	MIDIevent.cmd = (unsigned char)((UInt32)(p->params[3]));
	MIDIevent.d0 = (unsigned char)((UInt32)(p->params[2]));
	MIDIevent.d1 = (unsigned char)((UInt32)(p->params[1]));
	MIDIevent.flags = ASYS_AUDIOUNIT_MIDIFLAGS_WAITING;
	MIDIevent.kcycleidx = (int)((UInt32)(p->params[0]));
	result = asysn_audiounit_sendMIDIevent
	  (&MIDIevent, (asysn_audiounit_InstanceState *) obj);
      }
    else
      result = badComponentSelector;
    break;
  case kMusicDeviceSysExSelect:
  case kMusicDevicePrepareInstrumentSelect:
  case kMusicDeviceReleaseInstrumentSelect:
  case kMusicDeviceStartNoteSelect:
  case kMusicDeviceStopNoteSelect:
  default:
    result = badComponentSelector;
    break;
  case kComponentVersionSelect:
    result = 0x00010000;    /* major version 1, minor version 0 */
    break;
  case kComponentCanDoSelect:
    switch ((int)((SInt16)(p->params[0]))) {
    case kComponentOpenSelect:
    case kComponentCloseSelect:
    case kAudioUnitInitializeSelect:
    case kAudioUnitUninitializeSelect:
    case kAudioUnitGetPropertyInfoSelect:
    case kAudioUnitGetPropertySelect:
    case kAudioUnitSetPropertySelect:
    case kAudioUnitRenderSelect:
    case kAudioUnitResetSelect:
    case kAudioUnitAddPropertyListenerSelect:
    case kAudioUnitRemovePropertyListenerSelect:
    case kAudioUnitRemovePropertyListenerWithUserDataSelect:
    case kAudioUnitAddRenderNotifySelect:
    case kAudioUnitRemoveRenderNotifySelect:
    case kAudioUnitGetParameterSelect:
    case kAudioUnitSetParameterSelect:
      result = 1;
      break;
    case kMusicDeviceMIDIEventSelect:
      result = ASYS_AUDIOUNIT_MIDISUPPORT;
    default:
      result = 0;
      break;
    }
    break;
  }

  switch (p->what) { 
  case kAudioUnitRenderSelect:
    ASYS_AUDIOUNIT_WIRETAP_R(p, result);
    break;
  case kMusicDeviceMIDIEventSelect:
    ASYS_AUDIOUNIT_WIRETAP_M(p, result);
    break;
  default:
    ASYS_AUDIOUNIT_WIRETAP(p, result);
    break;
  }

  return result;
}

/****************************************************************/
/*         End of Part Three (The AudioUnit Component)          */
/****************************************************************/

/****************************************************************/
/*        End of Apple AudioUnit audio driver for sfront        */
/****************************************************************/
/****************************************************************/

#undef ASYS_HASOUTPUT
#undef ASYS_KSYNC


#define CSYS_DONE 0
#define CSYS_ERROR 1

#define CSYS_NONE 0
#define CSYS_SASLEVENTS 2
#define CSYS_EVENTS 3

#define CSYS_SASL_INSTR    0x00
#define CSYS_SASL_CONTROL  0x01
#define CSYS_SASL_TABLE    0x02
#define CSYS_SASL_ENDTIME  0x04
#define CSYS_SASL_TEMPO    0x05
#define CSYS_SASL_NOOP     0x06

#define CSYS_SASL_TGEN_SAMPLE       0x6F
#define CSYS_SASL_TGEN_DATA         0x70
#define CSYS_SASL_TGEN_RANDOM       0x71
#define CSYS_SASL_TGEN_STEP         0x72
#define CSYS_SASL_TGEN_LINESEG      0x73
#define CSYS_SASL_TGEN_EXPSEG       0x74
#define CSYS_SASL_TGEN_CUBICSEG     0x75
#define CSYS_SASL_TGEN_POLYNOMIAL   0x76
#define CSYS_SASL_TGEN_SPLINE       0x77
#define CSYS_SASL_TGEN_WINDOW       0x78
#define CSYS_SASL_TGEN_HARM         0x79
#define CSYS_SASL_TGEN_HARM_PHASE   0x7A
#define CSYS_SASL_TGEN_PERIODIC     0x7B
#define CSYS_SASL_TGEN_BUZZ         0x7C
#define CSYS_SASL_TGEN_CONCAT       0x7D
#define CSYS_SASL_TGEN_EMPTY        0x7E
#define CSYS_SASL_TGEN_DESTROY      0x7F


/*
#    Sfront, a SAOL to C translator    
#    This file: aucontrol control driver for sfront
#
# Copyright (c) 1999-2006, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  Redistributions of source code must retain the above copyright
#  notice, this list of conditions and the following disclaimer.
#
#  Redistributions in binary form must reproduce the above copyright
#  notice, this list of conditions and the following disclaimer in the
#  documentation and/or other materials provided with the distribution.
#
#  Neither the name of the University of California, Berkeley nor the
#  names of its contributors may be used to endorse or promote products
#  derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#    Maintainer: John Lazzaro, lazzaro@cs.berkeley.edu
*/

/****************************************************************/
/****************************************************************/
/*               audiounit control driver for sfront            */ 
/****************************************************************/

/*~~~~~~~~~~~~~~~~~*/
/* include headers */
/*_________________*/

/* for socket system */

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/uio.h>
#include <fcntl.h>

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*        MIDI event constants       */
/*  Must match audiounit.c versions  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* retry limit for socket writing */

#define CSYSI_AUCONTROL_RETRY_MAX  256

/* bitfield constants for MIDIevent flags variable */

#define CSYSI_AUCONTROL_MIDIFLAGS_WAITING 0x01u  /* queuing flag bit */

/* bitfield constants for SASLevent flags variable */

#define CSYSI_AUCONTROL_SASLFLAGS_WAITING 0x01u  /* queuing flag bit */

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   typedef for MIDI events   */
/*  Fields & order must match  */
/*  asysn_audiounit_MIDIevent  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct csysi_aucontrol_MIDIevent {
  unsigned char cmd;
  unsigned char d0;
  unsigned char d1;
  unsigned char flags;
  int kcycleidx;
} csysi_aucontrol_MIDIevent;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   typedef for SASL events   */
/*  Fields & order must match  */
/*  asysn_audiounit_SASLevent  */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct csysi_aucontrol_SASLevent {
  int index;
  Float32 value;
  unsigned char flags;
  int kcycleidx;
} csysi_aucontrol_SASLevent;

/*~~~~~~~~~~~~~~~~~*/
/* aucontrol state */
/*~~~~~~~~~~~~~~~~~*/

typedef struct csysi_aucontrol_state {
  int mpipe;
  csysi_aucontrol_MIDIevent nextMIDIevent;
  int spipe;
  csysi_aucontrol_SASLevent nextSASLevent;
} csysi_aucontrol_state;

/*~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* helper function externs */
/*_________________________*/

extern int csysi_aucontrol_midievent_read(csysi_aucontrol_state * mystate);
extern int csysi_aucontrol_saslevent_read(csysi_aucontrol_state * mystate);

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*     high-level functions: called by sfront engine            */
/*______________________________________________________________*/

/****************************************************************/
/*             initialization routine for control               */
/****************************************************************/

int csys_setup(ENGINE_PTR_DECLARE)
     
{
  csysi_aucontrol_state * mystate;
  int msize;
  char message[256];

  msize = sizeof(csysi_aucontrol_state);
  if (!(mystate = calloc(1, msize)))
    return CSYS_ERROR;
  asysn_audiounit_memstatus(mystate, msize, MADV_WILLNEED);

  EV(csys_state) = (void *) mystate;

  if ((EV(csys_argc) == 5) && 
      !strcmp(EV(csys_argv)[1], "-asys_audiounit_mpipe") &&
      !strcmp(EV(csys_argv)[3], "-asys_audiounit_spipe"))
    {
      mystate->mpipe = atoi(EV(csys_argv[2]));
      mystate->spipe = atoi(EV(csys_argv[4]));
    }

  sprintf(message, "Opening aucontrol driver for %s\n\n", 
	  (EV(csys_argc) >= 1) ? EV(csys_argv)[0] : "(unknown)");

  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);

  return CSYS_DONE;
}

/****************************************************************/
/*             polling routine for new data                     */
/****************************************************************/

int csys_newdata(ENGINE_PTR_DECLARE)
     
{
  csysi_aucontrol_state * mystate = (csysi_aucontrol_state *) EV(csys_state);
  int has_midi, has_sasl;

#if defined(CSYS_CDRIVER_AUCONTROLM)
  has_midi = (mystate->nextMIDIevent.flags & CSYSI_AUCONTROL_MIDIFLAGS_WAITING) ||
    (csysi_aucontrol_midievent_read(mystate) == CSYS_MIDIEVENTS);
#else
  has_midi = 0;
#endif

  has_sasl = (mystate->nextSASLevent.flags & CSYSI_AUCONTROL_SASLFLAGS_WAITING) ||
    (csysi_aucontrol_saslevent_read(mystate) == CSYS_SASLEVENTS);

  if (!(has_midi || has_sasl))    /* quick exit in common case */
    return CSYS_NONE;

  has_midi = has_midi && (mystate->nextMIDIevent.kcycleidx <= EV(kcycleidx));
  has_sasl = has_sasl && (mystate->nextSASLevent.kcycleidx <= EV(kcycleidx));

  if (!(has_midi || has_sasl))
    return CSYS_NONE;

  if (has_midi && has_sasl)
    return CSYS_EVENTS;

#if defined(CSYS_CDRIVER_AUCONTROLM)
  if (has_midi)
    return CSYS_MIDIEVENTS;
#endif

  if (has_sasl)
    return CSYS_SASLEVENTS;
}


#if defined(CSYS_CDRIVER_AUCONTROLM)

/****************************************************************/
/*                 processes a MIDI event                       */
/****************************************************************/

int csys_midievent(ENGINE_PTR_DECLARE_COMMA unsigned char * cmd,  
		   unsigned char * ndata, unsigned char * vdata, 
		   unsigned short * extchan, float * fval)

{
  csysi_aucontrol_state * mystate = (csysi_aucontrol_state *) EV(csys_state);

  if ((*cmd = mystate->nextMIDIevent.cmd) < CSYS_MIDI_SYSTEM)
    {
      *extchan = (*cmd) & 0x0Fu;  
      *ndata = mystate->nextMIDIevent.d0;
      *vdata = mystate->nextMIDIevent.d1;
    }
  else
    *cmd = CSYS_MIDI_NOOP;  /* filter MIDI System commands */

  if (csysi_aucontrol_midievent_read(mystate) == CSYS_NONE)
    {
      mystate->nextMIDIevent.flags &= ~(CSYSI_AUCONTROL_MIDIFLAGS_WAITING);
      return CSYS_NONE;
    }

  if (mystate->nextMIDIevent.kcycleidx <= EV(kcycleidx))
    return CSYS_MIDIEVENTS;

  return CSYS_NONE;
}

#endif

/****************************************************************/
/*                 processes a SASL event                       */
/****************************************************************/

int csys_saslevent(ENGINE_PTR_DECLARE_COMMA unsigned char * cmd, 
		   unsigned char * priority, unsigned short * id, 
		   unsigned short * label, float * fval, 
		   unsigned int * pnum, float ** p)

{
  csysi_aucontrol_state * mystate = (csysi_aucontrol_state *) EV(csys_state);

  *cmd = CSYS_SASL_CONTROL;
  *priority = 1;
  *id = CSYS_SASL_NOINSTR;
  *label = CSYS_NOLABEL;
  *pnum = mystate->nextSASLevent.index;
  *fval = mystate->nextSASLevent.value;
  
  if (csysi_aucontrol_saslevent_read(mystate) == CSYS_NONE)
    {
      mystate->nextSASLevent.flags &= ~(CSYSI_AUCONTROL_SASLFLAGS_WAITING);
      return CSYS_NONE;
    }

  if (mystate->nextSASLevent.kcycleidx <= EV(kcycleidx))
    return CSYS_SASLEVENTS;

  return CSYS_NONE;
}


/****************************************************************/
/*                  closing routine for control                 */
/****************************************************************/

void csys_shutdown(ENGINE_PTR_DECLARE)
     
{
  csysi_aucontrol_state * mystate = (csysi_aucontrol_state *) EV(csys_state);
  int msize;
  char message[256];

  /* flush unprocessed events from MIDI and SASL pipes */

#if defined(CSYS_CDRIVER_AUCONTROLM)
  do { } while (csysi_aucontrol_midievent_read(mystate) != CSYS_NONE);
#endif
  do { } while (csysi_aucontrol_saslevent_read(mystate) != CSYS_NONE);

  /* log driver close, and free mystate */

  sprintf(message, "Closing aucontrol driver for %s\n\n",
	  (EV(csys_argc) >= 1) ? EV(csys_argv)[0] : "(unknown)");

  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING(message);

  msize = sizeof(csysi_aucontrol_state);
  asysn_audiounit_memstatus(mystate, msize, MADV_FREE);
  free(mystate);

  return;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                     helper functions                         */
/*______________________________________________________________*/

#if defined(CSYS_CDRIVER_AUCONTROLM)

/****************************************************************/
/*              read a MIDI event from the socket               */
/****************************************************************/

int csysi_aucontrol_midievent_read(csysi_aucontrol_state * mystate)

{
  int retry = 0;
  int len;

  if (mystate->mpipe == 0)
      return CSYS_NONE;   /* no mpipe, so no MIDIevents */

  do {

    if (((len = read(mystate->mpipe, &(mystate->nextMIDIevent),
		     sizeof(csysi_aucontrol_MIDIevent))) < 0) && (errno == EAGAIN))
      return CSYS_NONE;        /* no MIDIevents in mpipe */

    if (len == sizeof(csysi_aucontrol_MIDIevent))
      return CSYS_MIDIEVENTS;  /* a MIDIevent was read  */

    if ((len >= 0) || (errno != EINTR) || (++retry > CSYSI_AUCONTROL_RETRY_MAX))
      {
	if (len == 0)
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: read() returned a zero-length MIDIevent\n");

	if (len > 0)
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: read() returned an incomplete MIDIevent\n");
	
	if ((len < 0) && (errno != EINTR))
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: errno other than EINTR or EAGAIN (MIDI)\n");

	if ((len < 0) && (retry > ASYS_AUDIOUNIT_RETRY_MAX))
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: Maximum retry for EINTR exceeded (MIDI)\n");

	return CSYS_NONE;     /* to do: clear flag, consider error reporting */
      }
    
  } while(1);   /* loop to try again when (errno == EINTR) */

  return CSYS_NONE;
}

#endif 

/****************************************************************/
/*              read a SASL event from the socket               */
/****************************************************************/

int csysi_aucontrol_saslevent_read(csysi_aucontrol_state * mystate)

{
  int retry = 0;
  int len;

  if (mystate->spipe == 0)
      return CSYS_NONE;   /* no spipe, so no SASLevents */

  do {

    if (((len = read(mystate->spipe, &(mystate->nextSASLevent),
		     sizeof(csysi_aucontrol_SASLevent))) < 0) && (errno == EAGAIN))
      return CSYS_NONE;        /* no SASLevents in spipe */

    if (len == sizeof(csysi_aucontrol_SASLevent))
      return CSYS_SASLEVENTS;  /* a SASLevent was read  */

    if ((len >= 0) || (errno != EINTR) || (++retry > CSYSI_AUCONTROL_RETRY_MAX))
      {
	if (len == 0)
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: read() returned a zero-length SASLevent\n");

	if (len > 0)
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: read() returned an incomplete SASLevent\n");
	
	if ((len < 0) && (errno != EINTR))
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: errno other than EINTR or EAGAIN (SASL)\n");

	if ((len < 0) && (retry > ASYS_AUDIOUNIT_RETRY_MAX))
	  ASYS_AUDIOUNIT_WIRETAP_PUTSTRING
	    ("\tError in csys_newdata: Maximum retry for EINTR exceeded (SASL)\n");

	return CSYS_NONE;     /* to do: clear flag, consider error reporting */
      }
    
  } while(1);   /* loop to try again when (errno == EINTR) */

  return CSYS_NONE;
}


/* handles termination in case of error */

void epr(int linenum, char * filename, char * token, char * message)

{


#if (ASYS_AUDIOUNIT_DEBUG_LEVEL >= 1)
  asysn_audiounit_logfile = fopen(ASYS_AUDIOUNIT_LOGFILE_NAME, "a");
  fprintf(asysn_audiounit_logfile, "\nRuntime Error.\n");
  if (filename != NULL)
    fprintf(asysn_audiounit_logfile, "Location: File %s near line %i.\n",filename, linenum);
  if (token != NULL)
    fprintf(asysn_audiounit_logfile, "While executing: %s.\n",token);
  if (message != NULL)
    fprintf(asysn_audiounit_logfile, "Potential problem: %s.\n",message);
  fprintf(asysn_audiounit_logfile, "Exiting program.\n\n");
  exit(-1);
  fclose(asysn_audiounit_logfile);
#endif /* (ASYS_AUDIOUNIT_DEBUG_LEVEL >= 1) */

}



float centroid__sym_rms4(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid_rms4_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid_rms4_buffer).len;i++)
      ret +=  NT(TBL_centroid_rms4_buffer).t[i]*NT(TBL_centroid_rms4_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid_rms4_scale));
   }
   else
   {
       i = NT(TBL_centroid_rms4_buffer).len = EV(ACYCLE);
    NT(TBL_centroid_rms4_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid_rms4_buffer).llmem=1;
    NV(centroid_rms4_scale) = 1.0F/NT(TBL_centroid_rms4_buffer).len;
    NVI(centroid_rms4_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid_rms4_return) = ret));

}



void centroid__sym_rms4_spec(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   x = NV(centroid_in_perceptual);
     i = NT(TBL_centroid_rms4_buffer).tend++;
     if (NT(TBL_centroid_rms4_buffer).tend == NT(TBL_centroid_rms4_buffer).len)
       NT(TBL_centroid_rms4_buffer).tend = 0;
     NT(TBL_centroid_rms4_buffer).t[i] = x;

}



float centroid__sym_rms5(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid_rms5_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid_rms5_buffer).len;i++)
      ret +=  NT(TBL_centroid_rms5_buffer).t[i]*NT(TBL_centroid_rms5_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid_rms5_scale));
   }
   else
   {
       i = NT(TBL_centroid_rms5_buffer).len = EV(ACYCLE);
    NT(TBL_centroid_rms5_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid_rms5_buffer).llmem=1;
    NV(centroid_rms5_scale) = 1.0F/NT(TBL_centroid_rms5_buffer).len;
    NVI(centroid_rms5_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid_rms5_return) = ret));

}



void centroid__sym_rms5_spec(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   x = NV(centroid_lp_perceptual);
     i = NT(TBL_centroid_rms5_buffer).tend++;
     if (NT(TBL_centroid_rms5_buffer).tend == NT(TBL_centroid_rms5_buffer).len)
       NT(TBL_centroid_rms5_buffer).tend = 0;
     NT(TBL_centroid_rms5_buffer).t[i] = x;

}



float centroid__sym_rms6(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid_rms6_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid_rms6_buffer).len;i++)
      ret +=  NT(TBL_centroid_rms6_buffer).t[i]*NT(TBL_centroid_rms6_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid_rms6_scale));
   }
   else
   {
       i = NT(TBL_centroid_rms6_buffer).len = EV(ACYCLE);
    NT(TBL_centroid_rms6_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid_rms6_buffer).llmem=1;
    NV(centroid_rms6_scale) = 1.0F/NT(TBL_centroid_rms6_buffer).len;
    NVI(centroid_rms6_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid_rms6_return) = ret));

}



void centroid__sym_rms6_spec(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;
   int i;

   x = NV(centroid_out_perceptual);
     i = NT(TBL_centroid_rms6_buffer).tend++;
     if (NT(TBL_centroid_rms6_buffer).tend == NT(TBL_centroid_rms6_buffer).len)
       NT(TBL_centroid_rms6_buffer).tend = 0;
     NT(TBL_centroid_rms6_buffer).t[i] = x;

}



float centroid__sym_sgn14(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;

   x = NV(centroid_out_db) - NV(centroid_in_db);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid_sgn14_return) = ret));

}



float centroid__sym_sgn21(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;

   x = NV(centroid_state);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid_sgn21_return) = ret));

}



float centroid__sym_sgn23(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float x;
   float ret;

   x = NV(centroid_tilt_db);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid_sgn23_return) = ret));

}



float centroid__sym_lopass44(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid_in_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_lopass44_b0)*input + NV(centroid_lopass44_d2);
      NV(centroid_lopass44_d2) = NV(centroid_lopass44_d1)-NV(centroid_lopass44_a1)*ret+NV(centroid_lopass44_b1)*input;
      NV(centroid_lopass44_d1) =            -NV(centroid_lopass44_a2)*ret+NV(centroid_lopass44_b2)*input;
    }
   else
    {
     if ((NVI(centroid_lopass44_kcyc)==0)||(NV(centroid_lopass44_ocut)!=cut))
      {
       NVI(centroid_lopass44_kcyc) = 1;
       NV(centroid_lopass44_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid_lopass44_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid_lopass44_b1)= 2.0F*NV(centroid_lopass44_b0);
       NV(centroid_lopass44_b2)= NV(centroid_lopass44_b0);
       NV(centroid_lopass44_a1)= 2.0F*NV(centroid_lopass44_b0)*(1.0F - c*c);
       NV(centroid_lopass44_a2)= NV(centroid_lopass44_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_lopass44_b0)*input + NV(centroid_lopass44_d2);
     NV(centroid_lopass44_d2) = NV(centroid_lopass44_d1)-NV(centroid_lopass44_a1)*ret+NV(centroid_lopass44_b1)*input;
     NV(centroid_lopass44_d1) =            -NV(centroid_lopass44_a2)*ret+NV(centroid_lopass44_b2)*input;
    }
   return((NV(centroid_lopass44_return) = ret));

}



float centroid__sym_hipass45(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid__sym_lopass44(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_hipass45_b0)*input + NV(centroid_hipass45_d2);
      NV(centroid_hipass45_d2) = NV(centroid_hipass45_d1)-NV(centroid_hipass45_a1)*ret+NV(centroid_hipass45_b1)*input;
      NV(centroid_hipass45_d1) =            -NV(centroid_hipass45_a2)*ret+NV(centroid_hipass45_b2)*input;
    }
   else
    {
     if ((NVI(centroid_hipass45_kcyc)==0)||(NV(centroid_hipass45_ocut)!=cut))
      {
       NVI(centroid_hipass45_kcyc) = 1;
        NV(centroid_hipass45_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid_hipass45_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid_hipass45_b1)= -2.0F*NV(centroid_hipass45_b0);
        NV(centroid_hipass45_b2)= NV(centroid_hipass45_b0);
        NV(centroid_hipass45_a1)= 2.0F*NV(centroid_hipass45_b0)*(c*c - 1.0F);
        NV(centroid_hipass45_a2)= NV(centroid_hipass45_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_hipass45_b0)*input + NV(centroid_hipass45_d2);
     NV(centroid_hipass45_d2) = NV(centroid_hipass45_d1)-NV(centroid_hipass45_a1)*ret+NV(centroid_hipass45_b1)*input;
     NV(centroid_hipass45_d1) =            -NV(centroid_hipass45_a2)*ret+NV(centroid_hipass45_b2)*input;
    }
   return((NV(centroid_hipass45_return) = ret));

}



float centroid__sym_lopass46(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid_lp_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_lopass46_b0)*input + NV(centroid_lopass46_d2);
      NV(centroid_lopass46_d2) = NV(centroid_lopass46_d1)-NV(centroid_lopass46_a1)*ret+NV(centroid_lopass46_b1)*input;
      NV(centroid_lopass46_d1) =            -NV(centroid_lopass46_a2)*ret+NV(centroid_lopass46_b2)*input;
    }
   else
    {
     if ((NVI(centroid_lopass46_kcyc)==0)||(NV(centroid_lopass46_ocut)!=cut))
      {
       NVI(centroid_lopass46_kcyc) = 1;
       NV(centroid_lopass46_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid_lopass46_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid_lopass46_b1)= 2.0F*NV(centroid_lopass46_b0);
       NV(centroid_lopass46_b2)= NV(centroid_lopass46_b0);
       NV(centroid_lopass46_a1)= 2.0F*NV(centroid_lopass46_b0)*(1.0F - c*c);
       NV(centroid_lopass46_a2)= NV(centroid_lopass46_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_lopass46_b0)*input + NV(centroid_lopass46_d2);
     NV(centroid_lopass46_d2) = NV(centroid_lopass46_d1)-NV(centroid_lopass46_a1)*ret+NV(centroid_lopass46_b1)*input;
     NV(centroid_lopass46_d1) =            -NV(centroid_lopass46_a2)*ret+NV(centroid_lopass46_b2)*input;
    }
   return((NV(centroid_lopass46_return) = ret));

}



float centroid__sym_hipass47(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid__sym_lopass46(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_hipass47_b0)*input + NV(centroid_hipass47_d2);
      NV(centroid_hipass47_d2) = NV(centroid_hipass47_d1)-NV(centroid_hipass47_a1)*ret+NV(centroid_hipass47_b1)*input;
      NV(centroid_hipass47_d1) =            -NV(centroid_hipass47_a2)*ret+NV(centroid_hipass47_b2)*input;
    }
   else
    {
     if ((NVI(centroid_hipass47_kcyc)==0)||(NV(centroid_hipass47_ocut)!=cut))
      {
       NVI(centroid_hipass47_kcyc) = 1;
        NV(centroid_hipass47_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid_hipass47_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid_hipass47_b1)= -2.0F*NV(centroid_hipass47_b0);
        NV(centroid_hipass47_b2)= NV(centroid_hipass47_b0);
        NV(centroid_hipass47_a1)= 2.0F*NV(centroid_hipass47_b0)*(c*c - 1.0F);
        NV(centroid_hipass47_a2)= NV(centroid_hipass47_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_hipass47_b0)*input + NV(centroid_hipass47_d2);
     NV(centroid_hipass47_d2) = NV(centroid_hipass47_d1)-NV(centroid_hipass47_a1)*ret+NV(centroid_hipass47_b1)*input;
     NV(centroid_hipass47_d1) =            -NV(centroid_hipass47_a2)*ret+NV(centroid_hipass47_b2)*input;
    }
   return((NV(centroid_hipass47_return) = ret));

}



float centroid__sym_lopass48(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid_out_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_lopass48_b0)*input + NV(centroid_lopass48_d2);
      NV(centroid_lopass48_d2) = NV(centroid_lopass48_d1)-NV(centroid_lopass48_a1)*ret+NV(centroid_lopass48_b1)*input;
      NV(centroid_lopass48_d1) =            -NV(centroid_lopass48_a2)*ret+NV(centroid_lopass48_b2)*input;
    }
   else
    {
     if ((NVI(centroid_lopass48_kcyc)==0)||(NV(centroid_lopass48_ocut)!=cut))
      {
       NVI(centroid_lopass48_kcyc) = 1;
       NV(centroid_lopass48_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid_lopass48_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid_lopass48_b1)= 2.0F*NV(centroid_lopass48_b0);
       NV(centroid_lopass48_b2)= NV(centroid_lopass48_b0);
       NV(centroid_lopass48_a1)= 2.0F*NV(centroid_lopass48_b0)*(1.0F - c*c);
       NV(centroid_lopass48_a2)= NV(centroid_lopass48_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_lopass48_b0)*input + NV(centroid_lopass48_d2);
     NV(centroid_lopass48_d2) = NV(centroid_lopass48_d1)-NV(centroid_lopass48_a1)*ret+NV(centroid_lopass48_b1)*input;
     NV(centroid_lopass48_d1) =            -NV(centroid_lopass48_a2)*ret+NV(centroid_lopass48_b2)*input;
    }
   return((NV(centroid_lopass48_return) = ret));

}



float centroid__sym_hipass49(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid__sym_lopass48(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid_hipass49_b0)*input + NV(centroid_hipass49_d2);
      NV(centroid_hipass49_d2) = NV(centroid_hipass49_d1)-NV(centroid_hipass49_a1)*ret+NV(centroid_hipass49_b1)*input;
      NV(centroid_hipass49_d1) =            -NV(centroid_hipass49_a2)*ret+NV(centroid_hipass49_b2)*input;
    }
   else
    {
     if ((NVI(centroid_hipass49_kcyc)==0)||(NV(centroid_hipass49_ocut)!=cut))
      {
       NVI(centroid_hipass49_kcyc) = 1;
        NV(centroid_hipass49_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid_hipass49_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid_hipass49_b1)= -2.0F*NV(centroid_hipass49_b0);
        NV(centroid_hipass49_b2)= NV(centroid_hipass49_b0);
        NV(centroid_hipass49_a1)= 2.0F*NV(centroid_hipass49_b0)*(c*c - 1.0F);
        NV(centroid_hipass49_a2)= NV(centroid_hipass49_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid_hipass49_b0)*input + NV(centroid_hipass49_d2);
     NV(centroid_hipass49_d2) = NV(centroid_hipass49_d1)-NV(centroid_hipass49_a1)*ret+NV(centroid_hipass49_b1)*input;
     NV(centroid_hipass49_d1) =            -NV(centroid_hipass49_a2)*ret+NV(centroid_hipass49_b2)*input;
    }
   return((NV(centroid_hipass49_return) = ret));

}


#undef NS
#define NS(x)  EV(ninstr)[0].x
#undef NSP
#define NSP  ENGINE_PTR 
#undef NT
#define NT(x)  EV(ninstr)[0].t[x]
#undef NV
#define NV(x)  EV(ninstr)[0].v[x].f
#undef NVI
#define NVI(x)  EV(ninstr)[0].v[x].i
#undef NVUI
#define NVUI(x)  EV(ninstr)[0].v[x].ui
#undef NVPS
#define NVPS(x)  EV(ninstr)[0].v[x].ps
#undef NVU
#define NVU(x)  EV(ninstr)[0].v[x]
#undef NP
#define NP(x)  EV(ninstr)[0].v[x].f
#undef NPI
#define NPI(x)  EV(ninstr)[0].v[x].i
#undef NPUI
#define NPUI(x)  EV(ninstr)[0].v[x].ui
#undef NPPS
#define NPPS(x)  EV(ninstr)[0].v[x].ps



float centroid1__sym_rms4(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid1_rms4_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid1_rms4_buffer).len;i++)
      ret +=  NT(TBL_centroid1_rms4_buffer).t[i]*NT(TBL_centroid1_rms4_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid1_rms4_scale));
   }
   else
   {
       i = NT(TBL_centroid1_rms4_buffer).len = EV(ACYCLE);
    NT(TBL_centroid1_rms4_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid1_rms4_buffer).llmem=1;
    NV(centroid1_rms4_scale) = 1.0F/NT(TBL_centroid1_rms4_buffer).len;
    NVI(centroid1_rms4_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid1_rms4_return) = ret));

}



void centroid1__sym_rms4_spec(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   x = NV(centroid1_in_perceptual);
     i = NT(TBL_centroid1_rms4_buffer).tend++;
     if (NT(TBL_centroid1_rms4_buffer).tend == NT(TBL_centroid1_rms4_buffer).len)
       NT(TBL_centroid1_rms4_buffer).tend = 0;
     NT(TBL_centroid1_rms4_buffer).t[i] = x;

}



float centroid1__sym_rms5(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid1_rms5_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid1_rms5_buffer).len;i++)
      ret +=  NT(TBL_centroid1_rms5_buffer).t[i]*NT(TBL_centroid1_rms5_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid1_rms5_scale));
   }
   else
   {
       i = NT(TBL_centroid1_rms5_buffer).len = EV(ACYCLE);
    NT(TBL_centroid1_rms5_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid1_rms5_buffer).llmem=1;
    NV(centroid1_rms5_scale) = 1.0F/NT(TBL_centroid1_rms5_buffer).len;
    NVI(centroid1_rms5_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid1_rms5_return) = ret));

}



void centroid1__sym_rms5_spec(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   x = NV(centroid1_lp_perceptual);
     i = NT(TBL_centroid1_rms5_buffer).tend++;
     if (NT(TBL_centroid1_rms5_buffer).tend == NT(TBL_centroid1_rms5_buffer).len)
       NT(TBL_centroid1_rms5_buffer).tend = 0;
     NT(TBL_centroid1_rms5_buffer).t[i] = x;

}



float centroid1__sym_rms6(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   if (NVI(centroid1_rms6_kcyc))
   {
     ret = 0.0F;
     for (i=0; i < NT(TBL_centroid1_rms6_buffer).len;i++)
      ret +=  NT(TBL_centroid1_rms6_buffer).t[i]*NT(TBL_centroid1_rms6_buffer).t[i];
     ret = (float)sqrt(ret*NV(centroid1_rms6_scale));
   }
   else
   {
       i = NT(TBL_centroid1_rms6_buffer).len = EV(ACYCLE);
    NT(TBL_centroid1_rms6_buffer).t=(float *)calloc(i,sizeof(float));
    NT(TBL_centroid1_rms6_buffer).llmem=1;
    NV(centroid1_rms6_scale) = 1.0F/NT(TBL_centroid1_rms6_buffer).len;
    NVI(centroid1_rms6_kcyc) = EV(kcycleidx);
    ret = 0.0F;
   }
   return((NV(centroid1_rms6_return) = ret));

}



void centroid1__sym_rms6_spec(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;
   int i;

   x = NV(centroid1_out_perceptual);
     i = NT(TBL_centroid1_rms6_buffer).tend++;
     if (NT(TBL_centroid1_rms6_buffer).tend == NT(TBL_centroid1_rms6_buffer).len)
       NT(TBL_centroid1_rms6_buffer).tend = 0;
     NT(TBL_centroid1_rms6_buffer).t[i] = x;

}



float centroid1__sym_sgn14(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;

   x = NV(centroid1_out_db) - NV(centroid1_in_db);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid1_sgn14_return) = ret));

}



float centroid1__sym_sgn21(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;

   x = NV(centroid1_state);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid1_sgn21_return) = ret));

}



float centroid1__sym_sgn23(ENGINE_PTR_DECLARE)
{
   float x;
   float ret;

   x = NV(centroid1_tilt_db);
   ret = (x == 0.0F) ? 0.0F : ((x>0.0F) ? 1.0F : -1.0F);
   return((NV(centroid1_sgn23_return) = ret));

}



float centroid1__sym_lopass44(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid1_in_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_lopass44_b0)*input + NV(centroid1_lopass44_d2);
      NV(centroid1_lopass44_d2) = NV(centroid1_lopass44_d1)-NV(centroid1_lopass44_a1)*ret+NV(centroid1_lopass44_b1)*input;
      NV(centroid1_lopass44_d1) =            -NV(centroid1_lopass44_a2)*ret+NV(centroid1_lopass44_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_lopass44_kcyc)==0)||(NV(centroid1_lopass44_ocut)!=cut))
      {
       NVI(centroid1_lopass44_kcyc) = 1;
       NV(centroid1_lopass44_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid1_lopass44_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid1_lopass44_b1)= 2.0F*NV(centroid1_lopass44_b0);
       NV(centroid1_lopass44_b2)= NV(centroid1_lopass44_b0);
       NV(centroid1_lopass44_a1)= 2.0F*NV(centroid1_lopass44_b0)*(1.0F - c*c);
       NV(centroid1_lopass44_a2)= NV(centroid1_lopass44_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_lopass44_b0)*input + NV(centroid1_lopass44_d2);
     NV(centroid1_lopass44_d2) = NV(centroid1_lopass44_d1)-NV(centroid1_lopass44_a1)*ret+NV(centroid1_lopass44_b1)*input;
     NV(centroid1_lopass44_d1) =            -NV(centroid1_lopass44_a2)*ret+NV(centroid1_lopass44_b2)*input;
    }
   return((NV(centroid1_lopass44_return) = ret));

}



float centroid1__sym_hipass45(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid1__sym_lopass44(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_hipass45_b0)*input + NV(centroid1_hipass45_d2);
      NV(centroid1_hipass45_d2) = NV(centroid1_hipass45_d1)-NV(centroid1_hipass45_a1)*ret+NV(centroid1_hipass45_b1)*input;
      NV(centroid1_hipass45_d1) =            -NV(centroid1_hipass45_a2)*ret+NV(centroid1_hipass45_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_hipass45_kcyc)==0)||(NV(centroid1_hipass45_ocut)!=cut))
      {
       NVI(centroid1_hipass45_kcyc) = 1;
        NV(centroid1_hipass45_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid1_hipass45_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid1_hipass45_b1)= -2.0F*NV(centroid1_hipass45_b0);
        NV(centroid1_hipass45_b2)= NV(centroid1_hipass45_b0);
        NV(centroid1_hipass45_a1)= 2.0F*NV(centroid1_hipass45_b0)*(c*c - 1.0F);
        NV(centroid1_hipass45_a2)= NV(centroid1_hipass45_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_hipass45_b0)*input + NV(centroid1_hipass45_d2);
     NV(centroid1_hipass45_d2) = NV(centroid1_hipass45_d1)-NV(centroid1_hipass45_a1)*ret+NV(centroid1_hipass45_b1)*input;
     NV(centroid1_hipass45_d1) =            -NV(centroid1_hipass45_a2)*ret+NV(centroid1_hipass45_b2)*input;
    }
   return((NV(centroid1_hipass45_return) = ret));

}



float centroid1__sym_lopass46(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid1_lp_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_lopass46_b0)*input + NV(centroid1_lopass46_d2);
      NV(centroid1_lopass46_d2) = NV(centroid1_lopass46_d1)-NV(centroid1_lopass46_a1)*ret+NV(centroid1_lopass46_b1)*input;
      NV(centroid1_lopass46_d1) =            -NV(centroid1_lopass46_a2)*ret+NV(centroid1_lopass46_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_lopass46_kcyc)==0)||(NV(centroid1_lopass46_ocut)!=cut))
      {
       NVI(centroid1_lopass46_kcyc) = 1;
       NV(centroid1_lopass46_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid1_lopass46_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid1_lopass46_b1)= 2.0F*NV(centroid1_lopass46_b0);
       NV(centroid1_lopass46_b2)= NV(centroid1_lopass46_b0);
       NV(centroid1_lopass46_a1)= 2.0F*NV(centroid1_lopass46_b0)*(1.0F - c*c);
       NV(centroid1_lopass46_a2)= NV(centroid1_lopass46_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_lopass46_b0)*input + NV(centroid1_lopass46_d2);
     NV(centroid1_lopass46_d2) = NV(centroid1_lopass46_d1)-NV(centroid1_lopass46_a1)*ret+NV(centroid1_lopass46_b1)*input;
     NV(centroid1_lopass46_d1) =            -NV(centroid1_lopass46_a2)*ret+NV(centroid1_lopass46_b2)*input;
    }
   return((NV(centroid1_lopass46_return) = ret));

}



float centroid1__sym_hipass47(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid1__sym_lopass46(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_hipass47_b0)*input + NV(centroid1_hipass47_d2);
      NV(centroid1_hipass47_d2) = NV(centroid1_hipass47_d1)-NV(centroid1_hipass47_a1)*ret+NV(centroid1_hipass47_b1)*input;
      NV(centroid1_hipass47_d1) =            -NV(centroid1_hipass47_a2)*ret+NV(centroid1_hipass47_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_hipass47_kcyc)==0)||(NV(centroid1_hipass47_ocut)!=cut))
      {
       NVI(centroid1_hipass47_kcyc) = 1;
        NV(centroid1_hipass47_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid1_hipass47_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid1_hipass47_b1)= -2.0F*NV(centroid1_hipass47_b0);
        NV(centroid1_hipass47_b2)= NV(centroid1_hipass47_b0);
        NV(centroid1_hipass47_a1)= 2.0F*NV(centroid1_hipass47_b0)*(c*c - 1.0F);
        NV(centroid1_hipass47_a2)= NV(centroid1_hipass47_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_hipass47_b0)*input + NV(centroid1_hipass47_d2);
     NV(centroid1_hipass47_d2) = NV(centroid1_hipass47_d1)-NV(centroid1_hipass47_a1)*ret+NV(centroid1_hipass47_b1)*input;
     NV(centroid1_hipass47_d1) =            -NV(centroid1_hipass47_a2)*ret+NV(centroid1_hipass47_b2)*input;
    }
   return((NV(centroid1_hipass47_return) = ret));

}



float centroid1__sym_lopass48(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = NV(centroid1_out_mono);
   cut =  2000.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_lopass48_b0)*input + NV(centroid1_lopass48_d2);
      NV(centroid1_lopass48_d2) = NV(centroid1_lopass48_d1)-NV(centroid1_lopass48_a1)*ret+NV(centroid1_lopass48_b1)*input;
      NV(centroid1_lopass48_d1) =            -NV(centroid1_lopass48_a2)*ret+NV(centroid1_lopass48_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_lopass48_kcyc)==0)||(NV(centroid1_lopass48_ocut)!=cut))
      {
       NVI(centroid1_lopass48_kcyc) = 1;
       NV(centroid1_lopass48_ocut) = cut;
       e = 3.141593F*cut*EV(ATIME);
       c = 0.0031416F;
       if (e < 3.27249e-05F)
         c = 30557.8F;
       else
        if (e < 1.56765F)
          c = 1.0F/(float)tan(e);
       NV(centroid1_lopass48_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
       NV(centroid1_lopass48_b1)= 2.0F*NV(centroid1_lopass48_b0);
       NV(centroid1_lopass48_b2)= NV(centroid1_lopass48_b0);
       NV(centroid1_lopass48_a1)= 2.0F*NV(centroid1_lopass48_b0)*(1.0F - c*c);
       NV(centroid1_lopass48_a2)= NV(centroid1_lopass48_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_lopass48_b0)*input + NV(centroid1_lopass48_d2);
     NV(centroid1_lopass48_d2) = NV(centroid1_lopass48_d1)-NV(centroid1_lopass48_a1)*ret+NV(centroid1_lopass48_b1)*input;
     NV(centroid1_lopass48_d1) =            -NV(centroid1_lopass48_a2)*ret+NV(centroid1_lopass48_b2)*input;
    }
   return((NV(centroid1_lopass48_return) = ret));

}



float centroid1__sym_hipass49(ENGINE_PTR_DECLARE)
{
   float cut;
   float input;
   float ret;
   float c,e;

   input = centroid1__sym_lopass48(NSP);
   cut =  235.0F ;
   if (EV(acycleidx))
    {
      ret = NV(centroid1_hipass49_b0)*input + NV(centroid1_hipass49_d2);
      NV(centroid1_hipass49_d2) = NV(centroid1_hipass49_d1)-NV(centroid1_hipass49_a1)*ret+NV(centroid1_hipass49_b1)*input;
      NV(centroid1_hipass49_d1) =            -NV(centroid1_hipass49_a2)*ret+NV(centroid1_hipass49_b2)*input;
    }
   else
    {
     if ((NVI(centroid1_hipass49_kcyc)==0)||(NV(centroid1_hipass49_ocut)!=cut))
      {
       NVI(centroid1_hipass49_kcyc) = 1;
        NV(centroid1_hipass49_ocut) = cut;
        e = 3.141593F*cut*EV(ATIME);
        c = 318.309F;
        if (e < 3.27249e-05F)
          c = 3.27249e-05F;
        else
         if (e < 1.56765F)
           c = (float)tan(e);
        NV(centroid1_hipass49_b0)= 1.0F/(1.0F + 1.414214F*c + c*c);
        NV(centroid1_hipass49_b1)= -2.0F*NV(centroid1_hipass49_b0);
        NV(centroid1_hipass49_b2)= NV(centroid1_hipass49_b0);
        NV(centroid1_hipass49_a1)= 2.0F*NV(centroid1_hipass49_b0)*(c*c - 1.0F);
        NV(centroid1_hipass49_a2)= NV(centroid1_hipass49_b0)*(1.0F - 1.414214F*c + c*c);
      }
     ret = NV(centroid1_hipass49_b0)*input + NV(centroid1_hipass49_d2);
     NV(centroid1_hipass49_d2) = NV(centroid1_hipass49_d1)-NV(centroid1_hipass49_a1)*ret+NV(centroid1_hipass49_b1)*input;
     NV(centroid1_hipass49_d1) =            -NV(centroid1_hipass49_a2)*ret+NV(centroid1_hipass49_b2)*input;
    }
   return((NV(centroid1_hipass49_return) = ret));

}


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps

void centroid_ipass(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{
   int i;

memset(&(NVU(0)), 0, centroid_ENDVAR*sizeof(NVU(0)));
memset(&(NT(0)), 0, centroid_ENDTBL*sizeof(NT(0)));
NV(centroid_sr3) =  3.0F  *  EV(ARATE) ;
 NV(centroid_acycle) =  EV(ARATE)  /  (float)  (  EV(KRATE) )
;
 NV(centroid__tvr0) =  1.0F  / NV(centroid_acycle);
 NV(centroid__tvr1) =  (  2400.0F  * NV(centroid_acycle))
;
 NV(centroid__tvr2) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr3) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr4) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr5) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr6) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr7) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr8) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr9) =  (  8.656170e+00F  * NV(centroid_acycle))
;
 NV(centroid__tvr10) =  1.0F  - NV(centroid_acycle);
 
}


#undef NS
#define NS(x)  EV(ninstr)[0].x
#undef NSP
#define NSP  ENGINE_PTR 
#undef NT
#define NT(x)  EV(ninstr)[0].t[x]
#undef NV
#define NV(x)  EV(ninstr)[0].v[x].f
#undef NVI
#define NVI(x)  EV(ninstr)[0].v[x].i
#undef NVUI
#define NVUI(x)  EV(ninstr)[0].v[x].ui
#undef NVPS
#define NVPS(x)  EV(ninstr)[0].v[x].ps
#undef NVU
#define NVU(x)  EV(ninstr)[0].v[x]
#undef NP
#define NP(x)  EV(ninstr)[0].v[x].f
#undef NPI
#define NPI(x)  EV(ninstr)[0].v[x].i
#undef NPUI
#define NPUI(x)  EV(ninstr)[0].v[x].ui
#undef NPPS
#define NPPS(x)  EV(ninstr)[0].v[x].ps

void centroid1_ipass(ENGINE_PTR_DECLARE)
{
   int i;

memset(&(NVU(0)), 0, centroid1_ENDVAR*sizeof(NVU(0)));
memset(&(NT(0)), 0, centroid1_ENDTBL*sizeof(NT(0)));
NV(centroid1_sr3) =  3.0F  *  EV(ARATE) ;
 NV(centroid1_acycle) =  EV(ARATE)  /  (float)  (  EV(KRATE) )
;
 NV(centroid1__tvr0) =  1.0F  / NV(centroid1_acycle);
 NV(centroid1__tvr1) =  (  2400.0F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr2) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr3) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr4) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr5) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr6) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr7) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr8) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr9) =  (  8.656170e+00F  * NV(centroid1_acycle))
;
 NV(centroid1__tvr10) =  1.0F  - NV(centroid1_acycle);
 
}


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps

void centroid_kpass(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{

   int i;

   NS(iline->itime) = ((float)(EV(kcycleidx) - NS(iline->kbirth)))*EV(KTIME);

 if  (  NS(iline->itime)  ==  0.0F )
 { NV(centroid_onedb_time) =  0.25F ;
 NV(centroid_capture_range) =  1.0F ;
 NV(centroid_db_per_kcycle) =  1.0F  /  (  EV(KRATE)  * NV(centroid_onedb_time))
;
 NV(centroid_tracking_range) = NV(centroid_capture_range) *  5.000000e-01F ;
 NV(centroid_quick_set) =  1.0F ;
 }
NV(centroid_sc_interp) =  0.0F ;
  if  (  ! NV(centroid_sc_update) &&  (  !  NS(iline->itime)  ||  ( NV(centroid_last_Spectral_Center) != NG(GBL_aup_Spectral_Center))
)
)
 { NV(centroid_sc_interp) =  1.0F ;
  if  ( NV(centroid_last_Spectral_Center))
 { NV(centroid_omega_mult) = ((float)(pow(NG(GBL_aup_Spectral_Center) / NV(centroid_last_Spectral_Center), NV(centroid__tvr0))));
 NV(centroid_omega_start) = NV(centroid_omega_mult) *  6.283185e+00F  * NV(centroid_last_Spectral_Center);
  if  ( NG(GBL_aup_Spectral_Center) > NV(centroid_last_Spectral_Center))
 { NV(centroid_sc_update) =  ( NG(GBL_aup_Spectral_Center) >  3.0F  * NV(centroid_last_Spectral_Center))
;
 }
 else  { NV(centroid_sc_update) =  (  3.0F  * NG(GBL_aup_Spectral_Center) < NV(centroid_last_Spectral_Center))
;
 }
}
 else  { NV(centroid_omega_mult) =  1.0F ;
 NV(centroid_omega_start) =  6.283185e+00F  * NG(GBL_aup_Spectral_Center);
 }
NV(centroid_last_Spectral_Center) = NG(GBL_aup_Spectral_Center);
 }
 if  ( NG(GBL_aup_Tracking_Time) != NV(centroid_onedb_time))
 { NV(centroid_onedb_time) = NG(GBL_aup_Tracking_Time);
 NV(centroid_db_per_kcycle) =  1.0F  /  (  EV(KRATE)  * NV(centroid_onedb_time))
;
 }
 if  ( NG(GBL_aup_Threshold) != NV(centroid_capture_range))
 { NV(centroid_capture_range) = NG(GBL_aup_Threshold);
 NV(centroid_tracking_range) = NG(GBL_aup_Threshold) *  5.000000e-01F ;
 }
NV(centroid_in_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = centroid__sym_rms4(NSP);
 NV(centroid_in_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = NV(centroid_in_log + ((int)(0.5F + (NV(centroid_bptr) ))))  * NV(centroid_in_log + ((int)(0.5F + (NV(centroid_bptr) )))) ;
 NV(centroid_lp_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = centroid__sym_rms5(NSP);
 NV(centroid_lp_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = NV(centroid_lp_log + ((int)(0.5F + (NV(centroid_bptr) ))))  * NV(centroid_lp_log + ((int)(0.5F + (NV(centroid_bptr) )))) ;
 NV(centroid_out_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = centroid__sym_rms6(NSP);
 NV(centroid_out_log + ((int)(0.5F + (NV(centroid_bptr) ))))  = NV(centroid_out_log + ((int)(0.5F + (NV(centroid_bptr) ))))  * NV(centroid_out_log + ((int)(0.5F + (NV(centroid_bptr) )))) ;
 NV(centroid_bptr) =  ( NV(centroid_bptr) ==  24.0F )
 ?  0.0F  : NV(centroid_bptr) +  1.0F ;
 NV(centroid_bidx) =  0.0F ;
 NV(centroid_in_rms) =  0.0F ;
 NV(centroid_lp_rms) =  0.0F ;
 NV(centroid_out_rms) =  0.0F ;
  while  ( NV(centroid_bidx) <  25.0F )
 { NV(centroid_in_rms) = NV(centroid_in_rms) + NV(centroid_in_log + ((int)(0.5F + (NV(centroid_bidx) )))) ;
 NV(centroid_lp_rms) = NV(centroid_lp_rms) + NV(centroid_lp_log + ((int)(0.5F + (NV(centroid_bidx) )))) ;
 NV(centroid_out_rms) = NV(centroid_out_rms) + NV(centroid_out_log + ((int)(0.5F + (NV(centroid_bidx) )))) ;
 NV(centroid_bidx) = NV(centroid_bidx) +  1.0F ;
 }
NV(centroid_in_rms) = ((float)sqrt(NV(centroid_in_rms) *  4.000000e-02F )
);
 NV(centroid_lp_rms) = ((float)sqrt(NV(centroid_lp_rms) *  4.000000e-02F )
);
 NV(centroid_out_rms) = ((float)sqrt(NV(centroid_out_rms) *  4.000000e-02F )
);
  if  ( NV(centroid_sc_update) ==  2.0F )
 { NV(centroid_in_rms) =  0.0F ;
 NV(centroid_lp_rms) =  0.0F ;
 NV(centroid_out_rms) =  0.0F ;
 NV(centroid_in_log + 0) =  0.0F ;
 NV(centroid_in_log + 1) =  0.0F ;
 NV(centroid_in_log + 2) =  0.0F ;
 NV(centroid_in_log + 3) =  0.0F ;
 NV(centroid_in_log + 4) =  0.0F ;
 NV(centroid_in_log + 5) =  0.0F ;
 NV(centroid_in_log + 6) =  0.0F ;
 NV(centroid_in_log + 7) =  0.0F ;
 NV(centroid_in_log + 8) =  0.0F ;
 NV(centroid_in_log + 9) =  0.0F ;
 NV(centroid_in_log + 10) =  0.0F ;
 NV(centroid_in_log + 11) =  0.0F ;
 NV(centroid_in_log + 12) =  0.0F ;
 NV(centroid_in_log + 13) =  0.0F ;
 NV(centroid_in_log + 14) =  0.0F ;
 NV(centroid_in_log + 15) =  0.0F ;
 NV(centroid_in_log + 16) =  0.0F ;
 NV(centroid_in_log + 17) =  0.0F ;
 NV(centroid_in_log + 18) =  0.0F ;
 NV(centroid_in_log + 19) =  0.0F ;
 NV(centroid_in_log + 20) =  0.0F ;
 NV(centroid_in_log + 21) =  0.0F ;
 NV(centroid_in_log + 22) =  0.0F ;
 NV(centroid_in_log + 23) =  0.0F ;
 NV(centroid_in_log + 24) =  0.0F ;
 NV(centroid_lp_log + 0) =  0.0F ;
 NV(centroid_lp_log + 1) =  0.0F ;
 NV(centroid_lp_log + 2) =  0.0F ;
 NV(centroid_lp_log + 3) =  0.0F ;
 NV(centroid_lp_log + 4) =  0.0F ;
 NV(centroid_lp_log + 5) =  0.0F ;
 NV(centroid_lp_log + 6) =  0.0F ;
 NV(centroid_lp_log + 7) =  0.0F ;
 NV(centroid_lp_log + 8) =  0.0F ;
 NV(centroid_lp_log + 9) =  0.0F ;
 NV(centroid_lp_log + 10) =  0.0F ;
 NV(centroid_lp_log + 11) =  0.0F ;
 NV(centroid_lp_log + 12) =  0.0F ;
 NV(centroid_lp_log + 13) =  0.0F ;
 NV(centroid_lp_log + 14) =  0.0F ;
 NV(centroid_lp_log + 15) =  0.0F ;
 NV(centroid_lp_log + 16) =  0.0F ;
 NV(centroid_lp_log + 17) =  0.0F ;
 NV(centroid_lp_log + 18) =  0.0F ;
 NV(centroid_lp_log + 19) =  0.0F ;
 NV(centroid_lp_log + 20) =  0.0F ;
 NV(centroid_lp_log + 21) =  0.0F ;
 NV(centroid_lp_log + 22) =  0.0F ;
 NV(centroid_lp_log + 23) =  0.0F ;
 NV(centroid_lp_log + 24) =  0.0F ;
 NV(centroid_out_log + 0) =  0.0F ;
 NV(centroid_out_log + 1) =  0.0F ;
 NV(centroid_out_log + 2) =  0.0F ;
 NV(centroid_out_log + 3) =  0.0F ;
 NV(centroid_out_log + 4) =  0.0F ;
 NV(centroid_out_log + 5) =  0.0F ;
 NV(centroid_out_log + 6) =  0.0F ;
 NV(centroid_out_log + 7) =  0.0F ;
 NV(centroid_out_log + 8) =  0.0F ;
 NV(centroid_out_log + 9) =  0.0F ;
 NV(centroid_out_log + 10) =  0.0F ;
 NV(centroid_out_log + 11) =  0.0F ;
 NV(centroid_out_log + 12) =  0.0F ;
 NV(centroid_out_log + 13) =  0.0F ;
 NV(centroid_out_log + 14) =  0.0F ;
 NV(centroid_out_log + 15) =  0.0F ;
 NV(centroid_out_log + 16) =  0.0F ;
 NV(centroid_out_log + 17) =  0.0F ;
 NV(centroid_out_log + 18) =  0.0F ;
 NV(centroid_out_log + 19) =  0.0F ;
 NV(centroid_out_log + 20) =  0.0F ;
 NV(centroid_out_log + 21) =  0.0F ;
 NV(centroid_out_log + 22) =  0.0F ;
 NV(centroid_out_log + 23) =  0.0F ;
 NV(centroid_out_log + 24) =  0.0F ;
 }
NV(centroid_lo_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid_lp_rms) +  9.999998e-07F )));
 NV(centroid_hi_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid_out_rms) - NV(centroid_lp_rms) +  9.999998e-07F )));
 NV(centroid_in_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid_in_rms) +  9.999998e-07F )));
 NV(centroid_out_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid_out_rms) +  9.999998e-07F )));
 NV(centroid_last_mmass) = NV(centroid_mmass);
  if  ( NV(centroid_sc_update) ==  0.0F )
 { NV(centroid_nextm) =  2.0F  * ((int)( ( NV(centroid_out_db) - NV(centroid_in_db) + centroid__sym_sgn14(NSP))
 *  5.000000e-01F )
);
  if  (  ! NV(centroid_qcount))
 { NV(centroid_mmass) = NV(centroid_mmass) - NV(centroid_mlog + ((int)(0.5F + (NV(centroid_mptr) ))))  + NV(centroid_nextm);
 NV(centroid_mlog + ((int)(0.5F + (NV(centroid_mptr) ))))  = NV(centroid_nextm);
 }
 else  { NV(centroid_midx) = NV(centroid_mptr) -  32.0F  * ((int)(NV(centroid_mptr) *  3.125000e-02F )
);
  while  ( NV(centroid_midx) <  120.0F )
 { NV(centroid_mmass) = NV(centroid_mmass) - NV(centroid_mlog + ((int)(0.5F + (NV(centroid_midx) ))))  + NV(centroid_nextm);
 NV(centroid_mlog + ((int)(0.5F + (NV(centroid_midx) ))))  = NV(centroid_nextm);
 NV(centroid_midx) = NV(centroid_midx) +  32.0F ;
 }
NV(centroid_qcount) = NV(centroid_qcount) -  1.0F ;
 }
}
 else  {  if  ( NV(centroid_sc_update) ==  1.0F )
 { NV(centroid_mmass) =  0.0F ;
 NV(centroid_mlog + 0) =  0.0F ;
 NV(centroid_mlog + 1) =  0.0F ;
 NV(centroid_mlog + 2) =  0.0F ;
 NV(centroid_mlog + 3) =  0.0F ;
 NV(centroid_mlog + 4) =  0.0F ;
 NV(centroid_mlog + 5) =  0.0F ;
 NV(centroid_mlog + 6) =  0.0F ;
 NV(centroid_mlog + 7) =  0.0F ;
 NV(centroid_mlog + 8) =  0.0F ;
 NV(centroid_mlog + 9) =  0.0F ;
 NV(centroid_mlog + 10) =  0.0F ;
 NV(centroid_mlog + 11) =  0.0F ;
 NV(centroid_mlog + 12) =  0.0F ;
 NV(centroid_mlog + 13) =  0.0F ;
 NV(centroid_mlog + 14) =  0.0F ;
 NV(centroid_mlog + 15) =  0.0F ;
 NV(centroid_mlog + 16) =  0.0F ;
 NV(centroid_mlog + 17) =  0.0F ;
 NV(centroid_mlog + 18) =  0.0F ;
 NV(centroid_mlog + 19) =  0.0F ;
 NV(centroid_mlog + 20) =  0.0F ;
 NV(centroid_mlog + 21) =  0.0F ;
 NV(centroid_mlog + 22) =  0.0F ;
 NV(centroid_mlog + 23) =  0.0F ;
 NV(centroid_mlog + 24) =  0.0F ;
 NV(centroid_mlog + 25) =  0.0F ;
 NV(centroid_mlog + 26) =  0.0F ;
 NV(centroid_mlog + 27) =  0.0F ;
 NV(centroid_mlog + 28) =  0.0F ;
 NV(centroid_mlog + 29) =  0.0F ;
 NV(centroid_mlog + 30) =  0.0F ;
 NV(centroid_mlog + 31) =  0.0F ;
 NV(centroid_mlog + 32) =  0.0F ;
 NV(centroid_mlog + 33) =  0.0F ;
 NV(centroid_mlog + 34) =  0.0F ;
 NV(centroid_mlog + 35) =  0.0F ;
 NV(centroid_mlog + 36) =  0.0F ;
 NV(centroid_mlog + 37) =  0.0F ;
 NV(centroid_mlog + 38) =  0.0F ;
 NV(centroid_mlog + 39) =  0.0F ;
 NV(centroid_mlog + 40) =  0.0F ;
 NV(centroid_mlog + 41) =  0.0F ;
 NV(centroid_mlog + 42) =  0.0F ;
 NV(centroid_mlog + 43) =  0.0F ;
 NV(centroid_mlog + 44) =  0.0F ;
 NV(centroid_mlog + 45) =  0.0F ;
 NV(centroid_mlog + 46) =  0.0F ;
 NV(centroid_mlog + 47) =  0.0F ;
 NV(centroid_mlog + 48) =  0.0F ;
 NV(centroid_mlog + 49) =  0.0F ;
 NV(centroid_mlog + 50) =  0.0F ;
 NV(centroid_mlog + 51) =  0.0F ;
 NV(centroid_mlog + 52) =  0.0F ;
 NV(centroid_mlog + 53) =  0.0F ;
 NV(centroid_mlog + 54) =  0.0F ;
 NV(centroid_mlog + 55) =  0.0F ;
 NV(centroid_mlog + 56) =  0.0F ;
 NV(centroid_mlog + 57) =  0.0F ;
 NV(centroid_mlog + 58) =  0.0F ;
 NV(centroid_mlog + 59) =  0.0F ;
 NV(centroid_mlog + 60) =  0.0F ;
 NV(centroid_mlog + 61) =  0.0F ;
 NV(centroid_mlog + 62) =  0.0F ;
 NV(centroid_mlog + 63) =  0.0F ;
 NV(centroid_mlog + 64) =  0.0F ;
 NV(centroid_mlog + 65) =  0.0F ;
 NV(centroid_mlog + 66) =  0.0F ;
 NV(centroid_mlog + 67) =  0.0F ;
 NV(centroid_mlog + 68) =  0.0F ;
 NV(centroid_mlog + 69) =  0.0F ;
 NV(centroid_mlog + 70) =  0.0F ;
 NV(centroid_mlog + 71) =  0.0F ;
 NV(centroid_mlog + 72) =  0.0F ;
 NV(centroid_mlog + 73) =  0.0F ;
 NV(centroid_mlog + 74) =  0.0F ;
 NV(centroid_mlog + 75) =  0.0F ;
 NV(centroid_mlog + 76) =  0.0F ;
 NV(centroid_mlog + 77) =  0.0F ;
 NV(centroid_mlog + 78) =  0.0F ;
 NV(centroid_mlog + 79) =  0.0F ;
 NV(centroid_mlog + 80) =  0.0F ;
 NV(centroid_mlog + 81) =  0.0F ;
 NV(centroid_mlog + 82) =  0.0F ;
 NV(centroid_mlog + 83) =  0.0F ;
 NV(centroid_mlog + 84) =  0.0F ;
 NV(centroid_mlog + 85) =  0.0F ;
 NV(centroid_mlog + 86) =  0.0F ;
 NV(centroid_mlog + 87) =  0.0F ;
 NV(centroid_mlog + 88) =  0.0F ;
 NV(centroid_mlog + 89) =  0.0F ;
 NV(centroid_mlog + 90) =  0.0F ;
 NV(centroid_mlog + 91) =  0.0F ;
 NV(centroid_mlog + 92) =  0.0F ;
 NV(centroid_mlog + 93) =  0.0F ;
 NV(centroid_mlog + 94) =  0.0F ;
 NV(centroid_mlog + 95) =  0.0F ;
 NV(centroid_mlog + 96) =  0.0F ;
 NV(centroid_mlog + 97) =  0.0F ;
 NV(centroid_mlog + 98) =  0.0F ;
 NV(centroid_mlog + 99) =  0.0F ;
 NV(centroid_mlog + 100) =  0.0F ;
 NV(centroid_mlog + 101) =  0.0F ;
 NV(centroid_mlog + 102) =  0.0F ;
 NV(centroid_mlog + 103) =  0.0F ;
 NV(centroid_mlog + 104) =  0.0F ;
 NV(centroid_mlog + 105) =  0.0F ;
 NV(centroid_mlog + 106) =  0.0F ;
 NV(centroid_mlog + 107) =  0.0F ;
 NV(centroid_mlog + 108) =  0.0F ;
 NV(centroid_mlog + 109) =  0.0F ;
 NV(centroid_mlog + 110) =  0.0F ;
 NV(centroid_mlog + 111) =  0.0F ;
 NV(centroid_mlog + 112) =  0.0F ;
 NV(centroid_mlog + 113) =  0.0F ;
 NV(centroid_mlog + 114) =  0.0F ;
 NV(centroid_mlog + 115) =  0.0F ;
 NV(centroid_mlog + 116) =  0.0F ;
 NV(centroid_mlog + 117) =  0.0F ;
 NV(centroid_mlog + 118) =  0.0F ;
 NV(centroid_mlog + 119) =  0.0F ;
 }
 else  { NV(centroid_sc_update) =  0.0F ;
 }
}
NV(centroid_mup_mult) = ((float)(pow( 10.0F ,  -  ( NV(centroid_mmass) - NV(centroid_last_mmass))
 / NV(centroid__tvr1))));
 NV(centroid_mup_start) = NV(centroid_mup_mult) * ((float)(pow( 10.0F ,  - NV(centroid_last_mmass) *  4.166667e-04F )));
 NV(centroid_mptr) =  ( NV(centroid_mptr) ==  119.0F )
 ?  0.0F  : NV(centroid_mptr) +  1.0F ;
  if  (  ( NV(centroid_lo_db) >  (-27.0F) )
 &&  ( NV(centroid_hi_db) >  (-27.0F) )
)
 {  if  ( NV(centroid_state) ==  0.0F )
 {  if  ( NV(centroid_lo_db) - NV(centroid_hi_db) > NV(centroid_capture_range))
 { NV(centroid_state) =  1.0F ;
 }
 if  ( NV(centroid_hi_db) - NV(centroid_lo_db) > NV(centroid_capture_range))
 { NV(centroid_state) =  (-1.0F) ;
 }
 if  ( NV(centroid_quick_set) && NV(centroid_state))
 { NV(centroid_quick_set) =  0.0F ;
 NV(centroid_quick_go) =  1.0F ;
 }
}
 else  {  if  ( NV(centroid_state) ==  1.0F )
 {  if  ( ((float)fabs(NV(centroid_hi_db) - NV(centroid_lo_db))
) < NV(centroid_tracking_range))
 { NV(centroid_state) =  0.0F ;
 }
 else  {  if  ( NV(centroid_hi_db) - NV(centroid_lo_db) > NV(centroid_tracking_range))
 { NV(centroid_state) =  (-1.0F) ;
 }
}
}
 else  {  if  ( ((float)fabs(NV(centroid_lo_db) - NV(centroid_hi_db))
) < NV(centroid_tracking_range))
 { NV(centroid_state) =  0.0F ;
 }
 else  {  if  ( NV(centroid_lo_db) - NV(centroid_hi_db) > NV(centroid_tracking_range))
 { NV(centroid_state) =  1.0F ;
 }
}
}
}
NV(centroid_tilt_db) = NV(centroid_tilt_db) + centroid__sym_sgn21(NSP) * NV(centroid_db_per_kcycle);
 }
 else  {  if  ( NV(centroid_tilt_db))
 {  if  ( ((float)fabs(NV(centroid_tilt_db))
) > NV(centroid_db_per_kcycle))
 { NV(centroid_tilt_db) = NV(centroid_tilt_db) - centroid__sym_sgn23(NSP) * NV(centroid_db_per_kcycle);
 }
 else  { NV(centroid_tilt_db) =  0.0F ;
 }
}
NV(centroid_state) =  0.0F ;
 NV(centroid_quick_set) =  1.0F ;
 }
 if  ( NV(centroid_sc_update) ==  1.0F )
 { NV(centroid_sc_update) =  2.0F ;
 NV(centroid_tilt_db) =  0.0F ;
 NV(centroid_state) =  0.0F ;
 }
 if  ( NV(centroid_quick_go))
 { NV(centroid_quick_go) =  0.0F ;
 NV(centroid_qcount) =  120.0F ;
 }
 if  (  !  NS(iline->itime)  ||  ( NV(centroid_last_tilt_db) != NV(centroid_tilt_db))
)
 {  if  ( NV(centroid_tilt_db) >  0.0F )
 {  if  ( NV(centroid_last_tilt_db) >  0.0F )
 { NV(centroid_lmult) = ((float)exp( (-5.0F)  *  ( NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr2))
);
 NV(centroid_hmult) = ((float)exp( ( NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr3))
);
 NV(centroid_lgain_start) = NV(centroid_lmult) * ((float)exp( (-5.0F)  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = NV(centroid_hmult) * ((float)exp(NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
 else  { NV(centroid_lmult) = ((float)exp( -  (  5.0F  * NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr4))
);
 NV(centroid_hmult) = ((float)exp( ( NV(centroid_tilt_db) -  5.0F  * NV(centroid_last_tilt_db))
 / NV(centroid__tvr5))
);
 NV(centroid_lgain_start) = NV(centroid_lmult) * ((float)exp( - NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = NV(centroid_hmult) * ((float)exp( 5.0F  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
}
 else  {  if  ( NV(centroid_last_tilt_db) >  0.0F )
 { NV(centroid_lmult) = ((float)exp( -  ( NV(centroid_tilt_db) -  5.0F  * NV(centroid_last_tilt_db))
 / NV(centroid__tvr6))
);
 NV(centroid_hmult) = ((float)exp( (  5.0F  * NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr7))
);
 NV(centroid_lgain_start) = NV(centroid_lmult) * ((float)exp( (-5.0F)  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = NV(centroid_hmult) * ((float)exp(NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
 else  { NV(centroid_lmult) = ((float)exp( -  ( NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr8))
);
 NV(centroid_hmult) = ((float)exp( 5.0F  *  ( NV(centroid_tilt_db) - NV(centroid_last_tilt_db))
 / NV(centroid__tvr9))
);
 NV(centroid_lgain_start) = NV(centroid_lmult) * ((float)exp( - NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = NV(centroid_hmult) * ((float)exp( 5.0F  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
}
NV(centroid_last_tilt_db) = NV(centroid_tilt_db);
 NG(GBL_aup_Treble_Meter) = NV(centroid_tilt_db);
 }
 else  {  if  ( NV(centroid_last_tilt_db) >  0.0F )
 {  if  (  ( NV(centroid_lmult) !=  1.0F )
 ||  ( NV(centroid_hmult) !=  1.0F )
)
 { NV(centroid_lgain_start) = ((float)exp( (-5.0F)  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = ((float)exp(NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
NV(centroid_lmult) =  1.0F ;
 NV(centroid_hmult) =  1.0F ;
 }
 else  {  if  (  ( NV(centroid_lmult) !=  1.0F )
 ||  ( NV(centroid_hmult) !=  1.0F )
)
 { NV(centroid_lgain_start) = ((float)exp( - NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid_hgain_start) = ((float)exp( 5.0F  * NV(centroid_last_tilt_db) *  1.155245e-01F )
);
 }
NV(centroid_lmult) =  1.0F ;
 NV(centroid_hmult) =  1.0F ;
 }
}

}


#undef NS
#define NS(x)  EV(ninstr)[0].x
#undef NSP
#define NSP  ENGINE_PTR 
#undef NT
#define NT(x)  EV(ninstr)[0].t[x]
#undef NV
#define NV(x)  EV(ninstr)[0].v[x].f
#undef NVI
#define NVI(x)  EV(ninstr)[0].v[x].i
#undef NVUI
#define NVUI(x)  EV(ninstr)[0].v[x].ui
#undef NVPS
#define NVPS(x)  EV(ninstr)[0].v[x].ps
#undef NVU
#define NVU(x)  EV(ninstr)[0].v[x]
#undef NP
#define NP(x)  EV(ninstr)[0].v[x].f
#undef NPI
#define NPI(x)  EV(ninstr)[0].v[x].i
#undef NPUI
#define NPUI(x)  EV(ninstr)[0].v[x].ui
#undef NPPS
#define NPPS(x)  EV(ninstr)[0].v[x].ps

void centroid1_kpass(ENGINE_PTR_DECLARE)
{

   int i;

   NS(iline->itime) = ((float)(EV(kcycleidx) - NS(iline->kbirth)))*EV(KTIME);

 if  (  NS(iline->itime)  ==  0.0F )
 { NV(centroid1_onedb_time) =  0.25F ;
 NV(centroid1_capture_range) =  1.0F ;
 NV(centroid1_db_per_kcycle) =  1.0F  /  (  EV(KRATE)  * NV(centroid1_onedb_time))
;
 NV(centroid1_tracking_range) = NV(centroid1_capture_range) *  5.000000e-01F ;
 NV(centroid1_quick_set) =  1.0F ;
 }
NV(centroid1_sc_interp) =  0.0F ;
  if  (  ! NV(centroid1_sc_update) &&  (  !  NS(iline->itime)  ||  ( NV(centroid1_last_Spectral_Center) != NG(GBL_aup_Spectral_Center))
)
)
 { NV(centroid1_sc_interp) =  1.0F ;
  if  ( NV(centroid1_last_Spectral_Center))
 { NV(centroid1_omega_mult) = ((float)(pow(NG(GBL_aup_Spectral_Center) / NV(centroid1_last_Spectral_Center), NV(centroid1__tvr0))));
 NV(centroid1_omega_start) = NV(centroid1_omega_mult) *  6.283185e+00F  * NV(centroid1_last_Spectral_Center);
  if  ( NG(GBL_aup_Spectral_Center) > NV(centroid1_last_Spectral_Center))
 { NV(centroid1_sc_update) =  ( NG(GBL_aup_Spectral_Center) >  3.0F  * NV(centroid1_last_Spectral_Center))
;
 }
 else  { NV(centroid1_sc_update) =  (  3.0F  * NG(GBL_aup_Spectral_Center) < NV(centroid1_last_Spectral_Center))
;
 }
}
 else  { NV(centroid1_omega_mult) =  1.0F ;
 NV(centroid1_omega_start) =  6.283185e+00F  * NG(GBL_aup_Spectral_Center);
 }
NV(centroid1_last_Spectral_Center) = NG(GBL_aup_Spectral_Center);
 }
 if  ( NG(GBL_aup_Tracking_Time) != NV(centroid1_onedb_time))
 { NV(centroid1_onedb_time) = NG(GBL_aup_Tracking_Time);
 NV(centroid1_db_per_kcycle) =  1.0F  /  (  EV(KRATE)  * NV(centroid1_onedb_time))
;
 }
 if  ( NG(GBL_aup_Threshold) != NV(centroid1_capture_range))
 { NV(centroid1_capture_range) = NG(GBL_aup_Threshold);
 NV(centroid1_tracking_range) = NG(GBL_aup_Threshold) *  5.000000e-01F ;
 }
NV(centroid1_in_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = centroid1__sym_rms4(NSP);
 NV(centroid1_in_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = NV(centroid1_in_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  * NV(centroid1_in_log + ((int)(0.5F + (NV(centroid1_bptr) )))) ;
 NV(centroid1_lp_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = centroid1__sym_rms5(NSP);
 NV(centroid1_lp_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = NV(centroid1_lp_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  * NV(centroid1_lp_log + ((int)(0.5F + (NV(centroid1_bptr) )))) ;
 NV(centroid1_out_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = centroid1__sym_rms6(NSP);
 NV(centroid1_out_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  = NV(centroid1_out_log + ((int)(0.5F + (NV(centroid1_bptr) ))))  * NV(centroid1_out_log + ((int)(0.5F + (NV(centroid1_bptr) )))) ;
 NV(centroid1_bptr) =  ( NV(centroid1_bptr) ==  24.0F )
 ?  0.0F  : NV(centroid1_bptr) +  1.0F ;
 NV(centroid1_bidx) =  0.0F ;
 NV(centroid1_in_rms) =  0.0F ;
 NV(centroid1_lp_rms) =  0.0F ;
 NV(centroid1_out_rms) =  0.0F ;
  while  ( NV(centroid1_bidx) <  25.0F )
 { NV(centroid1_in_rms) = NV(centroid1_in_rms) + NV(centroid1_in_log + ((int)(0.5F + (NV(centroid1_bidx) )))) ;
 NV(centroid1_lp_rms) = NV(centroid1_lp_rms) + NV(centroid1_lp_log + ((int)(0.5F + (NV(centroid1_bidx) )))) ;
 NV(centroid1_out_rms) = NV(centroid1_out_rms) + NV(centroid1_out_log + ((int)(0.5F + (NV(centroid1_bidx) )))) ;
 NV(centroid1_bidx) = NV(centroid1_bidx) +  1.0F ;
 }
NV(centroid1_in_rms) = ((float)sqrt(NV(centroid1_in_rms) *  4.000000e-02F )
);
 NV(centroid1_lp_rms) = ((float)sqrt(NV(centroid1_lp_rms) *  4.000000e-02F )
);
 NV(centroid1_out_rms) = ((float)sqrt(NV(centroid1_out_rms) *  4.000000e-02F )
);
  if  ( NV(centroid1_sc_update) ==  2.0F )
 { NV(centroid1_in_rms) =  0.0F ;
 NV(centroid1_lp_rms) =  0.0F ;
 NV(centroid1_out_rms) =  0.0F ;
 NV(centroid1_in_log + 0) =  0.0F ;
 NV(centroid1_in_log + 1) =  0.0F ;
 NV(centroid1_in_log + 2) =  0.0F ;
 NV(centroid1_in_log + 3) =  0.0F ;
 NV(centroid1_in_log + 4) =  0.0F ;
 NV(centroid1_in_log + 5) =  0.0F ;
 NV(centroid1_in_log + 6) =  0.0F ;
 NV(centroid1_in_log + 7) =  0.0F ;
 NV(centroid1_in_log + 8) =  0.0F ;
 NV(centroid1_in_log + 9) =  0.0F ;
 NV(centroid1_in_log + 10) =  0.0F ;
 NV(centroid1_in_log + 11) =  0.0F ;
 NV(centroid1_in_log + 12) =  0.0F ;
 NV(centroid1_in_log + 13) =  0.0F ;
 NV(centroid1_in_log + 14) =  0.0F ;
 NV(centroid1_in_log + 15) =  0.0F ;
 NV(centroid1_in_log + 16) =  0.0F ;
 NV(centroid1_in_log + 17) =  0.0F ;
 NV(centroid1_in_log + 18) =  0.0F ;
 NV(centroid1_in_log + 19) =  0.0F ;
 NV(centroid1_in_log + 20) =  0.0F ;
 NV(centroid1_in_log + 21) =  0.0F ;
 NV(centroid1_in_log + 22) =  0.0F ;
 NV(centroid1_in_log + 23) =  0.0F ;
 NV(centroid1_in_log + 24) =  0.0F ;
 NV(centroid1_lp_log + 0) =  0.0F ;
 NV(centroid1_lp_log + 1) =  0.0F ;
 NV(centroid1_lp_log + 2) =  0.0F ;
 NV(centroid1_lp_log + 3) =  0.0F ;
 NV(centroid1_lp_log + 4) =  0.0F ;
 NV(centroid1_lp_log + 5) =  0.0F ;
 NV(centroid1_lp_log + 6) =  0.0F ;
 NV(centroid1_lp_log + 7) =  0.0F ;
 NV(centroid1_lp_log + 8) =  0.0F ;
 NV(centroid1_lp_log + 9) =  0.0F ;
 NV(centroid1_lp_log + 10) =  0.0F ;
 NV(centroid1_lp_log + 11) =  0.0F ;
 NV(centroid1_lp_log + 12) =  0.0F ;
 NV(centroid1_lp_log + 13) =  0.0F ;
 NV(centroid1_lp_log + 14) =  0.0F ;
 NV(centroid1_lp_log + 15) =  0.0F ;
 NV(centroid1_lp_log + 16) =  0.0F ;
 NV(centroid1_lp_log + 17) =  0.0F ;
 NV(centroid1_lp_log + 18) =  0.0F ;
 NV(centroid1_lp_log + 19) =  0.0F ;
 NV(centroid1_lp_log + 20) =  0.0F ;
 NV(centroid1_lp_log + 21) =  0.0F ;
 NV(centroid1_lp_log + 22) =  0.0F ;
 NV(centroid1_lp_log + 23) =  0.0F ;
 NV(centroid1_lp_log + 24) =  0.0F ;
 NV(centroid1_out_log + 0) =  0.0F ;
 NV(centroid1_out_log + 1) =  0.0F ;
 NV(centroid1_out_log + 2) =  0.0F ;
 NV(centroid1_out_log + 3) =  0.0F ;
 NV(centroid1_out_log + 4) =  0.0F ;
 NV(centroid1_out_log + 5) =  0.0F ;
 NV(centroid1_out_log + 6) =  0.0F ;
 NV(centroid1_out_log + 7) =  0.0F ;
 NV(centroid1_out_log + 8) =  0.0F ;
 NV(centroid1_out_log + 9) =  0.0F ;
 NV(centroid1_out_log + 10) =  0.0F ;
 NV(centroid1_out_log + 11) =  0.0F ;
 NV(centroid1_out_log + 12) =  0.0F ;
 NV(centroid1_out_log + 13) =  0.0F ;
 NV(centroid1_out_log + 14) =  0.0F ;
 NV(centroid1_out_log + 15) =  0.0F ;
 NV(centroid1_out_log + 16) =  0.0F ;
 NV(centroid1_out_log + 17) =  0.0F ;
 NV(centroid1_out_log + 18) =  0.0F ;
 NV(centroid1_out_log + 19) =  0.0F ;
 NV(centroid1_out_log + 20) =  0.0F ;
 NV(centroid1_out_log + 21) =  0.0F ;
 NV(centroid1_out_log + 22) =  0.0F ;
 NV(centroid1_out_log + 23) =  0.0F ;
 NV(centroid1_out_log + 24) =  0.0F ;
 }
NV(centroid1_lo_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid1_lp_rms) +  9.999998e-07F )));
 NV(centroid1_hi_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid1_out_rms) - NV(centroid1_lp_rms) +  9.999998e-07F )));
 NV(centroid1_in_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid1_in_rms) +  9.999998e-07F )));
 NV(centroid1_out_db) = ((float)(90.0F+20.0F*(float)log10(1e-10F + NV(centroid1_out_rms) +  9.999998e-07F )));
 NV(centroid1_last_mmass) = NV(centroid1_mmass);
  if  ( NV(centroid1_sc_update) ==  0.0F )
 { NV(centroid1_nextm) =  2.0F  * ((int)( ( NV(centroid1_out_db) - NV(centroid1_in_db) + centroid1__sym_sgn14(NSP))
 *  5.000000e-01F )
);
  if  (  ! NV(centroid1_qcount))
 { NV(centroid1_mmass) = NV(centroid1_mmass) - NV(centroid1_mlog + ((int)(0.5F + (NV(centroid1_mptr) ))))  + NV(centroid1_nextm);
 NV(centroid1_mlog + ((int)(0.5F + (NV(centroid1_mptr) ))))  = NV(centroid1_nextm);
 }
 else  { NV(centroid1_midx) = NV(centroid1_mptr) -  32.0F  * ((int)(NV(centroid1_mptr) *  3.125000e-02F )
);
  while  ( NV(centroid1_midx) <  120.0F )
 { NV(centroid1_mmass) = NV(centroid1_mmass) - NV(centroid1_mlog + ((int)(0.5F + (NV(centroid1_midx) ))))  + NV(centroid1_nextm);
 NV(centroid1_mlog + ((int)(0.5F + (NV(centroid1_midx) ))))  = NV(centroid1_nextm);
 NV(centroid1_midx) = NV(centroid1_midx) +  32.0F ;
 }
NV(centroid1_qcount) = NV(centroid1_qcount) -  1.0F ;
 }
}
 else  {  if  ( NV(centroid1_sc_update) ==  1.0F )
 { NV(centroid1_mmass) =  0.0F ;
 NV(centroid1_mlog + 0) =  0.0F ;
 NV(centroid1_mlog + 1) =  0.0F ;
 NV(centroid1_mlog + 2) =  0.0F ;
 NV(centroid1_mlog + 3) =  0.0F ;
 NV(centroid1_mlog + 4) =  0.0F ;
 NV(centroid1_mlog + 5) =  0.0F ;
 NV(centroid1_mlog + 6) =  0.0F ;
 NV(centroid1_mlog + 7) =  0.0F ;
 NV(centroid1_mlog + 8) =  0.0F ;
 NV(centroid1_mlog + 9) =  0.0F ;
 NV(centroid1_mlog + 10) =  0.0F ;
 NV(centroid1_mlog + 11) =  0.0F ;
 NV(centroid1_mlog + 12) =  0.0F ;
 NV(centroid1_mlog + 13) =  0.0F ;
 NV(centroid1_mlog + 14) =  0.0F ;
 NV(centroid1_mlog + 15) =  0.0F ;
 NV(centroid1_mlog + 16) =  0.0F ;
 NV(centroid1_mlog + 17) =  0.0F ;
 NV(centroid1_mlog + 18) =  0.0F ;
 NV(centroid1_mlog + 19) =  0.0F ;
 NV(centroid1_mlog + 20) =  0.0F ;
 NV(centroid1_mlog + 21) =  0.0F ;
 NV(centroid1_mlog + 22) =  0.0F ;
 NV(centroid1_mlog + 23) =  0.0F ;
 NV(centroid1_mlog + 24) =  0.0F ;
 NV(centroid1_mlog + 25) =  0.0F ;
 NV(centroid1_mlog + 26) =  0.0F ;
 NV(centroid1_mlog + 27) =  0.0F ;
 NV(centroid1_mlog + 28) =  0.0F ;
 NV(centroid1_mlog + 29) =  0.0F ;
 NV(centroid1_mlog + 30) =  0.0F ;
 NV(centroid1_mlog + 31) =  0.0F ;
 NV(centroid1_mlog + 32) =  0.0F ;
 NV(centroid1_mlog + 33) =  0.0F ;
 NV(centroid1_mlog + 34) =  0.0F ;
 NV(centroid1_mlog + 35) =  0.0F ;
 NV(centroid1_mlog + 36) =  0.0F ;
 NV(centroid1_mlog + 37) =  0.0F ;
 NV(centroid1_mlog + 38) =  0.0F ;
 NV(centroid1_mlog + 39) =  0.0F ;
 NV(centroid1_mlog + 40) =  0.0F ;
 NV(centroid1_mlog + 41) =  0.0F ;
 NV(centroid1_mlog + 42) =  0.0F ;
 NV(centroid1_mlog + 43) =  0.0F ;
 NV(centroid1_mlog + 44) =  0.0F ;
 NV(centroid1_mlog + 45) =  0.0F ;
 NV(centroid1_mlog + 46) =  0.0F ;
 NV(centroid1_mlog + 47) =  0.0F ;
 NV(centroid1_mlog + 48) =  0.0F ;
 NV(centroid1_mlog + 49) =  0.0F ;
 NV(centroid1_mlog + 50) =  0.0F ;
 NV(centroid1_mlog + 51) =  0.0F ;
 NV(centroid1_mlog + 52) =  0.0F ;
 NV(centroid1_mlog + 53) =  0.0F ;
 NV(centroid1_mlog + 54) =  0.0F ;
 NV(centroid1_mlog + 55) =  0.0F ;
 NV(centroid1_mlog + 56) =  0.0F ;
 NV(centroid1_mlog + 57) =  0.0F ;
 NV(centroid1_mlog + 58) =  0.0F ;
 NV(centroid1_mlog + 59) =  0.0F ;
 NV(centroid1_mlog + 60) =  0.0F ;
 NV(centroid1_mlog + 61) =  0.0F ;
 NV(centroid1_mlog + 62) =  0.0F ;
 NV(centroid1_mlog + 63) =  0.0F ;
 NV(centroid1_mlog + 64) =  0.0F ;
 NV(centroid1_mlog + 65) =  0.0F ;
 NV(centroid1_mlog + 66) =  0.0F ;
 NV(centroid1_mlog + 67) =  0.0F ;
 NV(centroid1_mlog + 68) =  0.0F ;
 NV(centroid1_mlog + 69) =  0.0F ;
 NV(centroid1_mlog + 70) =  0.0F ;
 NV(centroid1_mlog + 71) =  0.0F ;
 NV(centroid1_mlog + 72) =  0.0F ;
 NV(centroid1_mlog + 73) =  0.0F ;
 NV(centroid1_mlog + 74) =  0.0F ;
 NV(centroid1_mlog + 75) =  0.0F ;
 NV(centroid1_mlog + 76) =  0.0F ;
 NV(centroid1_mlog + 77) =  0.0F ;
 NV(centroid1_mlog + 78) =  0.0F ;
 NV(centroid1_mlog + 79) =  0.0F ;
 NV(centroid1_mlog + 80) =  0.0F ;
 NV(centroid1_mlog + 81) =  0.0F ;
 NV(centroid1_mlog + 82) =  0.0F ;
 NV(centroid1_mlog + 83) =  0.0F ;
 NV(centroid1_mlog + 84) =  0.0F ;
 NV(centroid1_mlog + 85) =  0.0F ;
 NV(centroid1_mlog + 86) =  0.0F ;
 NV(centroid1_mlog + 87) =  0.0F ;
 NV(centroid1_mlog + 88) =  0.0F ;
 NV(centroid1_mlog + 89) =  0.0F ;
 NV(centroid1_mlog + 90) =  0.0F ;
 NV(centroid1_mlog + 91) =  0.0F ;
 NV(centroid1_mlog + 92) =  0.0F ;
 NV(centroid1_mlog + 93) =  0.0F ;
 NV(centroid1_mlog + 94) =  0.0F ;
 NV(centroid1_mlog + 95) =  0.0F ;
 NV(centroid1_mlog + 96) =  0.0F ;
 NV(centroid1_mlog + 97) =  0.0F ;
 NV(centroid1_mlog + 98) =  0.0F ;
 NV(centroid1_mlog + 99) =  0.0F ;
 NV(centroid1_mlog + 100) =  0.0F ;
 NV(centroid1_mlog + 101) =  0.0F ;
 NV(centroid1_mlog + 102) =  0.0F ;
 NV(centroid1_mlog + 103) =  0.0F ;
 NV(centroid1_mlog + 104) =  0.0F ;
 NV(centroid1_mlog + 105) =  0.0F ;
 NV(centroid1_mlog + 106) =  0.0F ;
 NV(centroid1_mlog + 107) =  0.0F ;
 NV(centroid1_mlog + 108) =  0.0F ;
 NV(centroid1_mlog + 109) =  0.0F ;
 NV(centroid1_mlog + 110) =  0.0F ;
 NV(centroid1_mlog + 111) =  0.0F ;
 NV(centroid1_mlog + 112) =  0.0F ;
 NV(centroid1_mlog + 113) =  0.0F ;
 NV(centroid1_mlog + 114) =  0.0F ;
 NV(centroid1_mlog + 115) =  0.0F ;
 NV(centroid1_mlog + 116) =  0.0F ;
 NV(centroid1_mlog + 117) =  0.0F ;
 NV(centroid1_mlog + 118) =  0.0F ;
 NV(centroid1_mlog + 119) =  0.0F ;
 }
 else  { NV(centroid1_sc_update) =  0.0F ;
 }
}
NV(centroid1_mup_mult) = ((float)(pow( 10.0F ,  -  ( NV(centroid1_mmass) - NV(centroid1_last_mmass))
 / NV(centroid1__tvr1))));
 NV(centroid1_mup_start) = NV(centroid1_mup_mult) * ((float)(pow( 10.0F ,  - NV(centroid1_last_mmass) *  4.166667e-04F )));
 NV(centroid1_mptr) =  ( NV(centroid1_mptr) ==  119.0F )
 ?  0.0F  : NV(centroid1_mptr) +  1.0F ;
  if  (  ( NV(centroid1_lo_db) >  (-27.0F) )
 &&  ( NV(centroid1_hi_db) >  (-27.0F) )
)
 {  if  ( NV(centroid1_state) ==  0.0F )
 {  if  ( NV(centroid1_lo_db) - NV(centroid1_hi_db) > NV(centroid1_capture_range))
 { NV(centroid1_state) =  1.0F ;
 }
 if  ( NV(centroid1_hi_db) - NV(centroid1_lo_db) > NV(centroid1_capture_range))
 { NV(centroid1_state) =  (-1.0F) ;
 }
 if  ( NV(centroid1_quick_set) && NV(centroid1_state))
 { NV(centroid1_quick_set) =  0.0F ;
 NV(centroid1_quick_go) =  1.0F ;
 }
}
 else  {  if  ( NV(centroid1_state) ==  1.0F )
 {  if  ( ((float)fabs(NV(centroid1_hi_db) - NV(centroid1_lo_db))
) < NV(centroid1_tracking_range))
 { NV(centroid1_state) =  0.0F ;
 }
 else  {  if  ( NV(centroid1_hi_db) - NV(centroid1_lo_db) > NV(centroid1_tracking_range))
 { NV(centroid1_state) =  (-1.0F) ;
 }
}
}
 else  {  if  ( ((float)fabs(NV(centroid1_lo_db) - NV(centroid1_hi_db))
) < NV(centroid1_tracking_range))
 { NV(centroid1_state) =  0.0F ;
 }
 else  {  if  ( NV(centroid1_lo_db) - NV(centroid1_hi_db) > NV(centroid1_tracking_range))
 { NV(centroid1_state) =  1.0F ;
 }
}
}
}
NV(centroid1_tilt_db) = NV(centroid1_tilt_db) + centroid1__sym_sgn21(NSP) * NV(centroid1_db_per_kcycle);
 }
 else  {  if  ( NV(centroid1_tilt_db))
 {  if  ( ((float)fabs(NV(centroid1_tilt_db))
) > NV(centroid1_db_per_kcycle))
 { NV(centroid1_tilt_db) = NV(centroid1_tilt_db) - centroid1__sym_sgn23(NSP) * NV(centroid1_db_per_kcycle);
 }
 else  { NV(centroid1_tilt_db) =  0.0F ;
 }
}
NV(centroid1_state) =  0.0F ;
 NV(centroid1_quick_set) =  1.0F ;
 }
 if  ( NV(centroid1_sc_update) ==  1.0F )
 { NV(centroid1_sc_update) =  2.0F ;
 NV(centroid1_tilt_db) =  0.0F ;
 NV(centroid1_state) =  0.0F ;
 }
 if  ( NV(centroid1_quick_go))
 { NV(centroid1_quick_go) =  0.0F ;
 NV(centroid1_qcount) =  120.0F ;
 }
 if  (  !  NS(iline->itime)  ||  ( NV(centroid1_last_tilt_db) != NV(centroid1_tilt_db))
)
 {  if  ( NV(centroid1_tilt_db) >  0.0F )
 {  if  ( NV(centroid1_last_tilt_db) >  0.0F )
 { NV(centroid1_lmult) = ((float)exp( (-5.0F)  *  ( NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr2))
);
 NV(centroid1_hmult) = ((float)exp( ( NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr3))
);
 NV(centroid1_lgain_start) = NV(centroid1_lmult) * ((float)exp( (-5.0F)  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = NV(centroid1_hmult) * ((float)exp(NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
 else  { NV(centroid1_lmult) = ((float)exp( -  (  5.0F  * NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr4))
);
 NV(centroid1_hmult) = ((float)exp( ( NV(centroid1_tilt_db) -  5.0F  * NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr5))
);
 NV(centroid1_lgain_start) = NV(centroid1_lmult) * ((float)exp( - NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = NV(centroid1_hmult) * ((float)exp( 5.0F  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
}
 else  {  if  ( NV(centroid1_last_tilt_db) >  0.0F )
 { NV(centroid1_lmult) = ((float)exp( -  ( NV(centroid1_tilt_db) -  5.0F  * NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr6))
);
 NV(centroid1_hmult) = ((float)exp( (  5.0F  * NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr7))
);
 NV(centroid1_lgain_start) = NV(centroid1_lmult) * ((float)exp( (-5.0F)  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = NV(centroid1_hmult) * ((float)exp(NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
 else  { NV(centroid1_lmult) = ((float)exp( -  ( NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr8))
);
 NV(centroid1_hmult) = ((float)exp( 5.0F  *  ( NV(centroid1_tilt_db) - NV(centroid1_last_tilt_db))
 / NV(centroid1__tvr9))
);
 NV(centroid1_lgain_start) = NV(centroid1_lmult) * ((float)exp( - NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = NV(centroid1_hmult) * ((float)exp( 5.0F  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
}
NV(centroid1_last_tilt_db) = NV(centroid1_tilt_db);
 NG(GBL_aup_Treble_Meter) = NV(centroid1_tilt_db);
 }
 else  {  if  ( NV(centroid1_last_tilt_db) >  0.0F )
 {  if  (  ( NV(centroid1_lmult) !=  1.0F )
 ||  ( NV(centroid1_hmult) !=  1.0F )
)
 { NV(centroid1_lgain_start) = ((float)exp( (-5.0F)  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = ((float)exp(NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
NV(centroid1_lmult) =  1.0F ;
 NV(centroid1_hmult) =  1.0F ;
 }
 else  {  if  (  ( NV(centroid1_lmult) !=  1.0F )
 ||  ( NV(centroid1_hmult) !=  1.0F )
)
 { NV(centroid1_lgain_start) = ((float)exp( - NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 NV(centroid1_hgain_start) = ((float)exp( 5.0F  * NV(centroid1_last_tilt_db) *  1.155245e-01F )
);
 }
NV(centroid1_lmult) =  1.0F ;
 NV(centroid1_hmult) =  1.0F ;
 }
}

}


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps

void centroid_apass(ENGINE_PTR_DECLARE_COMMA struct ninstr_types * nstate)
{

centroid__sym_rms4_spec(NSP);
;
centroid__sym_rms5_spec(NSP);
;
centroid__sym_rms6_spec(NSP);
;
 if  ( NV(centroid_acount))
 { NV(centroid_lgain) = NV(centroid_lgain) * NV(centroid_lmult);
 NV(centroid_hgain) = NV(centroid_hgain) * NV(centroid_hmult);
 NV(centroid_mup) = NV(centroid_mup) * NV(centroid_mup_mult);
  if  ( NV(centroid_sc_interp))
 { NV(centroid_omega) = NV(centroid_omega) * NV(centroid_omega_mult);
 }
NV(centroid_acount) = NV(centroid_acount) +  1.0F ;
 }
 else  { NV(centroid_lgain) = NV(centroid_lgain_start);
 NV(centroid_hgain) = NV(centroid_hgain_start);
 NV(centroid_mup) = NV(centroid_mup_start);
  if  ( NV(centroid_sc_interp))
 { NV(centroid_omega) = NV(centroid_omega_start);
 }
NV(centroid_acount) = NV(centroid__tvr10);
 }
 if  ( NV(centroid_sc_interp))
 { NV(centroid_n) =  1.0F  /  ( NV(centroid_sr3) + NV(centroid_omega))
;
 NV(centroid_a0) =  2.0F  * NV(centroid_omega) * NV(centroid_n);
 NV(centroid_b1) =  ( NV(centroid_sr3) - NV(centroid_omega))
 * NV(centroid_n);
 }
NV(centroid_lp) = NV(centroid_a0) *  0  + NV(centroid_b1) * NV(centroid_lp);
 NV(centroid_out) = NV(centroid_hgain) *  0  +  ( NV(centroid_lgain) - NV(centroid_hgain))
 * NV(centroid_lp);
 TB(BUS_output_bus + 0) += NV(centroid_mup) * NV(centroid_out);
TB(BUS_output_bus + 1) += NV(centroid_mup) * NV(centroid_out);
NV(centroid_in_mono) =  0  +  0 ;
 NV(centroid_lp_mono) =  ( NV(centroid_lp +  0 ) + NV(centroid_lp +  1 ))
 * NV(centroid_lgain);
 NV(centroid_out_mono) = NV(centroid_out +  0 ) + NV(centroid_out +  1 );
 NV(centroid_in_perceptual) = centroid__sym_hipass45(NSP);
 NV(centroid_lp_perceptual) = centroid__sym_hipass47(NSP);
 NV(centroid_out_perceptual) = centroid__sym_hipass49(NSP);
 }


#undef NS
#define NS(x)  EV(ninstr)[0].x
#undef NSP
#define NSP  ENGINE_PTR 
#undef NT
#define NT(x)  EV(ninstr)[0].t[x]
#undef NV
#define NV(x)  EV(ninstr)[0].v[x].f
#undef NVI
#define NVI(x)  EV(ninstr)[0].v[x].i
#undef NVUI
#define NVUI(x)  EV(ninstr)[0].v[x].ui
#undef NVPS
#define NVPS(x)  EV(ninstr)[0].v[x].ps
#undef NVU
#define NVU(x)  EV(ninstr)[0].v[x]
#undef NP
#define NP(x)  EV(ninstr)[0].v[x].f
#undef NPI
#define NPI(x)  EV(ninstr)[0].v[x].i
#undef NPUI
#define NPUI(x)  EV(ninstr)[0].v[x].ui
#undef NPPS
#define NPPS(x)  EV(ninstr)[0].v[x].ps

void centroid1_apass(ENGINE_PTR_DECLARE)
{

centroid1__sym_rms4_spec(NSP);
;
centroid1__sym_rms5_spec(NSP);
;
centroid1__sym_rms6_spec(NSP);
;
 if  ( NV(centroid1_acount))
 { NV(centroid1_lgain) = NV(centroid1_lgain) * NV(centroid1_lmult);
 NV(centroid1_hgain) = NV(centroid1_hgain) * NV(centroid1_hmult);
 NV(centroid1_mup) = NV(centroid1_mup) * NV(centroid1_mup_mult);
  if  ( NV(centroid1_sc_interp))
 { NV(centroid1_omega) = NV(centroid1_omega) * NV(centroid1_omega_mult);
 }
NV(centroid1_acount) = NV(centroid1_acount) +  1.0F ;
 }
 else  { NV(centroid1_lgain) = NV(centroid1_lgain_start);
 NV(centroid1_hgain) = NV(centroid1_hgain_start);
 NV(centroid1_mup) = NV(centroid1_mup_start);
  if  ( NV(centroid1_sc_interp))
 { NV(centroid1_omega) = NV(centroid1_omega_start);
 }
NV(centroid1_acount) = NV(centroid1__tvr10);
 }
 if  ( NV(centroid1_sc_interp))
 { NV(centroid1_n) =  1.0F  /  ( NV(centroid1_sr3) + NV(centroid1_omega))
;
 NV(centroid1_a0) =  2.0F  * NV(centroid1_omega) * NV(centroid1_n);
 NV(centroid1_b1) =  ( NV(centroid1_sr3) - NV(centroid1_omega))
 * NV(centroid1_n);
 }
NV(centroid1_lp + 0) = NV(centroid1_a0) *  TB(BUS_input_bus + 0)  + NV(centroid1_b1) * NV(centroid1_lp + 0);
 NV(centroid1_lp + 1) = NV(centroid1_a0) *  TB(BUS_input_bus + 1)  + NV(centroid1_b1) * NV(centroid1_lp + 1);
 NV(centroid1_out + 0) = NV(centroid1_hgain) *  TB(BUS_input_bus + 0)  +  ( NV(centroid1_lgain) - NV(centroid1_hgain))
 * NV(centroid1_lp + 0);
 NV(centroid1_out + 1) = NV(centroid1_hgain) *  TB(BUS_input_bus + 1)  +  ( NV(centroid1_lgain) - NV(centroid1_hgain))
 * NV(centroid1_lp + 1);
 TB(BUS_output_bus + 0) += NV(centroid1_mup) * NV(centroid1_out + 0);
TB(BUS_output_bus + 1) += NV(centroid1_mup) * NV(centroid1_out + 1);
NV(centroid1_in_mono) =  TB(BUS_input_bus + 0) +  TB(BUS_input_bus + 1);
 NV(centroid1_lp_mono) =  ( NV(centroid1_lp +  0 ) + NV(centroid1_lp +  1 ))
 * NV(centroid1_lgain);
 NV(centroid1_out_mono) = NV(centroid1_out +  0 ) + NV(centroid1_out +  1 );
 NV(centroid1_in_perceptual) = centroid1__sym_hipass45(NSP);
 NV(centroid1_lp_perceptual) = centroid1__sym_hipass47(NSP);
 NV(centroid1_out_perceptual) = centroid1__sym_hipass49(NSP);
 }


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps


#undef NS
#define NS(x) 0
#undef NSP
#define NSP  ENGINE_PTR_COMMA NULL
#undef NT
#define NT(x)  EV(gtables)[x]
#undef NV
#define NV(x)  EV(global)[x].f
#undef NVI
#define NVI(x)  EV(global)[x].i
#undef NVUI
#define NVUI(x)  EV(global)[x].ui
#undef NVPS
#define NVPS(x)  EV(global)[x].ps
#undef NVU
#define NVU(x)  EV(global)[x]
#undef NP
#define NP(x)  EV(global)[x].f
#undef NPI
#define NPI(x)  EV(global)[x].i
#undef NPUI
#define NPUI(x)  EV(global)[x].ui
#undef NPPS
#define NPPS(x)  EV(global)[x].ps


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps

float table_global_aup_Spectral_Parameters_pinfo[4] = { 
0.0F,0.0F,0.0F,0.0F};

float table_global_aup_Spectral_Center_pinfo[4] = { 
65.0F,650.0F,6500.0F,65.0F};

float table_global_aup_Treble_Meter_pinfo[4] = { 
-12.0F,0.0F,12.0F,-12.0F};

float table_global_aup_Servo_Loop_Parameters_pinfo[4] = { 
0.0F,0.0F,0.0F,0.0F};

float table_global_aup_Tracking_Time_pinfo[4] = { 
1.000000e-01F,2.000000e-01F,4.000000e-01F,1.000000e-01F};

float table_global_aup_Threshold_pinfo[4] = { 
1.000000e-01F,1.0F,10.0F,1.000000e-01F};



#undef NS
#define NS(x) 0
#undef NSP
#define NSP  ENGINE_PTR_COMMA NULL
#undef NT
#define NT(x)  EV(gtables)[x]
#undef NV
#define NV(x)  EV(global)[x].f
#undef NVI
#define NVI(x)  EV(global)[x].i
#undef NVUI
#define NVUI(x)  EV(global)[x].ui
#undef NVPS
#define NVPS(x)  EV(global)[x].ps
#undef NVU
#define NVU(x)  EV(global)[x]
#undef NP
#define NP(x)  EV(global)[x].f
#undef NPI
#define NPI(x)  EV(global)[x].i
#undef NPUI
#define NPUI(x)  EV(global)[x].ui
#undef NPPS
#define NPPS(x)  EV(global)[x].ps



ENGINE_PTR_TYPE system_init(int argc, char **argv, float sample_rate)
{
   int i;
   ENGINE_PTR_CREATE_SEMICOLON

   int j;

   ENGINE_PTR_NULLRETURN_SEMICOLON

   engine_init(ENGINE_PTR_COMMA sample_rate);

   srand(((unsigned int)time(0))|1);
   EV(asys_argc) = argc;
   EV(asys_argv) = argv;

   EV(csys_argc) = argc;
   EV(csys_argv) = argv;


   memset(&(NVU(GBL_STARTVAR)), 0, 
          (GBL_ENDVAR-GBL_STARTVAR)*sizeof(NVU(0)));
   memset(&(NT(0)), 0, GBL_ENDTBL*sizeof(NT(0)));
   NT(TBL_GBL_aup_Spectral_Parameters_pinfo).lenf = (float)(NT(TBL_GBL_aup_Spectral_Parameters_pinfo).len = 3);
   NT(TBL_GBL_aup_Spectral_Parameters_pinfo).tend = 2;
   NT(TBL_GBL_aup_Spectral_Parameters_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Spectral_Parameters_pinfo).t = (float *)(table_global_aup_Spectral_Parameters_pinfo);
   NT(TBL_GBL_aup_Spectral_Center_pinfo).lenf = (float)(NT(TBL_GBL_aup_Spectral_Center_pinfo).len = 3);
   NT(TBL_GBL_aup_Spectral_Center_pinfo).tend = 2;
   NT(TBL_GBL_aup_Spectral_Center_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Spectral_Center_pinfo).t = (float *)(table_global_aup_Spectral_Center_pinfo);
   NT(TBL_GBL_aup_Treble_Meter_pinfo).lenf = (float)(NT(TBL_GBL_aup_Treble_Meter_pinfo).len = 3);
   NT(TBL_GBL_aup_Treble_Meter_pinfo).tend = 2;
   NT(TBL_GBL_aup_Treble_Meter_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Treble_Meter_pinfo).t = (float *)(table_global_aup_Treble_Meter_pinfo);
   NT(TBL_GBL_aup_Servo_Loop_Parameters_pinfo).lenf = (float)(NT(TBL_GBL_aup_Servo_Loop_Parameters_pinfo).len = 3);
   NT(TBL_GBL_aup_Servo_Loop_Parameters_pinfo).tend = 2;
   NT(TBL_GBL_aup_Servo_Loop_Parameters_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Servo_Loop_Parameters_pinfo).t = (float *)(table_global_aup_Servo_Loop_Parameters_pinfo);
   NT(TBL_GBL_aup_Tracking_Time_pinfo).lenf = (float)(NT(TBL_GBL_aup_Tracking_Time_pinfo).len = 3);
   NT(TBL_GBL_aup_Tracking_Time_pinfo).tend = 2;
   NT(TBL_GBL_aup_Tracking_Time_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Tracking_Time_pinfo).t = (float *)(table_global_aup_Tracking_Time_pinfo);
   NT(TBL_GBL_aup_Threshold_pinfo).lenf = (float)(NT(TBL_GBL_aup_Threshold_pinfo).len = 3);
   NT(TBL_GBL_aup_Threshold_pinfo).tend = 2;
   NT(TBL_GBL_aup_Threshold_pinfo).oconst = 3.0F*EV(ATIME);
   NT(TBL_GBL_aup_Threshold_pinfo).t = (float *)(table_global_aup_Threshold_pinfo);

#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps

      memset(&(TB(0)), 0, ENDBUS*sizeof(TB(0)));


  ENGINE_PTR_RETURN_SEMICOLON
}


#undef NS
#define NS(x) 0
#undef NSP
#define NSP  ENGINE_PTR_COMMA NULL
#undef NT
#define NT(x)  EV(gtables)[x]
#undef NV
#define NV(x)  EV(global)[x].f
#undef NVI
#define NVI(x)  EV(global)[x].i
#undef NVUI
#define NVUI(x)  EV(global)[x].ui
#undef NVPS
#define NVPS(x)  EV(global)[x].ps
#undef NVU
#define NVU(x)  EV(global)[x]
#undef NP
#define NP(x)  EV(global)[x].f
#undef NPI
#define NPI(x)  EV(global)[x].i
#undef NPUI
#define NPUI(x)  EV(global)[x].ui
#undef NPPS
#define NPPS(x)  EV(global)[x].ps



void effects_init(ENGINE_PTR_DECLARE)
{

EV(e_centroid)[0].noteon = TOBEPLAYED;
EV(e_centroid)[0].starttime = 0.0F;
EV(e_centroid)[0].abstime = 0.0F;
EV(e_centroid)[0].released = 0;
EV(e_centroid)[0].turnoff = 0;
EV(e_centroid)[0].time = 0.0F;
EV(e_centroid)[0].itime = 0.0F;
EV(e_centroid)[0].sdur = -1.0F;
EV(e_centroid)[0].kbirth = EV(kbase);


}


#undef NS
#define NS(x) nstate->x
#undef NSP
#define NSP  ENGINE_PTR_COMMA nstate
#undef NT
#define NT(x)  nstate->t[x]
#undef NV
#define NV(x)  nstate->v[x].f
#undef NVI
#define NVI(x)  nstate->v[x].i
#undef NVUI
#define NVUI(x)  nstate->v[x].ui
#undef NVPS
#define NVPS(x)  nstate->v[x].ps
#undef NVU
#define NVU(x)  nstate->v[x]
#undef NP
#define NP(x)  nstate->v[x].f
#undef NPI
#define NPI(x)  nstate->v[x].i
#undef NPUI
#define NPUI(x)  nstate->v[x].ui
#undef NPPS
#define NPPS(x)  nstate->v[x].ps


#undef NS
#define NS(x) 0
#undef NSP
#define NSP  ENGINE_PTR_COMMA NULL
#undef NT
#define NT(x)  EV(gtables)[x]
#undef NV
#define NV(x)  EV(global)[x].f
#undef NVI
#define NVI(x)  EV(global)[x].i
#undef NVUI
#define NVUI(x)  EV(global)[x].ui
#undef NVPS
#define NVPS(x)  EV(global)[x].ps
#undef NVU
#define NVU(x)  EV(global)[x]
#undef NP
#define NP(x)  EV(global)[x].f
#undef NPI
#define NPI(x)  EV(global)[x].i
#undef NPUI
#define NPUI(x)  EV(global)[x].ui
#undef NPPS
#define NPPS(x)  EV(global)[x].ps



void shut_down(ENGINE_PTR_DECLARE)
   {

  if (EV(graceful_exit))
    {
      fprintf(stderr, "\nShutting down system ... please wait.\n");
      fprintf(stderr, "If no termination in 10 seconds, use Ctrl-C or Ctrl-\\ to force exit.\n");
      fflush(stderr);
    }
   asys_ioshutdown();

   csys_shutdown(ENGINE_PTR);

   ENGINE_PTR_DESTROY_SEMICOLON
   }

void main_apass(ENGINE_PTR_DECLARE)

{
  int busidx;
  instr_line * sysidx;

 for (sysidx=EV(cs_centroid___first);sysidx<=EV(cs_centroid___last);sysidx++)
  if (sysidx->noteon == PLAYING)
   centroid_apass(ENGINE_PTR_COMMA sysidx->nstate);
   centroid1_apass(ENGINE_PTR);

}

int main_kpass(ENGINE_PTR_DECLARE)

{
  instr_line * sysidx;

 for (sysidx=EV(cs_centroid___first);sysidx<=EV(cs_centroid___last);sysidx++)
  if (sysidx->noteon == PLAYING)
  {
   centroid_kpass(ENGINE_PTR_COMMA sysidx->nstate);
  }
   centroid1_kpass(ENGINE_PTR);

  return EV(graceful_exit);
}

void main_ipass(ENGINE_PTR_DECLARE)

{
  int i;
  instr_line * sysidx;

   unsigned char priority;
   unsigned short id;
   unsigned short label;
   unsigned int pnum;
   float * p;
   float oldtime;
   instr_line * oldest;
   int ne, newevents, netevents;
   unsigned char cmd;
   float fval;

   newevents = csys_newdata(ENGINE_PTR);
   if (newevents & CSYS_SASLEVENTS)
   do
   {
     ne = csys_saslevent(ENGINE_PTR_COMMA &cmd,&priority,&id,&label,&fval,&pnum,&p);
     switch(cmd) {
      case CSYS_SASL_ENDTIME:
       if (fval <= EV(scorebeats))
       {
         EV(endtime) = EV(scorebeats);
         EV(endkcycle) = EV(kcycleidx);
       }
       else
       {
         EV(endtime) = fval;
         EV(endkcycle) = EV(kbase) + (int)(EV(KRATE)*(EV(endtime)-EV(scorebase))*
                     (60.0F/EV(tempo)));
       }
       break;
      case CSYS_SASL_TEMPO:
       if (fval <= 0.0F)
         break;
       EV(kbase) = EV(kcycleidx);
       EV(scorebase) = EV(scorebeats);
       EV(tempo) = fval;
       EV(scoremult) = 1.666667e-02F*EV(KTIME)*EV(tempo);
       EV(endkcycle) = EV(kbase) + (int)(EV(KRATE)*(EV(endtime)-EV(scorebase))*
                   (60.0F/EV(tempo)));
       break;
      case CSYS_SASL_CONTROL:
       if (id > CSYS_SASL_NOINSTR)
         break;
       if (EV(maxsc) == CSYS_SASL_MAXCONTROL)
         break;
       EV(saslcontrol)[EV(maxsc)].id = id;
       EV(saslcontrol)[EV(maxsc)].label = label;
       EV(saslcontrol)[EV(maxsc)].fptr = pnum;
       EV(saslcontrol)[EV(maxsc)++].fval = fval;
       break;
      case CSYS_SASL_TABLE:
       if (id > GBL_ENDTBL)
         break;
       EV(sasltable)[EV(maxtb)].tgen = label;
       EV(sasltable)[EV(maxtb)].pnum = pnum;
       EV(sasltable)[EV(maxtb)].p = NULL;
       if (pnum)
        {
          EV(sasltable)[EV(maxtb)].p = (float *)
                                 calloc(pnum+1, sizeof(float));
          while ((pnum--) > 0)
            EV(sasltable)[EV(maxtb)].p[pnum] = p[pnum];
        }
       EV(sasltable)[EV(maxtb)++].index = id;
       break;
      case CSYS_SASL_INSTR:
       if (id >= CSYS_MAXSASLINSTR)
         break;
       if (*EV(cs_next)[id] == NULL)
         *EV(cs_next)[id] = *EV(cs_first)[id] = *EV(cs_last)[id];
       else
         {
           oldest = NULL;
           oldtime = MAXENDTIME;
           *EV(cs_next)[id] = *EV(cs_first)[id];
           while ((*EV(cs_next)[id]) < (*EV(cs_end)[id]))
            {
              if (((*EV(cs_next)[id])->noteon == NOTUSEDYET)||
                  ((*EV(cs_next)[id])->noteon == ALLDONE))
                break;
              if ((*EV(cs_next)[id])->starttime < oldtime)
               {
                 oldest = (*EV(cs_next)[id]);
                 oldtime = (*EV(cs_next)[id])->starttime;
               }
              (*EV(cs_next)[id])++;
            }
           if ((*EV(cs_next)[id]) == (*EV(cs_end)[id]))
            {
              *EV(cs_next)[id] = oldest;
              oldest->nstate->iline = NULL;
            }
           if ((*EV(cs_next)[id]) > (*EV(cs_last)[id]))
            *EV(cs_last)[id] = *EV(cs_next)[id];
         }
       (*EV(cs_next)[id])->starttime = EV(scorebeats) - EV(scoremult);
       (*EV(cs_next)[id])->endtime = MAXENDTIME;
       (*EV(cs_next)[id])->abstime = 0.0F;
       (*EV(cs_next)[id])->time = 0.0F;
       (*EV(cs_next)[id])->sdur = fval;
       (*EV(cs_next)[id])->released = 0;
       (*EV(cs_next)[id])->turnoff = 0;
       (*EV(cs_next)[id])->noteon = TOBEPLAYED;
       (*EV(cs_next)[id])->numchan = 0;
       (*EV(cs_next)[id])->notenum = 0;
       (*EV(cs_next)[id])->preset  = 0;
       (*EV(cs_next)[id])->label  = label;
       pnum = (pnum > MAXPFIELDS) ? MAXPFIELDS : pnum;
       while ((pnum--) > 0)
        (*EV(cs_next)[id])->p[pnum] = p[pnum];
       break;
      case CSYS_SASL_NOOP:
       break;
      default:
       break;
   }
    } while (ne != CSYS_DONE);

 for (sysidx=EV(cs_centroid___first);sysidx<=EV(cs_centroid___last);sysidx++)
  {
  switch(sysidx->noteon) {
   case PLAYING:
   if (sysidx->released)
    {
     if (sysidx->turnoff)
      {
        sysidx->noteon = ALLDONE;
        for (i = 0; i < centroid_ENDTBL; i++)
         if (sysidx->nstate->t[i].llmem)
           free(sysidx->nstate->t[i].t);
        sysidx->nstate->iline = NULL;
      }
     else
      {
        sysidx->abstime -= EV(KTIME);
        if (sysidx->abstime < 0.0F)
         {
           sysidx->noteon = ALLDONE;
           for (i = 0; i < centroid_ENDTBL; i++)
            if (sysidx->nstate->t[i].llmem)
             free(sysidx->nstate->t[i].t);
           sysidx->nstate->iline = NULL;
         }
        else
         sysidx->turnoff = sysidx->released = 0;
      }
    }
   else
    {
     if (sysidx->turnoff)
      {
       sysidx->released = 1;
      }
     else
      {
        if (sysidx->endtime <= EV(scorebeats))
         {
           if (sysidx->abstime <= 0.0F)
             sysidx->turnoff = sysidx->released = 1;
           else
           {
             sysidx->abstime -= EV(KTIME);
             if (sysidx->abstime < 0.0F)
               sysidx->turnoff = sysidx->released = 1;
           }
         }
        else
          if ((sysidx->abstime < 0.0F) &&
           (1.666667e-2F*EV(tempo)*sysidx->abstime + 
               sysidx->endtime <= EV(scorebeats)))
            sysidx->turnoff = sysidx->released = 1;
      }
    }
   break;
   case TOBEPLAYED:
    if (sysidx->starttime <= EV(scorebeats))
     {
      sysidx->noteon = PLAYING;
      sysidx->notestate = EV(nextstate);
      sysidx->nstate = &(EV(ninstr)[EV(nextstate)]);
      if (sysidx->sdur >= 0.0F)
        sysidx->endtime = EV(scorebeats) + sysidx->sdur;
      sysidx->kbirth = EV(kcycleidx);
      EV(ninstr)[EV(nextstate)].iline = sysidx;
      sysidx->time = (EV(kcycleidx)-1)*EV(KTIME);
       EV(oldstate) = EV(nextstate);
       EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
       while ((EV(oldstate) != EV(nextstate)) && 
              (EV(ninstr)[EV(nextstate)].iline != NULL))
         EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
       if (EV(oldstate) == EV(nextstate))
       {
         EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
         while ((EV(oldstate) != EV(nextstate)) && 
             (EV(ninstr)[EV(nextstate)].iline->time == 0.0F) &&
             (EV(ninstr)[EV(nextstate)].iline->noteon == PLAYING))
           EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
         EV(ninstr)[EV(nextstate)].iline->noteon = ALLDONE;
         EV(ninstr)[EV(nextstate)].iline = NULL;
       }
      centroid_ipass(ENGINE_PTR_COMMA sysidx->nstate);
    }
   break;
   default:
   break;
   }
 }
  while (EV(cs_centroid___last)->noteon == ALLDONE)
   if (EV(cs_centroid___last) == EV(cs_centroid___first))
    {
     EV(cs_centroid___first) = &EV(cs_centroid)[1];
     EV(cs_centroid___last) =  &EV(cs_centroid)[0];
     EV(cs_centroid___next) =  NULL;
     EV(cs_centroid___last)->noteon = TOBEPLAYED;
     break;
    }
   else
    EV(cs_centroid___last)--;
   for (i = 0; i < EV(maxsc); i++)
    {
      if (EV(saslcontrol)[i].id == CSYS_SASL_NOINSTR)
        {
          if (EV(saslcontrol)[i].fptr < GBL_ENDVAR)
           NG(EV(saslcontrol)[i].fptr) = EV(saslcontrol)[i].fval;
         }
       else
        {
          for (sysidx= *EV(cs_first)[EV(saslcontrol)[i].id];
               sysidx<= *EV(cs_last)[EV(saslcontrol)[i].id];sysidx++)
            if ((sysidx->label == EV(saslcontrol)[i].label)&&
                (sysidx->noteon == PLAYING))
              sysidx->nstate->v[EV(saslcontrol)[i].fptr].f
                                   = EV(saslcontrol)[i].fval;
        }
    }
   EV(maxsc) = 0;


}

void main_initpass(ENGINE_PTR_DECLARE)

{

  EV(e_centroid)[0].noteon = PLAYING;
  EV(e_centroid)[0].notestate = EV(nextstate);
  EV(e_centroid)[0].endtime = MAXENDTIME;
  EV(e_centroid)[0].nstate = &(EV(ninstr)[EV(nextstate)]);
  EV(ninstr)[EV(nextstate)].iline = &EV(e_centroid)[0];
       EV(oldstate) = EV(nextstate);
       EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
       while ((EV(oldstate) != EV(nextstate)) && 
              (EV(ninstr)[EV(nextstate)].iline != NULL))
         EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
       if (EV(oldstate) == EV(nextstate))
       {
         EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
         while ((EV(oldstate) != EV(nextstate)) && 
             (EV(ninstr)[EV(nextstate)].iline->time == 0.0F) &&
             (EV(ninstr)[EV(nextstate)].iline->noteon == PLAYING))
           EV(nextstate) = (EV(nextstate)+1) % MAXSTATE;
         EV(ninstr)[EV(nextstate)].iline->noteon = ALLDONE;
         EV(ninstr)[EV(nextstate)].iline = NULL;
       }
   centroid1_ipass(ENGINE_PTR);


   if (csys_setup(ENGINE_PTR) != CSYS_DONE)
   epr(0,NULL,NULL,"control input device unavailable");

   if (asys_iosetup((int) EV(ARATE), ASYS_ICHAN, ASYS_OCHAN, ASYS_ITYPENAME, ASYS_OTYPENAME,"audiounit", "audiounit",ASYS_TIMEOPTION) != ASYS_DONE)
   epr(0,NULL,NULL,"audio i/o device unavailable");
   EV(endkcycle) = EV(kbase) + (int) 
      (EV(KRATE)*(EV(endtime) - EV(scorebase))*(60.0F/EV(tempo)));

   ksyncinit();


}

void main_control(ENGINE_PTR_DECLARE)

{
   int i;

   for (i = 0; i < EV(maxtb); i++)
    {
      switch (EV(sasltable)[i].tgen) {
       case CSYS_SASL_TGEN_SAMPLE:
        tgen_sample(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_DATA:
        tgen_data(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_RANDOM:
        tgen_random(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_STEP:
        tgen_step(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_LINESEG:
        tgen_lineseg(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_EXPSEG:
        tgen_expseg(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_CUBICSEG:
        tgen_cubicseg(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_POLYNOMIAL:
        tgen_polynomial(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_SPLINE:
        tgen_spline(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_WINDOW  :
        tgen_window(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_HARM :
        tgen_harm(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_HARM_PHASE:
        tgen_harm_phase(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_PERIODIC:
        tgen_periodic(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_BUZZ :
        tgen_buzz(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_CONCAT:
        tgen_concat_global(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_EMPTY :
        tgen_empty(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
       case CSYS_SASL_TGEN_DESTROY:
        tgen_destroy(ENGINE_PTR_COMMA &(EV(gtables)[EV(sasltable)[i].index]),
                    EV(sasltable)[i].pnum, EV(sasltable)[i].p);
        break;
        }
    }
   EV(maxtb) = 0;

}
